/*
 * ===============================================
 * Disjunctive Propagator Implementation
 * ===============================================
 */

#include <set>
#include <string>
#include <cmath>

#include "../core/held_karp_relax.hpp"
#include "disjunctive.hpp"
#include <ilcp/cpext.h>

#define EPS_INF 1000
#define EPS_COST 0.0001



/**
 * Reserve node state/pre-processed info memory
 **/
void DisjunctivePropagator::reserve_memory() {

  //cout << "\n[Disjunctive] Reserving memory..." << endl;

  IloCP cp = get_CP();
  RevMDD* mdd = get_mdd();

  // initialize MDD node states
  // TODO: it is possible to compute more accurate bounds on the start / end time for each layer
  states = new( cp.getHeap() ) DisjunctiveState**[mdd->get_num_layers()];
  for( int l = 0; l < mdd->get_num_layers(); l++ ) {
    states[l] = new( cp.getHeap() ) DisjunctiveState*[mdd->get_max_width()];
    for( int w = 0; w < mdd->get_max_width(); w++ ) {
      if( obj_type == Makespan ) {
	states[l][w] = new( cp.getHeap() ) DisjunctiveState(cp, 
							    acts.size(), 
							    0, obj_var.getMax());
      } else {
	// TODO: compute a relaxation for max end time here
	states[l][w] = new( cp.getHeap() ) DisjunctiveState(cp, acts.size(), 
							    0, INF);
      }
    }
  }

  // initialize node sketches
  sketches = new( cp.getHeap() ) DisjunctiveSketch*[mdd->get_max_arc_id()+1];
  for( int w = 0; w < mdd->get_max_arc_id()+1; w++ ) {
    sketches[w] = new( cp.getHeap() ) DisjunctiveSketch;
  }

  // initialize group of node refinement
  // true: group A; false: group B
  group_node_ref = new(cp.getHeap()) bool[mdd->get_max_arc_id()+1];

  // initialize auxiliaries
  //num_node_checked = new( cp.getHeap() ) int[mdd->get_max_width()];
  //nodes_checked = new( cp.getHeap() ) SparseSet()

  min_est = new( cp.getHeap() ) int[mdd->get_max_width()];
  min_size_implied = new( cp.getHeap() ) int[mdd->get_max_width()];

  node_infeasible = new( cp.getHeap() ) bool[mdd->get_max_width()];
  act_checked = new( cp.getHeap() ) bool[acts.size()];

  // allocate and compute pre-processed data
  min_setup_from = new( cp.getHeap() ) int[acts.size()];
  min_setup_to = new( cp.getHeap() ) int[acts.size()];

  max_setup_from = new( cp.getHeap() ) int[acts.size()];
  max_setup_to = new( cp.getHeap() ) int[acts.size()];


  // todo: possibly disconsider depot / terminal
  for( unsigned int i = 0; i < acts.size(); i++ ) {
    min_setup_from[i] = INF;
    min_setup_to[i] = INF;
    max_setup_from[i] = -1;
    max_setup_to[i] = -1;
    for( unsigned int j = 0; j < acts.size(); j++ ) {
      if( i != j ) {
	min_setup_from[i] = MIN(min_setup_from[i], setup[i][j]);
	min_setup_to[i] = MIN(min_setup_to[i], setup[j][i]);

	if( setup[i][j] != INF ) {
	  max_setup_from[i] = MAX(max_setup_from[i], setup[i][j]);
	}

	if( setup[j][i] != INF ) {
	  max_setup_to[i] = MAX(max_setup_to[i], setup[j][i]);
	}
      }
    }
  }

  has_previous = new( cp.getHeap() ) bool[acts.size()];
  memset(has_previous, false, acts.size()*sizeof(bool));

  has_next = new( cp.getHeap() ) bool[acts.size()];
  memset(has_next, false, acts.size()*sizeof(bool));

  previous_act = new( cp.getHeap() ) int[acts.size()];
  next_act = new( cp.getHeap() ) int[acts.size()];

  precedence_status = new( cp.getHeap() ) int*[acts.size()];
  for( unsigned int i = 0; i < acts.size(); ++i ) {
    precedence_status[i] = new( cp.getHeap() ) int[acts.size()];
    memset( precedence_status[i], P_NOTCOMPUTED, sizeof(int)*acts.size() );
  }

  //mem_min_penalty = new( cp.getHeap() ) int*[mdd->get_num_layers()];
  //mem_min_tardy = new( cp.getHeap() ) int*[mdd->get_num_layers()];
  mem_min_sumsetup = new( cp.getHeap() ) double*[mdd->get_num_layers()];
  mem_min_lagrangian = new( cp.getHeap() ) double*[mdd->get_num_layers()];
  mem_min_est = new( cp.getHeap() ) int*[mdd->get_num_layers()];
  mem_max_lft = new( cp.getHeap() ) int*[mdd->get_num_layers()];
  mem_filter = new( cp.getHeap() ) double*[mdd->get_num_layers()];
  mem_max_supply = new( cp.getHeap() ) int*[mdd->get_num_layers()];
  for( int l = 0; l < mdd->get_num_layers(); ++l ) {
    //mem_min_penalty[l] = new( cp.getHeap() ) int[mdd->get_max_arc_id()+1];
    //mem_min_tardy[l] = new( cp.getHeap() ) int[mdd->get_max_arc_id()+1];
    mem_min_sumsetup[l] = new( cp.getHeap() ) double[mdd->get_max_arc_id()+1];
    mem_min_lagrangian[l] = new( cp.getHeap() ) double[mdd->get_max_arc_id()+1];
    mem_min_est[l] = new( cp.getHeap() ) int[mdd->get_max_arc_id()+1];
    mem_max_lft[l] = new( cp.getHeap() ) int[mdd->get_max_arc_id()+1];
    mem_filter[l] = new( cp.getHeap() ) double[mdd->get_max_arc_id()+1];
    mem_max_supply[l] = new( cp.getHeap() ) int[mdd->get_max_arc_id()+1];
  }

  temp_set.resize(acts.size(), false);
  aux_v.reserve(acts.size());
  may_be_refined = new( cp.getHeap() ) bool[mdd->get_max_width()];
  node_checked = new( cp.getHeap() ) bool[mdd->get_max_width()];

  exact_acts_array = new( cp.getHeap() ) int[acts.size()];
  exact_acts.reserve(acts.size());

  act_est = new( cp.getHeap() ) SetMinValueList<int>(cp, acts.size());
  act_lft = new( cp.getHeap() ) SetMaxValueList(cp, acts.size());

  is_act_exact_layer = new( cp.getHeap() ) IlcRevBool*[mdd->get_num_layers()];
  for( int l = 0; l < mdd->get_num_layers(); ++l ) {
    is_act_exact_layer[l] = new( cp.getHeap() ) IlcRevBool[acts.size()];
    for( int i = 0; i < (int)acts.size(); ++i ) {
      is_act_exact_layer[l][i].setValue(cp, false);
    }
  }

  is_act_exact = new( cp.getHeap() ) IlcRevBool[mdd->get_num_layers()];
  for( int i = 0; i < (int)acts.size(); ++i ) {
    is_act_exact[i].setValue(cp, false);
  }

  valid_acts.reserve( acts.size() );


  // Refinement
  requires_refinement = new( cp.getHeap() ) IlcRevBool[mdd->get_num_layers()];
  for( int i = 1; i < mdd->get_num_layers()-1; ++i ) {
    requires_refinement[i].setValue(cp, true);
  }
  requires_refinement[0].setValue(cp, false);
  requires_refinement[mdd->get_num_layers()-1].setValue(cp, false);

  node_ref_state = new( cp.getHeap() ) RefineState[mdd->get_max_width()];

  available_nodes.reserve(mdd->get_max_width());
  taken_nodes.reserve(mdd->get_max_width());

  set_counter = new (cp.getHeap()) SetCounterList(cp, mdd->get_max_width());

  disable_refinement.setValue(cp, false);
  //cout << "[Disjunctive] done.\n\n";

  // allocate array for arcs in the optimum path
  arcs_shortest_path = new(cp.getHeap()) RevMDD::Arc*[acts.size()];

  mult = new(cp.getHeap()) double[mdd->get_num_layers()];
  if (td_multipliers) {
    mult[0] = 1.0;
    for (int l = 1; l < mdd->get_num_layers(); ++l) {
      mult[l] = (acts.size() - l);
      cout << "mult " << l << " --> " << mult[l] << endl;
    }
  } else {
    for (int l = 0; l < mdd->get_num_layers(); ++l) {
      mult[l] = 1.0;
    }
  }
  
}



/**
 * Set initial activity states
 * TODO: do this at every search node?
 **/
void DisjunctivePropagator::set_initial_states() {

  //cout << "[Disjunctive] Warning: no initial state set" << endl;
  //cout << endl;
  return;
  //
  //	assert( (int)acts.size() + 1 == get_mdd()->get_num_layers() );

  //	IloCP cp = get_CP();
  //	RevMDD* mdd = get_mdd();

  cout << endl;
  cout << "[Disjunctive] Setting initial activity states..." << endl;

  int num_acts = acts.size();

  vector<int> est_r1(num_acts);
  vector<int> lft_r1(num_acts);
  vector<int> process_times(num_acts);
  int min_release = INF;
  int max_deadline = -1;
  vector<SpanningTree::Edge> travel_times;

  for( int i = 0; i < num_acts; ++i ) {
    est_r1[i] = acts[i]->release + acts[i]->processing;
    lft_r1[i] = acts[i]->deadline - acts[i]->processing;
    process_times[i] = acts[i]->processing;
    min_release = MIN(min_release, acts[i]->release);
    max_deadline = MAX(max_deadline, acts[i]->deadline);

    for( int j = i+1; j < num_acts; ++j ) {
      SpanningTree::Edge edge;
      edge.u = i;
      edge.v = j;
      edge.val = MIN( setup[i][j], setup[j][i] );
      travel_times.push_back(edge);
    }
  }

  // compute spanning tree relaxation
  SpanningTree spt(num_acts, travel_times);
  vector<int> tree = spt.compute();

  // compute relaxed states
  sort(est_r1.begin(), est_r1.end());
  sort(lft_r1.begin(), lft_r1.end());
  sort(process_times.begin(), process_times.end());

  DisjunctiveState* state;

  // EST relaxation
  int est_r2 = min_release;
  for( int k = 0; k < num_acts; ++k) {

    // state is always the single node of the (k+1)-th layer
    state = get_node_state(k+1, 0);

    // update relaxations
    est_r2 += process_times[k];
    if( k > 0 ) {
      est_r2 += travel_times[tree[k-1]].val;
    }

    // compute relaxation
    int est = MAX(est_r1[k], est_r2);
    //cout << "EST - layer " << k+1 << " --> " << est << endl;
    state->earliest_start.setValue(get_CP(), est);
  }

  // LFT relaxation
  if( max_deadline < INF ) {
    int lft_r2 = max_deadline;
    for( int k = num_acts-1; k >= 0; --k) {

      // state is always the single node of the k-th layer
      state = get_node_state(k, 0);

      // update relaxations
      lft_r2 -= process_times[num_acts-1-k];

      // compute relaxation
      int lft = MIN(lft_r1[k], lft_r2);
      //cout << "LFT: layer " << k << " --> " << lft << endl;
      state->latest_finish.setValue(get_CP(), lft);
    }
  }

  //cout << "[Disjunctive] done.\n\n";
  return;
}



/**
 * Filter layer in a top-down perspective
 **/
void DisjunctivePropagator::filter_layer_topdown(int layer) {

  //  if (layer == 0) {
  //cout << "Objective function: " << obj_var << endl;
  //}

  RevMDD* mdd = get_mdd();
  IloCP cp = get_CP();

  //cout << "Layer " << layer << endl;

  // pointer to the previous arc considered, for removal purposes
  RevMDD::Arc* previous = mdd->arcs_layer[layer].get_header();

  // remove infeasible arcs from layer
  for( RevMDD::Arc* arc = mdd->arcs_layer[layer].get_first();
       arc != NULL;
       arc = mdd->arcs_layer[layer].get_next(arc) )
    {
      if (is_incoming_infeasible(layer, arc)) {
	// arc is infeasible: remove from layer
	previous->next.setValue(cp, arc->next.getValue());
	//cout << "\tEliminated " << arc->get_source() << "," << arc->get_val() << endl;
      } else {
	previous = arc;	
      }
    }
}


/**
 * Filter layer in a bottom up perspective
 **/
void DisjunctivePropagator::filter_layer_bottomup(int layer) {
  RevMDD* mdd = get_mdd();
  IloCP cp = get_CP();

  //cout << "\tLayer " << layer << endl;

  // pointer to the previous arc considered, for removal purposes
  RevMDD::Arc* previous = mdd->arcs_layer[layer].get_header();

  // remove infeasible arcs from layer
  for( RevMDD::Arc* arc = mdd->arcs_layer[layer].get_first();
       arc != NULL;
       arc = mdd->arcs_layer[layer].get_next(arc) )
    {
      if (is_outgoing_infeasible(layer, arc)) {
	// arc is infeasible: remove from layer
	previous->next.setValue(cp, arc->next.getValue());
	//cout << "\t\t[BT] Eliminated arc " << layer << "," << arc->get_val() << endl;
      } else {
	previous = arc;
	//cout << "\t\tDid NOT eliminate arc " << layer << "," << arc->get_val() << endl;
      }
    }

  //cout << "\tEnd layer " << layer << endl;
}


/**
 * Partition refinement group (refinement). Max MDD width must be observed
 **/
void DisjunctivePropagator::partition_refinement_group(
						       int prev_layer,
						       RefinementGroup *ref_groups,
						       int &n_groups)
{
  //if (disable_refinement.getValue()) {
  //  return;
  //}

  // -----------------------------------------------------------
  // Begin Node Branching
  // -----------------------------------------------------------

  // Check exactness with respect to the alldiff conditions
  if (!is_layer_exact_v[prev_layer+1]->getValue()) {
    // alldiff-based refinement
    partition_exact_acts(prev_layer, ref_groups, n_groups);
  }

  // if it is the last layer to be refined, check if MDD
  // is already exact by verifying if all layers are exact
  if (prev_layer+1 == get_mdd()->get_num_layers()-2 && !is_mdd_exact_v.getValue()) {
    bool all_exact = true;
    for (int l = 1; l < get_mdd()->get_num_layers()-1 && all_exact; ++l) {
      all_exact = is_layer_exact_v[l]->getValue();
    }
    if (all_exact) {
      is_mdd_exact_v.setValue(get_CP(), true);
      //cout << "MDD is exact!" << endl;
    }
  }
}



/**
 * Partition refinement group based on full extension
 **/
void DisjunctivePropagator::partition_full_extension(
						     int layer,
						     RefinementGroup *ref_groups,
						     int &n_groups)
{
  RevMDD* mdd = get_mdd();

  // compute how many extra nodes we require to perform full
  // extension of the layer

  int extra_nodes = 0;

  available_nodes.clear();
  for( int g = 0; g < mdd->get_max_width(); g++ ) {
    if( ref_groups[g]->size == 0 ) {
      available_nodes.push_back(g);
    } else {
      extra_nodes += ref_groups[g]->size - 1;
    }
  }

  if( extra_nodes <= (int)available_nodes.size() ) {
    // perform full extension; layer does not require any more refinement
    for( int g = 0; g < mdd->get_max_width(); g++ ) {
      while( ref_groups[g]->size > 1 ) {
	assert( available_nodes.size() > 0 );

	// transfer arc to new node
	ref_groups[available_nodes.back()]->add( ref_groups[g]->pop_up() );

	// update available nodes and number of groups
	available_nodes.pop_back();
	n_groups++;
      }
    }
    // previous layer must not require refinement as well
    if( !requires_refinement[layer].getValue() ) {
      requires_refinement[layer+1].setValue(get_CP(), false);
    }
  }
}



/**
 * Partition refinement group based on exact activities
 **/
void DisjunctivePropagator::partition_exact_acts(int layer, RefinementGroup *ref_groups, int &n_groups) {
  //cout << "[Disjunctive] Refinement - layer=" << layer;
  //cout << " - n_groups=" << n_groups << endl;

  RevMDD* mdd = get_mdd();
  assert (n_groups < mdd->get_max_width());

  // comparator for groups according to earliest start time (EST)
  Arc_Est_Comparator group_comp(sketches);

  // check nodes that are available or that can be refined
  available_nodes.clear();
  for (int g = 0; g < mdd->get_max_width(); ++g) {
    if (ref_groups[g]->is_empty()) {
      available_nodes.push_back(g);
      //cout << "\tGroup " << g << " is available" << endl;

    } else {
      if (ref_groups[g]->size > 1) {
	min_est[g] = INF;
	for (int i = 0; i < ref_groups[g]->size; ++i) {
	  min_est[g] = MIN(min_est[g],
			   sketches[ref_groups[g]->elements[i]->get_layer_id()]->min_sumsetup_R);
	}
	//cout << "\tGroup " << g << " may be refined; size=" << ref_groups[g]->size << endl;
      }
    }
  }

  //cout << "Refinement layer " << layer << endl;

  // priority queue of nodes to refine
  GroupESTComp node_comp(min_est);
  priority_queue<int, vector<int>, GroupESTComp> refine_set(node_comp);

  // Main loop: iterate on activity that must be exact in layer. Priority is given to
  // order defined in 'exact_acts'
  for (vector<int>::iterator it = exact_acts.begin(); it != exact_acts.end(); ++it) {
    if (is_act_exact_layer[layer][*it].getValue()) {
      continue;
    }

    int act = *it;

    //cout << "\tActivity " << act << endl;

    //cout << "\n\tChecking activity " << act << endl;

    // if layer can be completely refined w.r.t. act
    bool completely_refinable = true;
    bool all_terminals;

    priority_queue<int, vector<int>, GroupESTComp> refine_set(node_comp);

    // check if target nodes at next layer can be refined
    int groupA, groupB;
    for (int target_node = 0; target_node < mdd->get_max_width(); ++target_node) {      

      // do not consider sink for optional nodes
      if (optional) {
	all_terminals = true;
	for (int i = 0; i < ref_groups[target_node]->size && all_terminals; ++i) {
	  all_terminals = (ref_groups[target_node]->elements[i]->get_val() == terminal_opt_act);
	}
	if (all_terminals) {
	  continue;
	}
      }

      //cout << "\tTarget - node=" << target_node << " - size=" << ref_groups[target_node]->size << endl;
      if (ref_groups[target_node]->size > 1) {     
	//cout << "\tChecking target: " << target_node << endl;
	
	// count how many arcs each new node would have
	groupA = 0;
	groupB = 0;
	
	for (int i = 0; i < ref_groups[target_node]->size; ++i) {
	  RevMDD::Arc* arc = ref_groups[target_node]->elements[i];	
	  //cout << "\t\t" << "arc source=" << arc->get_source() << " - val=" << arc->get_val() << endl;
	  if (arc->get_val() == act) {
	    groupA++;
	    group_node_ref[arc->get_layer_id()] = true;
	
	  } else if (get_node_state(layer, arc->get_source())->all_paths[act]) {
	    groupA++;
	    group_node_ref[arc->get_layer_id()] = true;
	  
	  } else if (!get_node_state(layer, arc->get_source())->some_paths[act]) {
	    groupB++;
	    group_node_ref[arc->get_layer_id()] = false;
	  
	  } else {
	    // there is a node that cannot be refined
	    //cout << "\t\tCannot be refined!" << endl;
	    groupA = 0;
	    completely_refinable = false;
	    break;
	  }
	}
	//cout << "\t\tgroupA= " << groupA << " - groupB=" << groupB << endl;
	if (groupA > 0 && groupB > 0) {
	  //cout << "\t\tAdded node " << target_node << " to refinement list" << endl;
	  refine_set.push(target_node);
	}
      }
    }
   
    // refinement
    while (!refine_set.empty()) {

      // take group with smallest objective from refine list
      int to_refine = refine_set.top();
      refine_set.pop();

      // take an available node
      int new_group = available_nodes.back();
      available_nodes.pop_back();

      // transfer arcs to new node
      RevMDD::Arc* arc;
      for( int i = 0; i < ref_groups[to_refine]->size; ) {
	arc = ref_groups[to_refine]->elements[i];
	if (group_node_ref[arc->get_layer_id()]) {
	  ref_groups[new_group]->add(arc);
	  ref_groups[to_refine]->elements[i] = ref_groups[to_refine]->pop_up();
	} else {
	  ++i;
	}
      }
      assert( ref_groups[to_refine]->size > 0 );
      assert( ref_groups[new_group]->size > 0 );
      //cout << "\t\tNode " << to_refine << " split to node " << new_group << endl;

      //cout << "sizeA: " << ref_groups[to_refine]->size  << endl;
      //cout << "sizeB: " << ref_groups[new_group]->size  << endl;
      if( ref_groups[to_refine]->size == 0 || ref_groups[new_group]->size == 0 ) {
	cout << "ERROR: refinement group has no arcs" << endl;
	exit(1);
      }

      //print_group(to_refine, ref_groups[to_refine]);
      //print_group(new_group, ref_groups[new_group]);

      // update number of groups
      n_groups++;

      // if width is met, refinement is concluded
      if( available_nodes.size() == 0 ) {
	return;
      }

      // update min_est of modified nodes
      if( ref_groups[new_group]->size > 1 ) {
	min_est[new_group] = INF;
	for( int i = 0; i < ref_groups[new_group]->size; ++i ) {
	  min_est[new_group] = MIN( min_est[new_group],
				    sketches[ref_groups[new_group]->elements[i]->get_layer_id()]->min_sumsetup_R );
	}
      }

      if( ref_groups[to_refine]->size > 1 ) {
	min_est[to_refine] = INF;
	for( int i = 0; i < ref_groups[to_refine]->size; ++i ) {
	  min_est[to_refine] = MIN( min_est[to_refine],
				    sketches[ref_groups[to_refine]->elements[i]->get_layer_id()]->min_sumsetup_R );
	}
      }
    }

    if (completely_refinable) {
      is_act_exact_layer[layer][act].setValue(get_CP(), true);
    }
  }

  // if this stage of the code was reached, no nodes need to be refined in this layer.
  // However, it is only exact if previous layer is exact as well
  if (is_layer_exact_v[layer]->getValue()) {
    //cout << "Layer " << layer+1 << " is now exact!" << endl;
    is_layer_exact_v[layer+1]->setValue(get_CP(), true);
  }
  
  return;
}


/**
 * Partition refinement group based exclusively on EST
 **/
void DisjunctivePropagator::partition_est(int layer, RefinementGroup *ref_groups, int &n_groups) {


  // RevMDD* mdd = get_mdd();
  // assert( n_groups < mdd->get_max_width( ));

  // /**
  //  * Check if it is possible to create one node for each arc sketch
  //  * in the refinement group.
  //  */

  // // Count total number of sketches and the number of available nodes
  // available_nodes.clear();
  // int total_size = 0;
  // for( int g = 0; g < mdd->get_max_width(); g++ ) {

  //   if( ref_groups[g]->is_empty() ) {
  //     // if group is empty, add to available nodes
  //     available_nodes.push_back(g);

  //   } else {
  //     // if group is not empty, count how many sketches we have to move
  //     if( ref_groups[g]->size > 1 ) {
  // 	total_size += ref_groups[g]->size - 1;
  //     }
  //   }
  // }

  // //	// If there is available space, create one node for each arc sketch
  // //	if( total_size <= (int) available_nodes.size() ) {
  // //		//cout << "Also full extension as well!!!" << endl;
  // //		for( int g = 0; g < mdd->get_max_width(); g++ ) {
  // //			while( ref_groups[g]->size > 1 ) {
  // //				assert( available_nodes.size() > 0 );
  // //
  // //				// transfer arc to new node
  // //				ref_groups[available_nodes.back()]->add( ref_groups[g]->pop_up() );
  // //
  // //				// update available nodes and number of groups
  // //				available_nodes.pop_back();
  // //				n_groups++;
  // //			}
  // //		}
  // //		return;
  // //	}

  // // TODO: test this!
  // //return;

  // /*
  //  * ===============================================
  //  *  Create priority queue of groups to refine
  //  * ===============================================
  //  */

  // RefGroup_Est_Comparator refgroup_comparator(min_est, min_size_implied);
  // priority_queue<int, vector<int>, RefGroup_Est_Comparator> group_queue(refgroup_comparator);

  // // reset node availability
  // available_nodes.clear();

  // // create comparator for groups according to earliest start time (EST)
  // Arc_Est_Comparator group_comp(sketches);

  // // sort each group according to lower bound
  // for( int g = 0; g < mdd->get_max_width(); g++ ) {

  //   if( ref_groups[g]->is_empty() ) {

  //     // if group is empty, add to available nodes
  //     available_nodes.push_back(g);

  //   } else {

  //     // only consider groups with more than one arc, since they are the ones that can be
  //     // refined

  //     if( ref_groups[g]->size > 1 ) {

  // 	// sort group
  // 	sort(ref_groups[g]->elements, ref_groups[g]->elements+ref_groups[g]->size, group_comp);
  // 	//min_dist[g] = get_min_dist_R(ref_groups[g]);
  // 	min_est[g] = get_min_est(ref_groups[g]);
  // 	min_size_implied[g] = get_min_all_paths(ref_groups[g]);

  // 	// print
  // 	//print_group(g, ref_groups[g]);

  // 	// add group to queue if min_size of size implied are distinct
  // 	if( is_est_distinct(ref_groups[g]) || is_min_all_paths_distinct(ref_groups[g]) ) {
  // 	  group_queue.push(g);

  // 	  //cout << "\t\tadded group to queue" << endl;

  // 	}

  //     }
  //   }
  // }

  // /*
  //  * =====================================================
  //  *  Refine until width is met or no more partitions
  //  *  are possible
  //  * =====================================================
  //  */

  // int to_refine, new_group;
  // while( n_groups < mdd->get_max_width() && !group_queue.empty() ) {

  //   // select group to refine
  //   to_refine = group_queue.top();
  //   group_queue.pop();

  //   //cout << "\n\tGroup to refine: " << to_refine << endl;

  //   // take new node
  //   assert( !available_nodes.empty() );
  //   new_group = available_nodes.back();
  //   available_nodes.pop_back();

  //   // increment number of existing groups
  //   n_groups++;

  //   // auxiliary
  //   RevMDD::Arc* arc;

  //   if( is_est_distinct(ref_groups[to_refine]) ) {

  //     /*
  //      * Refinement by min_dist_R
  //      */

  //     // transfer arcs
  //     while( sketches[ref_groups[to_refine]->get_last()->get_layer_id()]->earliest_start == min_est[to_refine] ) {

  // 	// take transfer arc
  // 	arc = ref_groups[to_refine]->pop_up();

  // 	// mark source node as modified
  // 	mdd->modified_out_node(layer, arc->get_source());

  // 	// add to new group
  // 	ref_groups[new_group]->add( arc );
  //     }

  //     assert( ref_groups[to_refine]->size > 0 );
  //     assert( ref_groups[new_group]->size > 0 );

  //   } else {

  //     /*
  //      * Refinement by implied up
  //      */

  //     // transfer arcs
  //     while( sketches[ref_groups[to_refine]->get_last()->get_layer_id()]->size_implied == min_size_implied[to_refine] ) {

  // 	// take transfer arc
  // 	arc = ref_groups[to_refine]->pop_up();

  // 	// mark source node as modified
  // 	mdd->modified_out_node(layer, arc->get_source());

  // 	// add to new group
  // 	ref_groups[new_group]->add( arc );
  //     }

  //     assert( ref_groups[to_refine]->size > 0 );
  //     assert( ref_groups[new_group]->size > 0 );
  //   }

  //   // sort new group
  //   sort(ref_groups[new_group]->elements, ref_groups[new_group]->elements+ref_groups[new_group]->size, group_comp);

  //   // update groups sketch info

  //   min_est[to_refine] = get_min_est(ref_groups[to_refine]);
  //   min_size_implied[to_refine] = get_min_all_paths(ref_groups[to_refine]);

  //   min_est[new_group] = get_min_est(ref_groups[new_group]);
  //   min_size_implied[new_group] = get_min_all_paths(ref_groups[new_group]);


  //   //		// print
  //   //		print_group(to_refine, ref_groups[to_refine]);
  //   //		print_group(new_group, ref_groups[new_group]);

  //   // add groups to queue if they can still be refined
  //   if( is_est_distinct(ref_groups[new_group]) || is_min_all_paths_distinct(ref_groups[new_group]) ) {
  //     group_queue.push(new_group);

  //     //			cout << "\t\tadded group " << new_group << " to queue" << endl;

  //   }

  //   //cout << endl;

  //   if( is_est_distinct(ref_groups[to_refine]) || is_min_all_paths_distinct(ref_groups[to_refine]) ) {
  //     group_queue.push(to_refine);

  //     //cout << "\t\tadded group " << to_refine << " to queue" << endl;

  //   }

  //   //cout << endl;
  // }

  // //cout << "refinement completed" << endl << endl;
  // //exit(1);
}



/**
 * Set the state of a node according to its incoming arcs.
 **/
void DisjunctivePropagator::setup_node_incoming(int layer, int node, RefinementGroup ref_group, bool new_node) {
  
  RevMDD* mdd = get_mdd();
  IloCP cp = get_CP();

  // initialize state of the new node
  DisjunctiveState* state = get_node_state(layer, node);

  state->some_paths.reset();
  state->all_paths.set();
  state->in_setup->clear();
  state->in_lagrangian->clear();
  state->in_est->clear();
  state->in_max_supply->clear();

  // if new node, we initialize the disjunctive states with the original node states
  if( new_node ) {
    DisjunctiveState* original_state = get_node_state(layer, ref_group->get_first()->get_target());
    state->min_sumsetup_T.setValue(cp, original_state->min_sumsetup_T.getValue());
    state->min_lagrangian_T.setValue(cp, original_state->min_lagrangian_T.getValue());
    state->earliest_start.setValue(cp, original_state->earliest_start.getValue());
    state->max_supply_T.setValue(cp, original_state->max_supply_T.getValue());
  }

  // initialize auxiliaries
  //memset( num_node_checked, 0, sizeof(int)*mdd->get_max_width() );
  set_counter->clear();

  double min_sum_setup_R = INF;
  double min_lagrangian_R = INF;
  int est = INF;
  int max_supply = -1;

  RevMDD::Arc* arc;
  for( int i = 0; i < ref_group->size; i++ ) {
    arc = ref_group->elements[i];

    // sum setup update
    min_sum_setup_R = MIN( min_sum_setup_R, 
			   sketches[arc->get_layer_id()]->min_sumsetup_R );
    state->in_setup->add(arc->get_val(), 
			 sketches[arc->get_layer_id()]->min_sumsetup_R); 
    
    // lagrangian update
    min_lagrangian_R = MIN( min_lagrangian_R, 
			    sketches[arc->get_layer_id()]->min_lagrangian_R );
    state->in_lagrangian->add(arc->get_val(), 
			      sketches[arc->get_layer_id()]->min_lagrangian_R); 
  
    // earliest start time update
    est = MIN( est, sketches[arc->get_layer_id()]->earliest_start ); 
    state->in_est->add(arc->get_val(), 
		       sketches[arc->get_layer_id()]->earliest_start);

    // max supply
    if (has_supply) {
      max_supply = MAX( max_supply, sketches[arc->get_layer_id()]->max_supply );
      state->in_max_supply->add(arc->get_val(),
				sketches[arc->get_layer_id()]->max_supply);
    }

    // some_paths update: value will always belong to state
    state->some_paths.set(arc->get_val(), true);

    // check how many times node was verified
    switch( set_counter->get_counter(arc->get_source()) ) {

    case 0:
      // update 'some_paths' state
      state->some_paths |= get_node_state(layer-1, arc->get_source())->some_paths;
      // update 'all_paths' state
      if( state->all_paths[arc->get_val()] ) {
	state->all_paths &= get_node_state(layer-1, arc->get_source())->all_paths;
	state->all_paths.set(arc->get_val(), true);

      } else {
	state->all_paths &= get_node_state(layer-1, arc->get_source())->all_paths;
      }
      set_counter->add(arc->get_source());
      break;

    case 1:
      // just intersect 'all_paths'
      state->all_paths &= get_node_state(layer-1, arc->get_source())->all_paths;
      set_counter->add(arc->get_source());
      break;

    default:
      break;
    }
  }

  // verify if some_paths defines a hall set
  state->hall_set = ((int)state->some_paths.count() == layer);

  // setup states
  state->min_sumsetup_R.setValue(cp, min_sum_setup_R);
  state->min_lagrangian_R.setValue(cp, min_lagrangian_R);
  state->earliest_start.setValue(cp, est);
  state->max_supply_R.setValue(cp, max_supply);

  // if last layer, update bounds
  if( layer == mdd->get_num_layers()-1 ) {
    if( state->min_sumsetup_R.getValue() >= obj_var.getMin() + EPS_COST ) {
      cp.add( obj_var >= state->min_sumsetup_R.getValue() );
    }
    if (obj_var_lag.getImpl() != NULL) {      
      int obj = floor(state->min_lagrangian_R.getValue());
      if( obj > obj_var_lag.getMin() ) {
       	cp.add( obj_var_lag >= obj );
      }
    }
    // cout << "\t\tobj_var = " << obj_var << endl;
    // if (obj_var_lag.getImpl() != NULL) {
    //   cout << "\t\tlagrangian = " << obj_var_lag << endl;
    // }    
  }
}




/**
 * Set the state of a node according to its outgoing arcs.
 **/
void DisjunctivePropagator::setup_node_outgoing(int layer, int node, RefinementGroup ref_group, bool new_node) {

  RevMDD* mdd = get_mdd();
  IloCP cp = get_CP();

  // initialize state of the new node
  DisjunctiveState* state = get_node_state(layer, node);

  state->some_paths.reset();
  state->all_paths.set();
  state->out_setup->clear();
  state->out_lagrangian->clear();
  state->out_lft->clear();
  state->out_max_supply->clear();

  double min_sum_setup_T = INF;
  double min_lagrangian_T = INF;
  int lft = -1;
  int max_supply = -1;

  // initialize auxiliaries
  //memset( num_node_checked, 0, sizeof(int)*mdd->get_max_width() );
  set_counter->clear();

  RevMDD::Arc* arc;
  for( int i = 0; i < ref_group->size; i++ ) {
    arc = ref_group->elements[i];

    // lft update
    lft = MAX( lft, sketches[arc->get_layer_id()]->latest_finish );
    state->out_lft->add(arc->get_val(), 
			  sketches[arc->get_layer_id()]->latest_finish);  
   
    // sum setup update
    min_sum_setup_T = MIN( min_sum_setup_T, sketches[arc->get_layer_id()]->min_sumsetup_T );
    state->out_setup->add(arc->get_val(), 
			  sketches[arc->get_layer_id()]->min_sumsetup_T);

    // lagrangian update
    min_lagrangian_T = MIN( min_lagrangian_T, sketches[arc->get_layer_id()]->min_lagrangian_T );
    state->out_lagrangian->add(arc->get_val(), 
			       sketches[arc->get_layer_id()]->min_lagrangian_T);

    // max supply
    if (has_supply) {
      max_supply = MAX( max_supply, sketches[arc->get_layer_id()]->max_supply );
      state->out_max_supply->add(arc->get_val(), 
				 sketches[arc->get_layer_id()]->max_supply);
    }

    // some_paths update: value will always belong to state
    state->some_paths.set(arc->get_val(), true);

    // check how many times node was verified
    switch( set_counter->get_counter(arc->get_target()) ) {
    case 0:
      // update 'some_paths' state
      state->some_paths |= get_node_state(layer+1, arc->get_target())->some_paths;
      // update 'all_paths' state
      if( state->all_paths[arc->get_val()] ) {
	state->all_paths &= get_node_state(layer+1, arc->get_target())->all_paths;
	state->all_paths.set(arc->get_val(), true);
      } else {
	state->all_paths &= get_node_state(layer+1, arc->get_target())->all_paths;
      }
      //num_node_checked[arc->get_target()]++;
      set_counter->add(arc->get_target());
      break;

    case 1:
      // just intersect 'all_paths'
      state->all_paths &= get_node_state(layer+1, arc->get_target())->all_paths;
      //num_node_checked[arc->get_target()]++;
      set_counter->add(arc->get_target());
      break;
    default:
      break;
    }
  }

  // verify if some_paths defines a hall set
  state->hall_set = ((int)state->some_paths.count() == ( mdd->get_num_layers() - (layer+1) ) );

  // setup states
  state->min_sumsetup_T.setValue(cp, min_sum_setup_T);
  state->min_lagrangian_T.setValue(cp, min_lagrangian_T);
  state->latest_finish.setValue(cp, lft);
  state->max_supply_T.setValue(cp, max_supply);

  // update bounds if first layer
  if( layer == 0 ) {
    //cout << "\tbound from bottom = " << state->min_sumsetup_T.getValue() << endl;
    if( state->min_sumsetup_T.getValue() >= obj_var.getMin()+EPS_COST ) {
      cp.add( obj_var >= state->min_sumsetup_T.getValue() );
    }
    if (obj_var_lag.getImpl() != NULL) {
      int obj = floor(state->min_lagrangian_T.getValue());
      if( obj > obj_var_lag.getMin() ) {
       	cp.add( obj_var_lag >= obj );
      }
    }
    // cout << "\t\tobj_var = " << obj_var << endl;
    // if (obj_var_lag.getImpl() != NULL) {
    //   cout << "\t\tlagrangian = " << obj_var_lag << endl;
    // }    
  }

  //cout << "layer=" << layer << " - node=" << node << " - max finish=" << lft << endl;
}



/**
 * Check if an in arc is not necessarily infeasible. It also
 * sets the corresponding arc sketch.
 *
 **/
bool DisjunctivePropagator::is_incoming_infeasible(int layer, RevMDD::Arc* arc) {
  if( is_incoming_infeasible_alldiff(layer, arc) ) {
    return true;
  }
  if( is_incoming_infeasible_disjunctive(layer, arc) ) {
    return true;
  }
  return false;
}


/**
 * Check if an in arc is not necessarily infeasible according to alldiff. It also
 * sets the corresponding arc sketch.
 *
 **/
inline bool DisjunctivePropagator::is_incoming_infeasible_alldiff(int layer, RevMDD::Arc* arc) {

  DisjunctiveState* source_state = get_node_state(layer, arc->get_source());


  // --------------------------------------------------
  // Optional conditions
  // --------------------------------------------------
  if (optional) {

    // terminal was already set
    if (source_state->all_paths[terminal_opt_act]) {
      return (arc->get_val() != terminal_opt_act);
    }

    // activity is not available to the resource
    if (opt_acts[arc->get_val()].isAbsent()) {
      return true;
    }

    // number of assigned activities to this resource
    if (arc->get_val() == terminal_opt_act) {
      if (num_optionals.getMax() < opt_acts.getSize() - layer - 1) {
	//cout << "\tLayer " << layer << " - num_acts = " << opt_acts.getSize() << " - opt = " << num_optionals << " - infeasible!" << endl;
	return true;
      }
    }
  }

  // --------------------------------------------------
  // All diff conditions
  // --------------------------------------------------

  // all_path condition
  if( source_state->all_paths[arc->get_val()] ) {
    return true;
  }

  // some paths condition
  if( source_state->hall_set && source_state->some_paths[arc->get_val()] )	{
    return true;
  }

  return false;
}

/**
 * Check if an in arc is not necessarily infeasible according to disjunctive. It also
 * sets the corresponding arc sketch.
 *
 **/
inline bool DisjunctivePropagator::is_incoming_infeasible_disjunctive(int layer, RevMDD::Arc* arc) {

  // get states
  DisjunctiveState* source_state = get_node_state(layer, arc->get_source());
  DisjunctiveState* target_state = get_node_state(layer+1, arc->get_target());

  // get arc sketch and initialize attributes
  DisjunctiveSketch* sketch = sketches[arc->get_layer_id()];

  // check optional conditions
  if (optional && arc->get_val() == terminal_opt_act 
      && source_state->all_paths[terminal_opt_act]) 
    {
      sketch->earliest_start = source_state->earliest_start.getValue();
      sketch->min_sumsetup_R = source_state->min_sumsetup_R.getValue();
      sketch->min_lagrangian_R = source_state->min_lagrangian_R.getValue();


      mem_min_est[layer][arc->get_layer_id()] = sketch->earliest_start;
      mem_min_sumsetup[layer][arc->get_layer_id()] = sketch->min_sumsetup_R;
      mem_min_lagrangian[layer][arc->get_layer_id()] = sketch->min_lagrangian_R;

      if (has_supply) {
	sketch->max_supply = source_state->max_supply_R.getValue();
	mem_max_supply[layer][arc->get_layer_id()] = sketch->max_supply;
      }

      return false;
    }

  sketch->earliest_start = (layer > 0 ? 
			    INF : acts[arc->get_val()]->start_time.getEndMin());
  sketch->min_sumsetup_R = (layer > 0 ? INF : 0);
  sketch->min_lagrangian_R = (layer > 0 ? INF : 0);
  sketch->max_supply = (layer > 0 ? -1 : 0);
  
  // check activities that are previous to this one
  if (layer > 0) {
    int prev_act;
    int arc_est;
    int max_supply = 0;

    for (int i = 0; i < source_state->in_setup->size; ++i) {
      prev_act = source_state->in_setup->list[i];
      if (prev_act != arc->get_val() && !precedes_activity(arc->get_val(), prev_act)) {

	// earliest start time if previous activity is prev_act
	arc_est = source_state->in_est->value(prev_act) + setup[prev_act][arc->get_val()];
	
	// maximum supply
	if (has_supply) {
	  max_supply = source_state->in_max_supply->value(prev_act) + supply[arc->get_val()];	
	}

	// check feasibility
	if ((arc_est + acts[arc->get_val()]->processing)
	    <= acts[arc->get_val()]->start_time.getEndMax()
	    && max_supply >= min_supply) {

	  // update earliest start time
	  sketch->earliest_start = std::min(sketch->earliest_start, arc_est);

	  // update supply
	  sketch->max_supply = std::max(sketch->max_supply, max_supply);

	  // update sum of setup times
	  sketch->min_sumsetup_R = std::min(sketch->min_sumsetup_R,
					    source_state->in_setup->value(prev_act) 
					    + mult[layer] 
					    * setup[prev_act][arc->get_val()]);
	  
	  // update minimum lagrangian cost
	  sketch->min_lagrangian_R = std::min(sketch->min_lagrangian_R,
					      source_state->in_lagrangian->value(prev_act) 
					      + mult[layer] 
					      * setup[prev_act][arc->get_val()]);
	}
      }
    }
  }

  //
  // Objective function conditions
  //
  if (sketch->min_sumsetup_R == INF 
      || sketch->min_sumsetup_R >= obj_var.getMax() + EPS_COST) {
    return true;
  }  
  mem_min_sumsetup[layer][arc->get_layer_id()] = sketch->min_sumsetup_R;

  //cout << "Set " << layer << " -- " << arc->get_source() << " -- " << arc->get_val() << endl;
  //cout << "\tmin sumsetup = " << mem_min_sumsetup[layer][arc->get_layer_id()] << endl;

  // fix lagrangian cost 
  sketch->min_lagrangian_R += cost_val_layer[arc->get_val()][layer]; 
  mem_min_lagrangian[layer][arc->get_layer_id()] = sketch->min_lagrangian_R;

  // 
  // Earliest start time conditions
  //  
  sketch->earliest_start = std::max(sketch->earliest_start,
				    (int)acts[arc->get_val()]->start_time.getEndMin());
    
  // memoize value
  mem_min_est[layer][arc->get_layer_id()] = sketch->earliest_start;

  //
  // Maximum supply
  //
  mem_max_supply[layer][arc->get_layer_id()] = sketch->max_supply;


   /**
   * Precedence information
   */
  if( has_input_prec ) {

    // --------------------------------
    // Definitely precedes filtering
    // --------------------------------
    if( source_state->all_paths.intersects(act_precedes[arc->get_val()]) ) {
      return true;
    }

    // --------------------------------
    // Definitely succeeds filtering
    // --------------------------------
    temp_set.set();
    temp_set -= source_state->some_paths;
    if( temp_set.intersects(act_succeeds[arc->get_val()]) ) {
      return true;
    }
  }
  
  return false;
}


/**
 * Check if an in arc is not necessarily infeasible
 **/
bool DisjunctivePropagator::is_outgoing_infeasible(int layer, RevMDD::Arc* arc) {
  if( is_outgoing_infeasible_alldiff(layer, arc) ) {
    return true;
  }
  if( is_outgoing_infeasible_disjunctive(layer, arc) ) {
    return true;
  }
  return false;
}

/**
 * Check if an in arc is not necessarily infeasible according to alldiff
 **/
inline bool DisjunctivePropagator::is_outgoing_infeasible_alldiff(int layer, RevMDD::Arc* arc) {

  DisjunctiveState* target_state = get_node_state(layer+1, arc->get_target());

  // --------------------------------------------------
  // Optional conditions
  // --------------------------------------------------
  if (optional) {

    // terminal condition
    if (target_state->all_paths[terminal_opt_act] && arc->get_val() == terminal_opt_act) {
      return false;
    }

    // all_path condition
    if( target_state->all_paths[arc->get_val()] ) {
      return true;
    }

    return false;
  }

  // --------------------------------------------------
  // All diff conditions
  // --------------------------------------------------

  // all_path condition
  if( target_state->all_paths[arc->get_val()] ) {
    return true;
  }

  // some paths condition
  if( target_state->hall_set && target_state->some_paths[arc->get_val()] )	{
    return true;
  }

  return false;
}


/**
 * Check if an in arc is not necessarily infeasible accordding to disjunctive
 **/
inline bool DisjunctivePropagator::is_outgoing_infeasible_disjunctive(int layer, RevMDD::Arc* arc) {

  // get states
  DisjunctiveState* source_state = get_node_state(layer, arc->get_source());
  DisjunctiveState* target_state = get_node_state(layer+1, arc->get_target());

  // get arc sketch and initialize min sum of setup times
  DisjunctiveSketch* sketch = sketches[arc->get_layer_id()];

 // check optional conditions
  if (optional && arc->get_val() == terminal_opt_act 
      && target_state->all_paths[terminal_opt_act]) 
    {
      sketch->latest_finish = target_state->latest_finish.getValue();
      sketch->min_sumsetup_T = target_state->min_sumsetup_T.getValue();
      sketch->min_lagrangian_T = target_state->min_lagrangian_T.getValue();
      sketch->max_supply = target_state->max_supply_T.getValue();

      mem_max_lft[layer][arc->get_layer_id()] = sketch->latest_finish + acts[arc->get_val()]->processing;

      return false;
    }


  sketch->latest_finish = (layer < get_mdd()->get_num_layers()-2 ?
			   -1 : acts[arc->get_val()]->start_time.getEndMax());
  sketch->min_sumsetup_T = (layer < get_mdd()->get_num_layers()-2 ? INF : 0);
  sketch->min_lagrangian_T = (layer < get_mdd()->get_num_layers()-2 ? INF : 0);
  sketch->max_supply = (layer < get_mdd()->get_num_layers()-2 ? -1 : 0);
  //cout << "initial = " << sketch->latest_finish << endl;
			   
  // check activities that are next to this one
  if (layer < get_mdd()->get_num_layers()-2) {
    int next_act;
    int arc_lft;
    int max_supply = 0;

    for (int i = 0; i < target_state->out_setup->size; ++i) {

      next_act = target_state->out_setup->list[i];
      if (next_act != arc->get_val() && !precedes_activity(next_act, arc->get_val())) {

	// maximum time to start activity
	arc_lft = target_state->out_lft->value(next_act) - setup[arc->get_val()][next_act];

	// maximum supply
	if (has_supply) {
	  max_supply = target_state->out_max_supply->value(next_act);
	}

	// check feasibility conditions
	if (arc_lft - acts[arc->get_val()]->processing 
	    >= mem_min_est[layer][arc->get_layer_id()]
	    && mem_max_supply[layer][arc->get_layer_id()] + max_supply >= 0)  {
  
	   // minimum earliest start time
	   sketch->latest_finish = MAX(sketch->latest_finish, arc_lft);

	   // max supply
	   if (has_supply) {
	     sketch->max_supply = MAX(sketch->max_supply, (max_supply + supply[arc->get_val()])) ;
	   }

	   // sum of setup times
	   sketch->min_sumsetup_T = MIN(sketch->min_sumsetup_T,
					target_state->out_setup->value(next_act)
					+ mult[layer+1] 
					* setup[arc->get_val()][next_act]);

	   // lagrangian
	   sketch->min_lagrangian_T = MIN(sketch->min_lagrangian_T,
					  target_state->out_lagrangian->value(next_act)
					  + mult[layer+1] 
					  * setup[arc->get_val()][next_act]);
	}
      }
    }
  }

  // 
  // Sum setup times
  //
  
  // check feasibility conditions
  if (sketch->min_sumsetup_T == INF) {
    return true;
  }
  
  // Consider now activities that are previous to this one. 
  // cout << "\tChecking " << layer << " -- " << arc->get_source() << " -- " << arc->get_val() << endl;
  // cout << "\tmin sumsetup = " << mem_min_sumsetup[layer][arc->get_layer_id()] << endl;
  // cout << "\tsketch = " << sketch->min_sumsetup_T << endl;
  if (((int)mem_min_sumsetup[layer][arc->get_layer_id()] + sketch->min_sumsetup_T)
      > obj_var.getMax()) 
    {
      return true;
    }
  //cout << "done!" << endl;

  //
  // Lagrangian cost
  //
  
  // Consider now activities that are previous to this one. 
  if (obj_var_lag.getImpl() != NULL) {

    int obj = floor(mem_min_lagrangian[layer][arc->get_layer_id()] + 
		    sketch->min_lagrangian_T);
    if ( obj > obj_var_lag.getMax() ) {
      return true;
    }
  }

  // add additional cost of value in layer
  sketch->min_lagrangian_T += cost_val_layer[arc->get_val()][layer];

  //
  // Latest finish time
  //
  sketch->latest_finish = MIN( sketch->latest_finish,
			       acts[arc->get_val()]->start_time.getEndMax())
    - acts[arc->get_val()]->processing;
  
  // check feasibility conditions
  if (mem_min_est[layer][arc->get_layer_id()] > sketch->latest_finish) {
    return true;
  }
    
  // memoize value
  mem_max_lft[layer][arc->get_layer_id()] = sketch->latest_finish + acts[arc->get_val()]->processing;

  /**
   * Precedence information
   */
  if( has_input_prec ) {

    // --------------------------------
    // Definitely succeeds filtering
    // --------------------------------
    if( target_state->all_paths.intersects(act_succeeds[arc->get_val()]) ) {
      return true;
    }

    // --------------------------------
    // Definitely preceeds filtering
    // --------------------------------
    temp_set.set();
    temp_set -= target_state->some_paths;
    if( temp_set.intersects(act_precedes[arc->get_val()]) ) {
      return true;
    }
  }
 
  return false;
}

/**
 * Update activities start/end time
 */
void DisjunctivePropagator::update_start_end_times() {
  
  RevMDD* mdd = get_mdd();

  // collect est and lft for each activity
  act_est->clear();
  act_lft->clear();
  for (int l = 0; l < mdd->get_num_layers(); ++l) {
    for (RevMDD::Arc* arc = mdd->arcs_layer[l].get_first(); 
	 arc != NULL; arc = mdd->arcs_layer[l].get_next(arc)) {
      act_est->add(arc->get_val(), mem_min_est[l][arc->get_layer_id()]);
      act_lft->add(arc->get_val(), mem_max_lft[l][arc->get_layer_id()]);
    }    
  }

 
  // Stats stats;
  // stats.register_name("testA", 0);
  // stats.start_timer("testA");
  // cout << "testA:" << endl;
  // long int totalA = 0;
  // for (int i = 0; i < 1000000000; ++i) {
  //   for (RevMDD::Arc* arc = mdd->arcs_layer[20].get_first(); 
  // 	 arc != NULL; arc = mdd->arcs_layer[20].get_next(arc)) {
  //     totalA++;
  //   }    
  // }
  // stats.end_timer("testA");

  // stats.register_name("testB", 0);
  // stats.start_timer("testB");
  // deque<RevMDD::Arc*> arcs_vector;
  // for (RevMDD::Arc* arc = mdd->arcs_layer[20].get_first(); 
  //      arc != NULL; arc = mdd->arcs_layer[20].get_next(arc)) {
  //   arcs_vector.push_back(arc);
  // }
  // cout << "testB:" << endl;
  // long int totalB = 0;
  // for (int i = 0; i < 1000000000; ++i) {
  //   for (deque<RevMDD::Arc*>::iterator it = arcs_vector.begin();
  // 	 it != arcs_vector.end();
  // 	 ++it) {
  //     totalB++;
  //   }
  // }
  // stats.end_timer("testB");

  // IloCP cp = get_CP();

  // cout << "testC:" << endl;
  // stats.register_name("testC", 0);
  // stats.start_timer("testC");
  // long int testC = 0;
  // int n = arcs_vector.size();
  // IlcRevInt** arrayInt = new IlcRevInt*[n];
  // for (int i = 0; i < n; ++i) {
  //   arrayInt[i] = new IlcRevInt(cp, i);
  // }
  // long int totalC = 0;
  // for (int i = 0; i < 1000000000; ++i) {
  //   for (int j = 0; j < n; ++j) {
  //     if (arcs_vector[arrayInt[j]->getValue()] != NULL) {
  // 	totalC++;
  //     }
  //   }
  // }
  // stats.end_timer("testC");
  


  // cout << "Here!" << " - totalA=" << totalA << " - totalB=" << totalB << endl;
  // cout << "timeA=" << stats.get_time("testA") << " - timeB=" << stats.get_time("testB") << endl;
  // cout << "timeC=" << stats.get_time("testC") << endl;
  // exit(1);


  
  // update start/end times of activities

  if (optional) {
    for (int act = 0; act < (int)acts.size(); ++act) {
      if (!act_est->contains(act) || !act_lft->contains(act)) {
	opt_acts[act].setAbsent();
      } else if (opt_acts[act].isPresent() && act != terminal_opt_act) {
	if (act_est->value(act) > acts[act]->start_time.getStartMin()) {
	  //cout << "\tUpdate start times: " << acts[act]->start_time;
	  acts[act]->start_time.setStartMin(act_est->value(act));
	  //cout << " - to: " << acts[act]->start_time << endl;
	}
	if (act_lft->value(act) < acts[act]->start_time.getEndMax()) {
	  //cout << "\tUpdate end times: " << acts[act]->start_time;
	  acts[act]->start_time.setEndMax(act_lft->value(act));
	  //cout << " - to: " << acts[act]->start_time << endl;
	}	
      }
    }
  } else {
    for (int act = 0; act < (int)acts.size(); ++act) {
      if (!act_est->contains(act) || !act_lft->contains(act)) {
	cout << "ERROR: MDD does not contain activity " << act << endl;
	get_mdd()->export_to_gml("graphs/error.gml");
	exit(1);
      } else {
	if (act_est->value(act) > acts[act]->start_time.getStartMin()) {
	  //cout << "\tUpdate start times: " << acts[act]->start_time;
	  acts[act]->start_time.setStartMin(act_est->value(act));
	  //cout << " - to: " << acts[act]->start_time << endl;
	}
	if (act_lft->value(act) < acts[act]->start_time.getEndMax()) {
	  //cout << "\tUpdate end times: " << acts[act]->start_time;
	  acts[act]->start_time.setEndMax(act_lft->value(act));
	  //cout << " - to: " << acts[act]->start_time << endl;
	}
      }
    }    
  }

  //exit(1);
}


/** 
 * Filter MDD using a particular cost structure and upper bound 
 */
void DisjunctivePropagator::filter(double** cost, const double upper_bound) {
  
  RevMDD* mdd = get_mdd();
  IloCP cp = get_CP();
  
  //mdd->export_to_gml("test.gml");

  // clean nodes
  for( int l = 0; l < mdd->get_num_layers(); ++l ) {
    for( int w = 0; w < mdd->get_max_width(); ++w ) {
      states[l][w]->in_cost->clear();
      states[l][w]->cost_tmp = INF;
    }
  }
  root_state()->cost_tmp = 0;
  root_state()->in_cost->clear();

  // compute shortest path

  DisjunctiveState *source_state, *target_state;
  RevMDD::Arc* prev_arc = NULL;

  // first layer
  for( RevMDD::Arc* arc = mdd->arcs_layer[0].get_first();
       arc != NULL;
       arc = mdd->arcs_layer[0].get_next(arc) )
    {
      states[1][arc->get_target()]->in_cost->add(arc->get_val(), cost[arc->get_val()][0], arc);
      mem_filter[0][arc->get_layer_id()] = cost[arc->get_val()][0];
    }

  // remaining layers
  double cost_arc, cost_arc_tmp;
  for( int l = 1; l < mdd->get_num_layers()-1; ++l ) {
    //cout << "Layer " << l << endl;
    for( RevMDD::Arc* arc = mdd->arcs_layer[l].get_first();
	 arc != NULL;
	 arc = mdd->arcs_layer[l].get_next(arc) )
      {
	source_state = states[l][arc->get_source()];
	target_state = states[l+1][arc->get_target()];

	if (source_state->in_cost->size == 0) {
	  cout << "Error: no incoming nodes during filtering" << endl;
	  exit(1);
	}

	cost_arc = -999999999;
	int prev_value;
	for( int i = 0; i < source_state->in_cost->size; ++i ) {
	  prev_value = source_state->in_cost->list[i];
	  if( prev_value != arc->get_val()
	      &&
	      !precedes_activity(arc->get_val(), prev_value) )
	    {
	      cost_arc_tmp = source_state->in_cost->value(prev_value)
		+ setup[prev_value][arc->get_val()]
		+ cost[arc->get_val()][l];
	      if( cost_arc == -999999999 || cost_arc_tmp < cost_arc ) {
		cost_arc = cost_arc_tmp;
	      }
	    }
	}
	if( cost_arc != -999999999 ) {
	  target_state->in_cost->add(arc->get_val(), cost_arc, arc);
	} else {
	  // TODO: adjust!!
	  target_state->in_cost->add(arc->get_val(), 100000, arc);
	}

	// memoize cost of the arc
	mem_filter[l][arc->get_layer_id()] = cost_arc;
      }
  }


  // perfom now a bottom-up pass to clean-up infeasible arcs
  // (reuses in_cost, but now it means out_cost)
  
  // clean nodes
  for( int l = 0; l < mdd->get_num_layers(); ++l ) {
    for( int w = 0; w < mdd->get_max_width(); ++w ) {
      states[l][w]->in_cost->clear();
      states[l][w]->cost_tmp = INF;
    }
  }
  terminal_state()->cost_tmp = 0;
  terminal_state()->in_cost->clear();

  // previous to the last layer
  for( RevMDD::Arc* arc = mdd->arcs_layer[mdd->get_num_layers()-2].get_first();
       arc != NULL;
       arc = mdd->arcs_layer[mdd->get_num_layers()-2].get_next(arc) )
    {
      states[mdd->get_num_layers()-2][arc->get_source()]->in_cost->add(arc->get_val(), 
								       cost[arc->get_val()][mdd->get_num_layers()-2], 
								       arc);
    }

  // remaining layers

  for( int l = mdd->get_num_layers()-3; l >= 1; --l ) {

    // pointer to the previous arc considered, for removal purposes
    RevMDD::Arc* previous = mdd->arcs_layer[l].get_header();
    for( RevMDD::Arc* arc = mdd->arcs_layer[l].get_first();
	 arc != NULL;
	 arc = mdd->arcs_layer[l].get_next(arc) )
      {
	source_state = states[l][arc->get_source()];
	target_state = states[l+1][arc->get_target()];

	cost_arc = -999999999;
	int next_value;
	for( int i = 0; i < target_state->in_cost->size; ++i ) {
	  next_value = target_state->in_cost->list[i];
	  if( next_value != arc->get_val()
	      &&
	      !precedes_activity(next_value, arc->get_val()) )
	    {
	      cost_arc_tmp = target_state->in_cost->value(next_value)
		+ setup[arc->get_val()][next_value]
		+ cost[arc->get_val()][l];
	      if( cost_arc == -999999999 || cost_arc_tmp < cost_arc ) {
		cost_arc = cost_arc_tmp;
	      }
	    }
	}

	// check arc feasibility conditions
	if (cost_arc == -999999999) {
	  // no next activity: remove arc
	  previous->next.setValue(cp, arc->next.getValue());	
	} else if (mem_filter[l][arc->get_layer_id()] + cost_arc - cost[arc->get_val()][l] > upper_bound) {
	  // cost-based filtering
	  previous->next.setValue(cp, arc->next.getValue());
	  //cout << "\tfiltered - ub=" << upper_bound << " - arc=" << cost_arc+mem_filter[l][arc->get_layer_id()] << endl;
	} else {
	  // arc is feasible
	  previous = arc;
	  source_state->in_cost->add(arc->get_val(), cost_arc, arc);
	}	
      }
  }

  // clean dangling nodes
  mdd->clean_dangling_nodes(false);
}


/** Compute shortest path corresponding to optimal solution in M according to some cost structure.
    Returns arcs in the optimum solution as well. */
int* DisjunctivePropagator::compute_shortest_path(double** cost, double& sp_value, RevMDD::Arc** &sp_arcs) {
  
  // verify shortest path
  RevMDD* mdd = get_mdd();

  // clean nodes
  for( int l = 0; l < mdd->get_num_layers()-1; ++l ) {
    for( int w = 0; w < mdd->get_max_width(); ++w ) {
      states[l+1][w]->in_cost->clear();
      states[l+1][w]->cost_tmp = INF;
    }
  }
  root_state()->cost_tmp = 0;
  root_state()->in_cost->clear();

  // compute shortest path

  DisjunctiveState *source_state, *target_state;
  RevMDD::Arc* prev_arc = NULL;

  // first layer
  for( RevMDD::Arc* arc = mdd->arcs_layer[0].get_first();
       arc != NULL;
       arc = mdd->arcs_layer[0].get_next(arc) )
    {
      states[1][arc->get_target()]->in_cost->add(arc->get_val(), cost[arc->get_val()][0], arc);
      arc->p_hook = NULL;
    }

  // remaining layers
  double cost_arc, cost_arc_tmp;
  for( int l = 1; l < mdd->get_num_layers()-1; ++l ) {
    //cout << "Layer " << l << endl;
    for( RevMDD::Arc* arc = mdd->arcs_layer[l].get_first();
	 arc != NULL;
	 arc = mdd->arcs_layer[l].get_next(arc) )
      {
	source_state = states[l][arc->get_source()];
	target_state = states[l+1][arc->get_target()];

	if (source_state->in_cost->size == 0) {
	  get_CP().fail();
	}

	cost_arc = -999999999;
	int prev_value;
	for( int i = 0; i < source_state->in_cost->size; ++i ) {
	  prev_value = source_state->in_cost->list[i];
	  if( prev_value != arc->get_val()
	      &&
	      !precedes_activity(arc->get_val(), prev_value) )
	    {
	      cost_arc_tmp = source_state->in_cost->value(prev_value)
		//+ cost[prev_value][arc->get_val()];
		+ mult[l] * setup[prev_value][arc->get_val()]
		+ cost[arc->get_val()][l];
	      //cout << "\tcost arc tmp = " << cost_arc_tmp << endl;
	      if( cost_arc == -999999999 || cost_arc_tmp < cost_arc ) {
		cost_arc = cost_arc_tmp;
		prev_arc = (RevMDD::Arc*)source_state->in_cost->pointer(prev_value);
	      }
	    }
	}
	arc->p_hook = (void*)prev_arc;
	if( cost_arc != -999999999 ) {
	  target_state->in_cost->add(arc->get_val(), cost_arc, arc);
	}
      }
  }

  // compute shortest path value and store it in terminal node
  terminal_state()->cost_tmp = -999999999;
  void* last = NULL;
  for( int i = 0; i < terminal_state()->in_cost->size; ++i ) {
    //cout << "Terminal = " << terminal_state()->in_cost->list[i] << endl;
    if( terminal_state()->cost_tmp == -999999999 || 
	terminal_state()->cost_tmp > terminal_state()->in_cost->value(terminal_state()->in_cost->list[i]) ) {
      terminal_state()->cost_tmp = terminal_state()->in_cost->value(terminal_state()->in_cost->list[i]);
      last = terminal_state()->in_cost->pointer(terminal_state()->in_cost->list[i]);
    }
  }

  if( terminal_state()->cost_tmp == -1 ) {
    cout << "ERROR: MDD is not feasible" << endl;
    exit(1);
  }
  sp_value = terminal_state()->cost_tmp;

  // compose optimum solution
  int* optimum = terminal_state()->optimum_sol;
  int l = permutation.getSize()-1;
  //cout << "Arcs: " << endl;
  while( last != NULL ) {
    arcs_shortest_path[l] = ((RevMDD::Arc*)last);
    //cout << "\t" <<  ((RevMDD::Arc*)last)->get_val() << endl;
    optimum[l--] = ((RevMDD::Arc*)last)->get_val();
    last = ((RevMDD::Arc*)last)->p_hook;
  }
  sp_arcs = arcs_shortest_path;
  
  //  cout << "Optimum solution: ";
  //  for( int i = 0; i < (int)acts.size(); ++i ) {
  //    cout << optimum[i] << " ";
  //  }
  //  cout << endl;
 
  return optimum;
}


/**
 * Compute shortest path in M according to some cost structure, which corresponds to the best
 * solution of the MDD
 */
int* DisjunctivePropagator::compute_shortest_path(double** cost, double &sp_value) {

  // verify shortest path
  RevMDD* mdd = get_mdd();

  // clean nodes
  for( int l = 0; l < mdd->get_num_layers()-1; ++l ) {
    for( int w = 0; w < mdd->get_max_width(); ++w ) {
      states[l+1][w]->in_cost->clear();
      states[l+1][w]->cost_tmp = INF;
    }
  }
  root_state()->cost_tmp = 0;
  root_state()->in_cost->clear();

  // compute shortest path

  DisjunctiveState *source_state, *target_state;
  RevMDD::Arc* prev_arc = NULL;

  // first layer
  for( RevMDD::Arc* arc = mdd->arcs_layer[0].get_first();
       arc != NULL;
       arc = mdd->arcs_layer[0].get_next(arc) )
    {
      states[1][arc->get_target()]->in_cost->add(arc->get_val(), cost[arc->get_val()][0], arc);
      arc->p_hook = NULL;
    }

  // remaining layers
  double cost_arc, cost_arc_tmp;
  for( int l = 1; l < mdd->get_num_layers()-1; ++l ) {
    //cout << "Layer " << l << endl;
    for( RevMDD::Arc* arc = mdd->arcs_layer[l].get_first();
	 arc != NULL;
	 arc = mdd->arcs_layer[l].get_next(arc) )
      {
	source_state = states[l][arc->get_source()];
	target_state = states[l+1][arc->get_target()];

	if (source_state->in_cost->size == 0) {
	  get_CP().fail();
	}

	cost_arc = -999999999;
	int prev_value;
	for( int i = 0; i < source_state->in_cost->size; ++i ) {
	  prev_value = source_state->in_cost->list[i];
	  if (!precedes_activity(arc->get_val(), prev_value) )
	    {
	      cost_arc_tmp = source_state->in_cost->value(prev_value)
		//+ cost[prev_value][arc->get_val()];
		+ mult[l] * setup[prev_value][arc->get_val()]
		+ cost[arc->get_val()][l];
	      //cout << "\tcost arc tmp = " << cost_arc_tmp << endl;
	      if( cost_arc == -999999999 || cost_arc_tmp < cost_arc ) {
		cost_arc = cost_arc_tmp;
		prev_arc = (RevMDD::Arc*)source_state->in_cost->pointer(prev_value);
	      }
	    }
	}
	arc->p_hook = (void*)prev_arc;
	if( cost_arc != -999999999 ) {
	  target_state->in_cost->add(arc->get_val(), cost_arc, arc);
	}
      }
  }

  // compute shortest path value and store it in terminal node
  terminal_state()->cost_tmp = -999999999;
  void* last = NULL;
  for( int i = 0; i < terminal_state()->in_cost->size; ++i ) {
    //cout << "Terminal = " << terminal_state()->in_cost->list[i] << endl;
    if( terminal_state()->cost_tmp == -999999999 || 
	terminal_state()->cost_tmp > terminal_state()->in_cost->value(terminal_state()->in_cost->list[i]) ) {
      terminal_state()->cost_tmp = terminal_state()->in_cost->value(terminal_state()->in_cost->list[i]);
      last = terminal_state()->in_cost->pointer(terminal_state()->in_cost->list[i]);
    }
  }

  if( terminal_state()->cost_tmp == -1 ) {
    cout << "ERROR: MDD is not feasible" << endl;
    exit(1);
  }
  sp_value = terminal_state()->cost_tmp;

  // compose optimum solution
  int* optimum = terminal_state()->optimum_sol;
  int l = permutation.getSize()-1;
  while( last != NULL ) {
    optimum[l--] = ((RevMDD::Arc*)last)->get_val();
    last = ((RevMDD::Arc*)last)->p_hook;
  }

  //  cout << "Optimum solution: ";
  //  for( int i = 0; i < (int)acts.size(); ++i ) {
  //    cout << optimum[i] << " ";
  //  }
  //  cout << endl;
 
  return optimum;
}


/**
 * Compute shortest path in M, which corresponds to the best
 * solution of the MDD
 */
int* DisjunctivePropagator::compute_shortest_path() {

  // verify shortest path
  RevMDD* mdd = get_mdd();

  // clean nodes
  for( int l = 0; l < mdd->get_num_layers()-1; ++l ) {
    for( int w = 0; w < mdd->get_max_width(); ++w ) {
      states[l+1][w]->in_est->clear();
      states[l+1][w]->est_tmp = INF;
    }
  }
  root_state()->est_tmp = 0;
  root_state()->in_est->clear();


  // compute shortest path

  DisjunctiveState *source_state, *target_state;
  RevMDD::Arc* prev_arc = NULL;

  // first layer
  for( RevMDD::Arc* arc = mdd->arcs_layer[0].get_first();
       arc != NULL;
       arc = mdd->arcs_layer[0].get_next(arc) )
    {
      states[1][arc->get_target()]->in_est->add(arc->get_val(), 0, arc);
      arc->p_hook = NULL;
    }

  // remaining layers
  int est_arc, est_arc_tmp;
  for( int l = 1; l < mdd->get_num_layers()-1; ++l ) {
    for( RevMDD::Arc* arc = mdd->arcs_layer[l].get_first();
	 arc != NULL;
	 arc = mdd->arcs_layer[l].get_next(arc) )
      {
	source_state = states[l][arc->get_source()];
	target_state = states[l+1][arc->get_target()];

	est_arc = INF;
	int prev_value;
	for( int i = 0; i < source_state->in_est->size; ++i ) {
	  prev_value = source_state->in_est->list[i];
	  if( prev_value != arc->get_val()
	      &&
	      !precedes_activity(arc->get_val(), prev_value) )
	    {
	      est_arc_tmp = source_state->in_est->value(prev_value)
		+ mult[l] * setup[prev_value][arc->get_val()];
	      if( est_arc_tmp < est_arc ) {
		est_arc = est_arc_tmp;
		prev_arc = (RevMDD::Arc*)source_state->in_est->pointer(prev_value);
	      }
	    }
	}
	arc->p_hook = (void*)prev_arc;

	if( est_arc != INF ) {
	  target_state->in_est->add(arc->get_val(), est_arc, arc);
	}
      }
  }

  // compute shortest path value and store it in terminal node
  terminal_state()->est_tmp = INF;
  void* last = NULL;
  for( int i = 0; i < terminal_state()->in_est->size; ++i ) {
    if( terminal_state()->est_tmp > terminal_state()->in_est->value(terminal_state()->in_est->list[i]) ) {
      terminal_state()->est_tmp = terminal_state()->in_est->value(terminal_state()->in_est->list[i]);
      last = terminal_state()->in_est->pointer(terminal_state()->in_est->list[i]);
    }
  }

  if( terminal_state()->est_tmp == INF ) {
    cout << "ERROR: MDD is not feasible" << endl;
    exit(1);
  }

  // compose optimum solution
  int* optimum = terminal_state()->optimum_sol;
  int l = permutation.getSize()-1;
  while( last != NULL ) {
    optimum[l--] = ((RevMDD::Arc*)last)->get_val();
    last = ((RevMDD::Arc*)last)->p_hook;
  }

  //  cout << "Optimum solution: ";
  //  for( int i = 0; i < (int)acts.size(); ++i ) {
  //    cout << optimum[i] << " ";
  //  }
  //  cout << endl;

  return optimum;
}



/**
 * Simple edge for maximum spanning tree computation
 */
struct EdgeSPT {
  int v;
  int u;
  int weight;

  //  bool operator<(Edge &a) {
  //    return a.weight > weight;
  //  }
};

struct EdgeSPTComp {
  bool operator()(EdgeSPT a, EdgeSPT b) {
    return a.weight > b.weight;
  }
};


///**
// * Compute activities that have priority to be exact in any solution
// */
//void DisjunctivePropagator::compute_exact_acts() {
//
//  // collect activities that are not fixed
//  int num_acts = acts.size();
//
//  // create activity edges for maximum spanning tree
//  vector<EdgeSPT> edges;
//  for( int i = 0; i < num_acts-1; ++i ) {
//    for( int j = i+1; j < num_acts; ++j ) {
//      EdgeSPT edge;
//      edge.u = i;
//      edge.v = j;
//      edge.weight = MIN(setup[i][j]+acts[j]->processing, setup[j][i]+acts[i]->processing);
//      edges.push_back(edge);
//    }
//  }
//  sort(edges.begin(), edges.end(), EdgeSPTComp());
//
//  memset(act_checked, false, sizeof(bool)*num_acts);
//
//  for( int i = 0; i < (int)edges.size(); ++i ) {
//    if( !act_checked[edges[i].u] ) {
//      exact_acts.push_back(edges[i].u);
//      act_checked[edges[i].u] = true;
//    }
//    if( !act_checked[edges[i].v] ) {
//      exact_acts.push_back(edges[i].v);
//      act_checked[edges[i].v] = true;
//    }
//    if( (int)exact_acts.size() == num_acts ) {
//      break;
//    }
//  }
//
//  cout << "Exact activity priority: ";
//  for( int i = 0; i < (int)exact_acts.size(); ++i ) {
//    cout << exact_acts[i] << " ";
//  }
//  cout << endl;
//
//  //exit(1);
//}



/**
 * Compute activities that have priority to be exact in any solution
 */
void DisjunctivePropagator::compute_exact_acts() {

  //return;

  if( input_refinement_file ) {

    cout << "\n[Disjunctive] Exact activities priority (read from file): ";
    for( int i = 0; i < (int)exact_acts.size(); ++i ) {
      cout << exact_acts[i] << " ";
    }
    cout << endl << endl;


    return;
  }

  int num_acts = acts.size();

  if( random_exact_refinement ) {

    exact_acts.clear();
    for( int i = 0; i < num_acts; ++i ) {
      exact_acts.push_back(-1);
    }

    boost::random::uniform_int_distribution<> dist(0, num_acts-1);
    for( int act = 0; act < num_acts; ++act ) {
      int pos = dist( random_generator );
      while( exact_acts[pos] != -1 ) {
	pos = dist( random_generator );
      }
      exact_acts[pos] = act;
    }

    ofstream order_file("ref-order.txt");
    for( int i = 0; i < (int)exact_acts.size(); ++i ) {
      order_file << exact_acts[i] << " ";
    }
    order_file << "\n";
    order_file.close();


    cout << "\n[Disjunctive] Exact activities priority (random): ";
    for( int i = 0; i < (int)exact_acts.size(); ++i ) {
      cout << exact_acts[i] << " ";
    }
    cout << endl << endl;

    return;
  }


  // switch( obj_type ) {

  // case Makespan:
  //   {
  //     // collect activities that are not fixed
  //     memset(act_checked, false, sizeof(bool)*num_acts);
  //     for( int i = 0; i < permutation.getSize(); ++i ) {
  // 	if( permutation[i].isFixed() ) {
  // 	  act_checked[permutation[i].getValue()] = true;
  // 	}
  //     }

  //     exact_acts.clear();
  //     for( int i = 0; i < num_acts; ++i ) {
  // 	if( !act_checked[i] ) {
  // 	  exact_acts.push_back(i);
  // 	}
  //     }
  //     if( exact_acts.size() <= 2 )
  // 	return;

  //     // first activity has the maximum deadline, using the largest processing time
  //     // as tie-breaking
  //     int max_release = -1;
  //     int max_proc = -1;
  //     int first = INF;
  //     for( vector<int>::iterator it = exact_acts.begin(); it != exact_acts.end(); ++it ) {
  // 	int i = *it;
  // 	if( (max_release < acts[i]->start_time.getStartMin() + acts[i]->processing )
  // 	    || (max_release == acts[i]->start_time.getStartMin() + acts[i]->processing && max_proc < acts[i]->processing) )
  // 	  {
  // 	    first = it - exact_acts.begin();
  // 	    max_release = acts[i]->start_time.getStartMin() + acts[i]->processing;
  // 	    max_proc = acts[i]->processing;
  // 	  }
  //     }

  //     // put activity in the first position
  //     int tmp = exact_acts[0];
  //     exact_acts[0] = exact_acts[first];
  //     exact_acts[first] = tmp;

  //     // sort remaining activities by largest (processing time+min_setup_from)
  //     aux_v.resize(num_acts);
  //     for( vector<int>::iterator it = exact_acts.begin(); it != exact_acts.end(); ++it ) {
  // 	int i = *it;
  // 	aux_v[i] = (acts[i]->start_time.getStartMin() + acts[i]->processing + max_setup_from[i]);
  // 	//aux_v[i] = (acts[i]->start_time.getStartMin() + acts[i]->processing + min_setup_from[i]);
  // 	//aux_v[i] = max_setup_from[i] + acts[i]->processing;
  // 	if( acts[i]->weight > 0 ) {
  // 	  aux_v[i] *= acts[i]->weight;
  // 	}
  //     }
  //     sort(exact_acts.begin()+1, exact_acts.end(), IndexComparatorDescendingSTL(aux_v));
  //     //sort(exact_acts.begin(), exact_acts.end(), IndexComparatorDescendingSTL(aux_v));
  //     //sort(exact_acts.rbegin(), exact_acts.rend());

  //     cout << "\n[Disjunctive] Exact activities priority: ";
  //     for( int i = 0; i < (int)exact_acts.size(); ++i ) {
  // 	cout << exact_acts[i] << " ";
  //     }
  //     cout << endl << endl;
  //   }
  //   break;

  // case SumSetupTimes:
  //   {
      exact_acts.clear();
      vector<bool> is_ordered( num_acts, false );

      // first activity has the maximum release date, using the largest processing time
      // as tie-breaking
      int max_release = -1;
      int max_proc = -1;
      int first = INF;
      for ( int i = 0; i < num_acts; ++i ) {
	if( (max_release < acts[i]->start_time.getStartMin() + acts[i]->processing )
	    || (max_release == acts[i]->start_time.getStartMin() + acts[i]->processing && max_proc < acts[i]->processing) )
	  {
	    first = i;
	    max_release = acts[i]->start_time.getStartMin() + acts[i]->processing;
	    max_proc = acts[i]->processing;
	  }
      }

      // put activity in the first position
      exact_acts.push_back( first );
      is_ordered[ first ] = true;

      // add remaining activities according to sum of setup times
      while( (int)exact_acts.size() < num_acts ) {
	int max_sum = -1;
	int act = -1;
	for( int i = 0; i < num_acts; ++i ) {
	  if( is_ordered[i] )
	    continue;
	  int sum = 0;
	  for( vector<int>::iterator it = exact_acts.begin(); it != exact_acts.end(); ++it ) {
	    if( input_prec[*it][i] ) {
	      sum += setup[*it][i];
	    } else {
	      sum += setup[i][*it];
	    }
	  }
	  if( sum > max_sum ) {
	    max_sum = sum;
	    act = i;
	  }
	}
	//cout << "act: " << act << " - max_sum: " << max_sum << endl;
	exact_acts.push_back( act );
	is_ordered[act] = true;
      }

      // if optional, add terminal activity in the first priority list
      if (optional) {
      	vector<int> final;
      	final.push_back(terminal_opt_act);
      	final.insert(final.end(), exact_acts.begin(), exact_acts.end());
      	exact_acts = final;
      }

      cout << "\n[Disjunctive] Exact activities priority: ";
       for( int i = 0; i < (int)exact_acts.size(); ++i ) {
       	cout << exact_acts[i] << " ";
       }
       cout << endl << endl;
      //    }
  //   break;

  // case TotalTardiness:
  // case MaxTardiness:
  //   {
  //     // collect activities that are not fixed
  //     memset(act_checked, false, sizeof(bool)*num_acts);
  //     for( int i = 0; i < permutation.getSize(); ++i ) {
  // 	if( permutation[i].isFixed() ) {
  // 	  act_checked[permutation[i].getValue()] = true;
  // 	}
  //     }

  //     exact_acts.clear();
  //     for( int i = 0; i < num_acts; ++i ) {
  // 	if( !act_checked[i] ) {
  // 	  exact_acts.push_back(i);
  // 	}
  //     }
  //     if( exact_acts.size() <= 2 )
  // 	return;

  //     // first activity has the maximum deadline, using the largest processing time
  //     // as tie-breaking
  //     int max_release = -1;
  //     int max_proc = -1;
  //     int first = INF;
  //     for( vector<int>::iterator it = exact_acts.begin(); it != exact_acts.end(); ++it ) {
  // 	int i = *it;
  // 	if( (max_release < acts[i]->start_time.getStartMin() + acts[i]->processing )
  // 	    || (max_release == acts[i]->start_time.getStartMin() + acts[i]->processing && max_proc < acts[i]->processing) )
  // 	  {
  // 	    first = it - exact_acts.begin();
  // 	    max_release = acts[i]->start_time.getStartMin() + acts[i]->processing;
  // 	    max_proc = acts[i]->processing;
  // 	  }
  //     }

  //     // put activity in the first position
  //     int tmp = exact_acts[0];
  //     exact_acts[0] = exact_acts[first];
  //     exact_acts[first] = tmp;

  //     // sort remaining activities by largest (processing time+min_setup_from)
  //     aux_v.resize(num_acts);
  //     for( vector<int>::iterator it = exact_acts.begin(); it != exact_acts.end(); ++it ) {
  // 	int i = *it;
  // 	//aux_v[i] = (acts[i]->start_time.getStartMin() + acts[i]->processing + max_setup_from[i]);
  // 	//aux_v[i] = (acts[i]->start_time.getStartMin() + acts[i]->processing + min_setup_from[i]);
  // 	//aux_v[i] = (acts[i]->duedate - acts[i]->release - acts[i]->processing);
  // 	aux_v[i] = -acts[i]->duedate;
  // 	//			if( acts[i]->weight > 0 ) {
  // 	//				aux_v[i] *= acts[i]->weight;
  // 	//			}
  //     }
  //     //sort(exact_acts.begin()+1, exact_acts.end(), IndexComparatorDescendingSTL(aux_v));
  //     sort(exact_acts.begin(), exact_acts.end(), IndexComparatorDescendingSTL(aux_v));
  //     //sort(exact_acts.rbegin(), exact_acts.rend());

  //     cout << "\n[Disjunctive] Exact activities priority: ";
  //     for( int i = 0; i < (int)exact_acts.size(); ++i ) {
  // 	cout << exact_acts[i] << " ";
  //     }
  //     cout << endl << endl;

  //     //		exact_acts.clear();
  //     //			vector<bool> is_ordered( num_acts, false );
  //     //
  //     //			// first activity has the maximum release date, using the largest processing time
  //     //			// as tie-breaking
  //     //			int max_release = -1;
  //     //			int max_proc = -1;
  //     //			int first = INF;
  //     //			for ( int i = 0; i < num_acts; ++i ) {
  //     //				if( (max_release < acts[i]->start_time.getStartMin() + acts[i]->processing )
  //     //						|| (max_release == acts[i]->start_time.getStartMin() + acts[i]->processing && max_proc < acts[i]->processing) )
  //     //				{
  //     //					first = i;
  //     //					max_release = acts[i]->start_time.getStartMin() + acts[i]->processing;
  //     //					max_proc = acts[i]->processing;
  //     //				}
  //     //			}
  //     //
  //     //			// put activity in the first position
  //     //			exact_acts.push_back( first );
  //     //			is_ordered[ first ] = true;
  //     //
  //     //			// add remaining activities according to sum of setup times
  //     //			while( (int)exact_acts.size() < num_acts ) {
  //     //				int max_sum = -1;
  //     //				int act = -1;
  //     //				for( int i = 0; i < num_acts; ++i ) {
  //     //					if( is_ordered[i] )
  //     //						continue;
  //     //					int sum = 0;
  //     //					for( vector<int>::iterator it = exact_acts.begin(); it != exact_acts.end(); ++it ) {
  //     //						if( input_prec[*it][i] ) {
  //     //							sum += setup[*it][i];
  //     //						} else {
  //     //							sum += setup[i][*it];
  //     //						}
  //     //					}
  //     //					if( sum > max_sum ) {
  //     //						max_sum = sum;
  //     //						act = i;
  //     //					}
  //     //				}
  //     //				cout << "act: " << act << " - max_sum: " << max_sum << endl;
  //     //				exact_acts.push_back( act );
  //     //				is_ordered[act] = true;
  //     //			}
  //     //
  //     //			cout << "\n[Disjunctive] Exact activities priority: ";
  //     //			for( int i = 0; i < (int)exact_acts.size(); ++i ) {
  //     //				cout << exact_acts[i] << " ";
  //     //			}
  //     //			cout << endl << endl;
  //   }
  //   break;

  // case Latency:
  //   {
  //     // collect activities that are not fixed
  //     memset(act_checked, false, sizeof(bool)*num_acts);
  //     for( int i = 0; i < permutation.getSize(); ++i ) {
  // 	if( permutation[i].isFixed() ) {
  // 	  act_checked[permutation[i].getValue()] = true;
  // 	}
  //     }

  //     exact_acts.clear();
  //     for( int i = 0; i < num_acts; ++i ) {
  // 	if( !act_checked[i] ) {
  // 	  exact_acts.push_back(i);
  // 	}
  //     }
  //     if( exact_acts.size() <= 2 )
  // 	return;

  //     // first activity has the maximum deadline, using the largest processing time
  //     // as tie-breaking
  //     int max_release = -1;
  //     int max_proc = -1;
  //     int first = INF;
  //     for( vector<int>::iterator it = exact_acts.begin(); it != exact_acts.end(); ++it ) {
  // 	int i = *it;
  // 	if( (max_release < (acts[i]->start_time.getStartMin() + acts[i]->processing)*acts[i]->weight )
  // 	    || (max_release == (acts[i]->start_time.getStartMin() + acts[i]->processing)*acts[i]->weight
  // 		&& max_proc < acts[i]->processing) )
  // 	  {
  // 	    first = it - exact_acts.begin();
  // 	    max_release = acts[i]->weight*(acts[i]->start_time.getStartMin() + acts[i]->processing);
  // 	    max_proc = acts[i]->processing;
  // 	  }
  //     }

  //     // put activity in the first position
  //     int tmp = exact_acts[0];
  //     exact_acts[0] = exact_acts[first];
  //     exact_acts[first] = tmp;

  //     // sort remaining activities by largest (processing time+min_setup_from)
  //     aux_v.resize(num_acts);
  //     vector<int> v2(num_acts);
  //     for( vector<int>::iterator it = exact_acts.begin(); it != exact_acts.end(); ++it ) {
  // 	int i = *it;
  // 	aux_v[i] = (acts[i]->start_time.getStartMin() + acts[i]->processing + max_setup_from[i]);
  // 	//aux_v[i] = (acts[i]->start_time.getStartMin() + acts[i]->processing + min_setup_from[i]);
  // 	//aux_v[i] = max_setup_from[i] + acts[i]->processing;
  // 	//			aux_v[i] = -acts[i]->duedate;
  // 	if( acts[i]->weight > 0 ) {
  // 	  aux_v[i] *= acts[i]->weight;
  // 	}
  // 	//aux_v[i] = -acts[i]->weight;
  // 	//v2[i] = -(acts[i]->start_time.getStartMin() + acts[i]->processing + max_setup_from[i]);
  //     }
  //     sort(exact_acts.begin()+1, exact_acts.end(), IndexComparatorDescendingSTL(aux_v));
  //     //sort(exact_acts.begin()+1, exact_acts.end(), IndexComparatorDescendingSTL2D(aux_v, v2));
  //     //sort(exact_acts.begin(), exact_acts.end(), IndexComparatorDescendingSTL(aux_v));
  //     //sort(exact_acts.rbegin(), exact_acts.rend());

  //     cout << "\n[Disjunctive] Exact activities priority: ";
  //     for( int i = 0; i < (int)exact_acts.size(); ++i ) {
  // 	cout << exact_acts[i] << " ";
  //     }
  //     cout << endl << endl;
  //   }
  //   break;

  // default:
  //   cerr << "\nRefinement not specified!" << endl;
  //   exit(1);
  //   break;
  // }



  //	exact_acts.clear();
  //
  //	int n_requests = (num_acts-2)/2;
  //
  //	for( int i = 0; i < n_requests; ++i ) {
  //		exact_acts.push_back( 1 + i );
  //	}
  //
  //	for( int i = 0; i < n_requests; ++i ) {
  //		exact_acts.push_back( n_requests + 1 + i );
  //	}
  //
  //
  //	exact_acts.push_back( 0 );
  //	exact_acts.push_back( num_acts - 1 );
  //
  //	cout << "\n[Disjunctive] Exact activities priority: ";
  //	for( int i = 0; i < (int)exact_acts.size(); ++i ) {
  //		cout << exact_acts[i] << " ";
  //	}
  //	cout << endl << endl;
  //
  //	return;
  //
  //
  //
  //	vector<bool> is_ordered(num_acts, false);
  //	vector<int> act_order;
  //
  //	// select pair of activities with largest distance
  //	int a1 = 0, a2 = 0;
  //	for( int i = 0; i < num_acts; ++i ) {
  //		for( int j = 0; j < num_acts-1; ++j ) {  // exclude last task
  //			if( setup[i][j] > setup[a1][a2] ) {
  //				a1 = i;
  //				a2 = j;
  //			}
  //		}
  //	}
  //
  //	cout << "largest pair: " << a1 << "," << a2 << " ==> " << setup[a1][a2] << endl;
  //	is_ordered[a1] = true;
  //	is_ordered[a2] = true;
  //	act_order.push_back(a1);
  //	act_order.push_back(a2);
  //
  ////	while( (int)act_order.size() < num_acts ) {
  ////		int min_max = -1;
  ////		int act = -1;
  ////		for( int i = 0; i < num_acts; ++i ) {
  ////			if( is_ordered[i] )
  ////				continue;
  ////			int min = INF;
  ////			for( vector<int>::iterator it = act_order.begin(); it != act_order.end(); ++it ) {
  ////				if( input_prec[*it][i] ) {
  ////					min = MIN(min, setup[*it][i]);
  ////				} else if( input_prec[i][*it] ) {
  ////					min = MIN(min, setup[i][*it]);
  ////				} else {
  ////					min = MIN(min, setup[i][*it]);
  ////					min = MIN(min, setup[*it][i]);
  ////				}
  ////			}
  ////			if( min > min_max ) {
  ////				min_max = min;
  ////				act = i;
  ////			}
  ////		}
  ////		cout << "act: " << act << " - min_max: " << min_max << endl;
  ////		act_order.push_back(act);
  ////		is_ordered[act] = true;
  ////	}
  //
  //	while( (int)act_order.size() < num_acts ) {
  //		int min_max = -1;
  //		int act = -1;
  //		for( int i = 0; i < num_acts; ++i ) {
  //			if( is_ordered[i] )
  //				continue;
  //			int sum = 0;
  //			for( vector<int>::iterator it = act_order.begin(); it != act_order.end(); ++it ) {
  //				if( input_prec[*it][i] ) {
  //					sum += setup[*it][i];
  //				} else {
  //					sum += setup[i][*it];
  //				}
  //			}
  //			if( sum > min_max ) {
  //				min_max = sum;
  //				act = i;
  //			}
  //		}
  //		cout << "act: " << act << " - min_max: " << min_max << endl;
  //		act_order.push_back(act);
  //		is_ordered[act] = true;
  //	}
  //
  //
  //	exact_acts = act_order;
  //	cout << "\n[Disjunctive] Exact activities priority: ";
  //	for( int i = 0; i < (int)exact_acts.size(); ++i ) {
  //		cout << exact_acts[i] << " ";
  //	}
  //	cout << endl << endl;
  //
  //	return;
  //
  //////	exit(1);

  // ---------------------------------------------------------
  //
  //	// collect activities that are not fixed
  //	memset(act_checked, false, sizeof(bool)*num_acts);
  //	for( int i = 0; i < permutation.getSize(); ++i ) {
  //		if( permutation[i].isFixed() ) {
  //			act_checked[permutation[i].getValue()] = true;
  //		}
  //	}
  //
  //	exact_acts.clear();
  //	for( int i = 0; i < num_acts; ++i ) {
  //		if( !act_checked[i] ) {
  //			exact_acts.push_back(i);
  //		}
  //	}
  //	if( exact_acts.size() <= 2 )
  //		return;
  //
  //	// first activity has the maximum deadline, using the largest processing time
  //	// as tie-breaking
  //	int max_release = -1;
  //	int max_proc = -1;
  //	int first = INF;
  //	for( vector<int>::iterator it = exact_acts.begin(); it != exact_acts.end(); ++it ) {
  //		int i = *it;
  //		if( (max_release < acts[i]->start_time.getStartMin() + acts[i]->processing )
  //				|| (max_release == acts[i]->start_time.getStartMin() + acts[i]->processing && max_proc < acts[i]->processing) )
  //		{
  //			first = it - exact_acts.begin();
  //			max_release = acts[i]->start_time.getStartMin() + acts[i]->processing;
  //			max_proc = acts[i]->processing;
  //		}
  //	}
  //
  //	// put activity in the first position
  //	int tmp = exact_acts[0];
  //	exact_acts[0] = exact_acts[first];
  //	exact_acts[first] = tmp;
  //
  //	// sort remaining activities by largest (processing time+min_setup_from)
  //	aux_v.resize(num_acts);
  //	for( vector<int>::iterator it = exact_acts.begin(); it != exact_acts.end(); ++it ) {
  //		int i = *it;
  //		aux_v[i] = (acts[i]->start_time.getStartMin() + acts[i]->processing + max_setup_from[i]);
  //		//aux_v[i] = (acts[i]->start_time.getStartMin() + acts[i]->processing + min_setup_from[i]);
  //		//aux_v[i] = max_setup_from[i] + acts[i]->processing;
  //		if( acts[i]->weight > 0 ) {
  //			aux_v[i] *= acts[i]->weight;
  //		}
  //	}
  //	sort(exact_acts.begin()+1, exact_acts.end(), IndexComparatorDescendingSTL(aux_v));
  //	//sort(exact_acts.begin(), exact_acts.end(), IndexComparatorDescendingSTL(aux_v));
  //	//sort(exact_acts.rbegin(), exact_acts.rend());
  //
  //	// H !!!!!!!!!!!!!!!!
  //	//sort(exact_acts.begin(), exact_acts.end(), IndexComparatorDescendingSTL(aux_v));
  //
  //	cout << "\n[Disjunctive] Exact activities priority: ";
  //	for( int i = 0; i < (int)exact_acts.size(); ++i ) {
  //		cout << exact_acts[i] << " ";
  //	}
  //	cout << endl << endl;


  // ---------------------------------------------------------

  // for sum of setup times

}


/**
 * Get activity and position of the activity with best sum of setup times to T
 **/
void DisjunctivePropagator::get_activity_min_sum_setup_T(int& pos, int& act) {

  if( obj_type == TotalTardiness ) {
    RevMDD* mdd = get_mdd();
    //int prev_act = -1;
    int min_tard_T;
    DisjunctiveState *source_state, *target_state;

    //		pos = 0;
    //		while( mdd->arcs_layer[pos].is_unique() ) {
    //			prev_act = mdd->arcs_layer[pos++].get_first()->get_val();
    //		}

    //cout << "Layer " << pos << endl;
    min_tard_T = INF;

    if( pos > 0 ) {
      for( RevMDD::Arc* arc = mdd->arcs_layer[pos].get_first();
	   arc != NULL;
	   arc = mdd->arcs_layer[pos].get_next(arc) )
	{
	  source_state = get_node_state(pos, arc->get_source());
	  target_state = get_node_state(pos+1, arc->get_target());

	  if( min_tard_T > source_state->min_tard.getValue() + target_state->min_tard_T.getValue() ) {
	    act = arc->get_val();
	    min_tard_T = source_state->min_tard.getValue() + target_state->min_tard_T.getValue();
	    //cout << "\ttarget: " << arc->get_target() << " - val: " << arc->get_val() << " - min tard T: " << min_tard_T << endl;
	  }
	}
    } else {
      for( RevMDD::Arc* arc = mdd->arcs_layer[pos].get_first();
	   arc != NULL;
	   arc = mdd->arcs_layer[pos].get_next(arc) )
	{
	  source_state = get_node_state(pos, arc->get_source());
	  target_state = get_node_state(pos+1, arc->get_target());
	  if( source_state->min_tard.getValue() + target_state->min_tard_T.getValue() < min_tard_T ) {
	    act = arc->get_val();
	    min_tard_T = source_state->min_tard.getValue() + target_state->min_tard_T.getValue();
	    //cout << "\ttarget2: " << arc->get_target() << " - min tard T: " << target_state->min_tard_T.getValue() << endl;
	  }
	}
    }

    //exit(1);
    if( pos == mdd->get_num_layers()-1 ) {
      pos = -1;
    }
    return;
  }

  RevMDD* mdd = get_mdd();
  int prev_act = -1;
  int min_sum_setup_T;
  DisjunctiveState *source_state, *target_state;

  pos = 0;
  while( mdd->arcs_layer[pos].is_unique() ) {
    prev_act = mdd->arcs_layer[pos++].get_first()->get_val();
  }

  //cout << "Layer " << pos << endl;
  min_sum_setup_T = INF;

  if( pos > 0 ) {
    for( RevMDD::Arc* arc = mdd->arcs_layer[pos].get_first();
	 arc != NULL;
	 arc = mdd->arcs_layer[pos].get_next(arc) )
      {
	source_state = get_node_state(pos, arc->get_source());
	target_state = get_node_state(pos+1, arc->get_target());
	int relaxation_val = source_state->min_sumsetup_R.getValue()
	  + setup[prev_act][arc->get_val()]
	  + target_state->min_sumsetup_T.getValue();

	if( relaxation_val < min_sum_setup_T ) {
	  act = arc->get_val();
	  min_sum_setup_T = relaxation_val;
	  //cout << "\ttarget: " << arc->get_target() << " - val: " << arc->get_val() << " - min sum setup: " << min_sum_setup_T << endl;
	}
      }
  } else {
    for( RevMDD::Arc* arc = mdd->arcs_layer[pos].get_first();
	 arc != NULL;
	 arc = mdd->arcs_layer[pos].get_next(arc) )
      {
	target_state = get_node_state(pos+1, arc->get_target());
	if( target_state->min_sumsetup_T.getValue() < min_sum_setup_T ) {
	  act = arc->get_val();
	  min_sum_setup_T = target_state->min_sumsetup_T.getValue();
	  //cout << "\ttarget2: " << arc->get_target() << " - min sum setup: " << state->min_sumsetup_T.getValue() << endl;
	}
      }
  }

  //exit(1);
  if( pos == mdd->get_num_layers()-1 ) {
    pos = -1;
  }
  return;
}


/**
 * Get node with best sum of setup times to T of a particular layer
 **/
int DisjunctivePropagator::get_node_min_sum_setup_T(int layer) {
  RevMDD* mdd = get_mdd();

  if (layer == 0 || layer == mdd->get_num_layers() -1) {
    return 0;
  }

  // collect nodes in layer
  set<int> nodes;
  for( RevMDD::Arc* arc = mdd->arcs_layer[layer].get_first(); arc != NULL;
       arc = mdd->arcs_layer[layer].get_next(arc) )
    {
      nodes.insert( arc->get_source() );
    }

  // check node with minimum sum of setup times
  int min_sumsetup = INF;
  int node = -1;
  for (set<int>::iterator it = nodes.begin(); it != nodes.end(); ++it) {
    int est = get_node_state(layer, *it)->min_sumsetup_R.getValue()
      + get_node_state(layer, *it)->min_sumsetup_T.getValue();
    if (est < min_sumsetup) {
      node = *it;
      min_sumsetup = est;
    }
  }
  assert(node != -1);

  return node;
}


/**
 * Get node with best sum of setup times to T
 **/
void DisjunctivePropagator::get_node_min_sum_setup_T(int& layer, int& node) {

  RevMDD* mdd = get_mdd();
  int prev_act = -1;
  int min_sum_setup_T;
  //int act;
  DisjunctiveState *source_state, *target_state;

  int pos = 0;
  while( mdd->arcs_layer[pos].is_unique() ) {
    prev_act = mdd->arcs_layer[pos++].get_first()->get_val();
  }

  //cout << "Layer " << pos << endl;
  min_sum_setup_T = INF;
  layer = -1;
  node = -1;

  if( layer > 0 ) {
    for( RevMDD::Arc* arc = mdd->arcs_layer[pos].get_first();
	 arc != NULL;
	 arc = mdd->arcs_layer[pos].get_next(arc) )
      {
	source_state = get_node_state(pos, arc->get_source());
	target_state = get_node_state(pos+1, arc->get_target());
	int relaxation_val = source_state->min_sumsetup_R.getValue()
	  + setup[prev_act][arc->get_val()]
	  + target_state->min_sumsetup_T.getValue();

	if( relaxation_val < min_sum_setup_T ) {
	  //act = arc->get_val();
	  min_sum_setup_T = relaxation_val;
	  //cout << "\ttarget: " << arc->get_target() << " - val: " << arc->get_val() << " - min sum setup: " << min_sum_setup_T << endl;
	  node = arc->get_target();
	  layer = pos+1;
	}
      }
  } else {
    for( RevMDD::Arc* arc = mdd->arcs_layer[pos].get_first();
	 arc != NULL;
	 arc = mdd->arcs_layer[pos].get_next(arc) )
      {
	target_state = get_node_state(pos+1, arc->get_target());
	if( target_state->min_sumsetup_T.getValue() < min_sum_setup_T ) {
	  //act = arc->get_val();
	  min_sum_setup_T = target_state->min_sumsetup_T.getValue();
	  node = arc->get_target();
	  layer = pos+1;
	  //cout << "\ttarget2: " << arc->get_target() << " - min sum setup: " << target_state->min_sumsetup_T.getValue() << endl;
	}
      }
  }

  //cout << "Selected node " << node << " in layer " << layer << endl;
  if( pos == mdd->get_num_layers()-1 ) {
    pos = -1;
    //		layer = -1;
    //		node = -1;
  }
  return;
}



///**
// * Compute activities that have priority to be exact in any solution
// * TODO: should be recomputed at each search node?
// */
//void DisjunctivePropagator::compute_exact_acts() {
//  int num_acts = acts.size();
//
//  // first activity has the maximum deadline, using the largest processing time
//  // as tie-breaking
//
//  int max_release = -1;
//  int max_proc = -1;
//  for( int i = 0; i < num_acts; ++i) {
//    if( (max_release < acts[i]->start_time.getStartMin())
//        || (max_release == acts[i]->start_time.getStartMin() && max_proc < acts[i]->processing) )
//    {
//      exact_acts_array[0] = i;
//      max_release = acts[i]->start_time.getStartMin();
//      max_proc = acts[i]->processing;
//    }
//  }
//
//  // compose remaining indices
//  for( int i = 1; i < num_acts; ++i ) {
//    exact_acts_array[i] = (i-1 < exact_acts_array[0] ? i-1 : i);
//  }
//
//  // sort remaining activities by largest (processing time+min_setup_from)
//  aux_v.resize(num_acts);
//  for( int i = 0; i < num_acts; ++i ) {
//    aux_v[i] = (acts[i]->release + acts[i]->processing + min_setup_from[i]);
//    //aux_v[i] = (acts[i]->release + acts[i]->processing + max_setup_from[i]);
//    //aux_v[i] = -acts[i]->processing - min_setup_from[i];
//    //aux_v[i] = (acts[i]->weight*acts[i]->processing);
//    //aux_v[i] = acts[i]->weight;
//    //aux_v[i] = -acts[i]->weight;
//  }
//  sort(exact_acts_array+1, exact_acts_array+num_acts, IndexComparatorDescendingSTL(aux_v));
//
//  cout << "\n[Disjunctive] Exact activities priority: ";
//  for( int i = 0; i < num_acts; ++i ) {
//    cout << exact_acts_array[i] << " ";
//  }
//  cout << endl << endl;
//}


/**
 * Apply Held-Karp relaxation
 */
void DisjunctivePropagator::apply_help_karp_relaxation() {

  //return;

  RevMDD* mdd = get_mdd();
  HeldKarpsRelaxation hk_relax( acts.size() );

  // initialize root state
  get_node_state(0, 0)->in_est->clear();

  // build graph
  for ( int l = 0; l < mdd->get_num_layers()-1; ++l ) {

    //cout << "Layer " << l << endl;

    // clean incoming list of the next layer
    for ( RevMDD::Arc* arc = mdd->arcs_layer[l+1].get_first();
	  arc != NULL;
	  arc = mdd->arcs_layer[l+1].get_next(arc) )
      {
	get_node_state( l+1, arc->get_source() )->in_list->clear();
      }

    // create arcs
    for( RevMDD::Arc* arc = mdd->arcs_layer[l].get_first();
	 arc != NULL;
	 arc = mdd->arcs_layer[l].get_next(arc) )
      {
	//cout << "\tArc: " << arc->get_val() << endl;

	// add arcs to graph
	DisjunctiveState* state = get_node_state( l, arc->get_source() );
	for ( int i = 0; i < state->in_list->size; ++i ) {
	  int prev_value = state->in_list->elements[i];
	  //cout << "\tChecking " << prev_value << " <--> " << arc->get_val() << endl;
	  if ( prev_value != arc->get_val()
	       &&
	       !precedes_activity(arc->get_val(), prev_value) )
	    {
	      hk_relax.add_edge_min_weight(prev_value, arc->get_val(),
					   setup[prev_value][arc->get_val()]);
	    }
	}
	// add to target list
	get_node_state( l+1, arc->get_target() )->in_list->add( arc->get_val() );
      }
    //cout << endl;
  }

  // compute bound
  //hk_relax.print();
  int bound = hk_relax.compute( obj_var.getMax() );
  //	cout << "Current MDD Lower bound: " << obj_var << endl;
  //	cout << "Held-Karps bound: " << bound << endl;


  if( bound > obj_var.getMin() ) {
    //		cout << "Improved!" << endl;
    //		cout << "\tCurrent MDD Lower bound: " << obj_var << endl;
    //		cout << "\tHeld-Karps bound: " << bound << endl;
    get_CP().add( obj_var >= bound );
  }
  //cout << endl;
  //exit(1);
}



// -----------------------------------------------------------
// Begin Node Branching
// -----------------------------------------------------------


int* DisjunctivePropagator::get_shortest_path() {

  //	cout << "Calculating shortest path"<<endl;
  // compute shortest path value and store it in terminal node
  RevMDD* mdd = get_mdd();

  //Calculate the earliest start times of the activities on the arcs
  //and maintain pointers from arcs in layer l+1 to arcs in layer l,
  //thereby identifying the shortest path.
  calc_shortest_path();

  //Determine the incoming arc in the terminal state which has the earliest start time of the terminal activity
  terminal_state()->est_tmp = INF;
  RevMDD::Arc* last = NULL;
  //  cout << "terminal_state()->in_est->size: "<< to_string(terminal_state()->in_est->size) <<endl;
  for( int i = 0; i < terminal_state()->in_est->size; ++i ) {
    if( terminal_state()->est_tmp > terminal_state()->in_est->value(terminal_state()->in_est->list[i]) ) {
      terminal_state()->est_tmp = terminal_state()->in_est->value(terminal_state()->in_est->list[i]);
      last = (RevMDD::Arc*)terminal_state()->in_est->pointer(terminal_state()->in_est->list[i]);
    }
  }

  //	cout<< "Terminal EST: "<<to_string(terminal_state()->est_tmp) << " Last: " << to_string(((RevMDD::Arc*)last)->get_val())<<endl;

  //Check whether shortest path could be found. If the shortest path has length INF, no shortest path exist (MDD is disconnected)
  if( terminal_state()->est_tmp == INF ) {
    IloCP cp = get_CP();
    cp.add( obj_var >= INF);
    //cout << "ERROR: MDD is not feasible" << endl;
    //exit(1);
    cout << "WARNING: NO shortest path exists in the MDD (MDD is disconnected)" << endl;
    char filename[256];
    sprintf(filename, "graphs/disconnected.gml");
    mdd->export_to_gml(filename);
    return NULL;
  }

  // compose optimum solution
  int* optimum = terminal_state()->optimum_sol;
  int l = permutation.getSize()-1;
  while( last != NULL ) {
    optimum[l--] = ((RevMDD::Arc*)last)->get_val();
    last = last->pred_est;
  }

  //	cout << "Optimum solution: ";
  //	for( int i = 0; i < (int)acts.size(); ++i ) {
  //		cout << optimum[i] << " ";
  //	}
  //	cout << endl;

  return optimum;
}



/**
 * Prepares data structures to compute a shortest path in the graph
 */
void DisjunctivePropagator::calc_shortest_path() {
  //	cout << "Compute shortest path"<<endl;

  // verify shortest path
  RevMDD* mdd = get_mdd();

  // clean nodes
  for( int l = 0; l < mdd->get_num_layers()-1; ++l ) {
    for( int w = 0; w < mdd->get_max_width(); ++w ) {
      states[l+1][w]->in_est->clear();
      states[l+1][w]->est_tmp = INF;
    }
  }
  root_state()->est_tmp = 0;
  root_state()->in_est->clear();


  // compute shortest path

  DisjunctiveState *source_state, *target_state;
  RevMDD::Arc* prev_arc = NULL;

  // first layer
  for( RevMDD::Arc* arc = mdd->arcs_layer[0].get_first();
       arc != NULL;
       arc = mdd->arcs_layer[0].get_next(arc) )
    {
      states[1][arc->get_target()]->in_est->add(arc->get_val(), 0, arc);
      arc->pred_est = NULL;
    }

  // remaining layers
  int est_arc, est_arc_tmp;
  for( int l = 1; l < mdd->get_num_layers()-1; ++l ) {
    for( RevMDD::Arc* arc = mdd->arcs_layer[l].get_first();
	 arc != NULL;
	 arc = mdd->arcs_layer[l].get_next(arc) )
      {
	source_state = states[l][arc->get_source()];
	target_state = states[l+1][arc->get_target()];

	prev_arc=NULL;
	est_arc = INF;
	int prev_value;
	for( int i = 0; i < source_state->in_est->size; ++i ) { //Find earliest start time for source state + travel time to new state
	  prev_value = source_state->in_est->list[i];
	  if( prev_value != arc->get_val()
	      &&
	      !precedes_activity(arc->get_val(), prev_value) )
	    {

	      assert(setup[prev_value][arc->get_val()] != INF);
	      est_arc_tmp = source_state->in_est->value(prev_value)
		+ mult[l]*setup[prev_value][arc->get_val()]; //MARK_JK
	      if( est_arc_tmp < est_arc ) {
		est_arc = est_arc_tmp;
		prev_arc = (RevMDD::Arc*)source_state->in_est->pointer(prev_value);
	      }
	    }
	}

	//      cout<< "Source state:  ("<<to_string(l)<<","<<to_string(arc->get_source()) << ") ";
	//      cout<< "Dest state: "<<to_string(l+1)<<","<<to_string(arc->get_target()) << ") ";
	//      cout<< "Arc act:" << to_string(arc->get_val());
	//      cout<< " EST arc act: " << to_string(est_arc);
	//      cout<< " Prev arc act: " << prev_arc->get_val()<<endl;

	//      assert(prev_arc != NULL);
	arc->pred_est = prev_arc;

	if( est_arc != INF ) {
	  target_state->in_est->add(arc->get_val(), est_arc, arc);
	  assert(target_state->in_est->contains(arc->get_val()));
	  //TEMP CODE
	  //        if(target_state==terminal_state()){
	  //        	cout<< "Adding (" << arc->get_val() << ", "<< est_arc << " to terminal_state in_est"<<endl;
	  //        }
	  //END TEMP CODE
	}
      }
  }
}

// -----------------------------------------------------------
// End Node Branching
// -----------------------------------------------------------



//
// Node for flow model
//
struct FlowNode {
  int act;
  IloNumVarArray in_arcs;
  IloNumVarArray out_arcs;

  FlowNode(IloEnv env, int _act) 
    : act(_act),
      in_arcs(IloNumVarArray(env)),
      out_arcs(IloNumVarArray(env))
  { }
};

/** 
 * Create MDD flow model 
 * is_in: binary variables for each activity i, indicating if
 *        that variable belongs to the solution
 */
IloModel DisjunctivePropagator::get_flow_model(IloNumVarArray &is_in) {

  RevMDD* mdd = get_mdd();
  
  IloEnv env;
  IloModel model(env);

  // objective function
  IloExpr obj(env);

  // Variables of each activity
  IloArray<IloNumVarArray> act_array(env, acts.size());
  for (int i = 0; i < (int)acts.size(); ++i) {
    act_array[i] = IloNumVarArray(env);
  }

  // nodes of the flow formulation per mdd layer / mdd node
  vector< vector< vector<FlowNode*> > > flow_nodes(mdd->get_num_layers(),
						   vector< vector<FlowNode*> >(mdd->get_max_width()));

  
  // root mdd node has flow source node
  flow_nodes[0][0].push_back( new(env) FlowNode(env, -1) );

  // cost

  // create nodes
  for (int l = 0; l < mdd->get_num_layers(); ++l) {
    for( RevMDD::Arc* arc = mdd->arcs_layer[l].get_first();
	 arc != NULL;
	 arc = mdd->arcs_layer[l].get_next(arc) )
      {
	// Create new flow node for arc
	FlowNode* arc_node = new(env) FlowNode(env, arc->get_val());
	flow_nodes[l+1][arc->get_target()].push_back(arc_node);

	for (vector<FlowNode*>::iterator it = flow_nodes[l][arc->get_source()].begin(); it != flow_nodes[l][arc->get_source()].end(); ++it) {
	  FlowNode* parent = (*it);
	  if (parent->act == -1 || !precedes_activity(arc->get_val(), parent->act)) {
	    IloNumVar flow(env, 0, 1);
	    arc_node->in_arcs.add(flow);
	    parent->out_arcs.add(flow);

	    act_array[arc->get_val()].add(flow);

	    if (parent->act != -1) {
	      act_array[parent->act].add(flow);
	      obj += setup[parent->act][arc->get_val()] * flow;
	    }
	  }
	}
      }
  }

  // create flow balance constraints  

  // root node
  model.add( 1 == IloSum(flow_nodes[0][0][0]->out_arcs) );

  // intermmediate nodes
  for (int l = 1; l < mdd->get_num_layers()-1; ++l) {
    for (int w = 0; w < mdd->get_max_width(); ++w) {
      for (vector<FlowNode*>::iterator it = flow_nodes[l][w].begin(); it != flow_nodes[l][w].end(); ++it) {
	model.add( IloSum((*it)->out_arcs) == IloSum((*it)->in_arcs) );
      }    
    }
  }

  // terminal node
  model.add( IloSum(flow_nodes[mdd->get_num_layers()-1][0][0]->in_arcs) == 1);

  // in flow variables
  is_in = IloNumVarArray(env, acts.size(), 0, 1);
  for (int i = 0; i < (int)acts.size(); ++i) {
    model.add( is_in[i] <= IloSum(act_array[i]));
    for (int j = 0; j < act_array[i].getSize(); ++j) {
      model.add( is_in[i] >= act_array[i][j] );
    }
  }

  // add objective function
  model.add( IloMinimize(env, obj) );

  //exit(1);
  return model;
}




/**
 * Called when all variables are fixed.
 * For this constraint, we just fix the costs
 **/
void DisjunctivePropagator::all_vars_fixed() {

  IloCP cp = get_CP();

  int start_time;
  int sum_setup_val = 0;
  int makespan_val = acts[permutation[0].getValue()]->release 
    + acts[permutation[0].getValue()]->processing;
  int tardiness = 0;
  int max_tardiness = 0;
  int weighted_completion = acts[permutation[0].getValue()]->weight * makespan_val;

  // minimum time for first activity in the resource
  if (optional) {
    opt_acts[permutation[0].getValue()].setStartMin(MAX(acts[permutation[0].getValue()]->release, 
    							opt_acts[permutation[0].getValue()].getStartMin()));
  } else {
    acts[permutation[0].getValue()]->start_time.setStartMin(
							    MAX(acts[permutation[0].getValue()]->release, 
								acts[permutation[0].getValue()]->start_time.getStartMin()));
  }
  
  for( int i = 1; i < permutation.getSize(); i++ ) {

    // consider optional activities
    if (optional && permutation[i-1].getValue() == terminal_opt_act) {
      break;
    }

    // minimum start time of activity
    if (optional) {
      start_time = MAX(opt_acts[permutation[i].getValue()].getStartMin(),
      	       makespan_val + setup[acts[permutation[i-1].getValue()]->id][acts[permutation[i].getValue()]->id]);
    } else {
      start_time = MAX(acts[permutation[i].getValue()]->start_time.getStartMin(),
		       makespan_val + setup[acts[permutation[i-1].getValue()]->id][acts[permutation[i].getValue()]->id]);
    }

    // impose precedence relation with respect to previous activity

    if (optional) {
      IlcEndBeforeStart(opt_acts[permutation[i-1].getValue()],
      			opt_acts[permutation[i].getValue()],
      			setup[acts[permutation[i-1].getValue()]->id][acts[permutation[i].getValue()]->id]);
    } else {
      IlcEndBeforeStart(acts[permutation[i-1].getValue()]->start_time,
			acts[permutation[i].getValue()]->start_time,
			setup[acts[permutation[i-1].getValue()]->id][acts[permutation[i].getValue()]->id]);
    }

    // compute makespan / sum of setup times
    sum_setup_val += mult[i]*setup[acts[permutation[i-1].getValue()]->id][acts[permutation[i].getValue()]->id];
    makespan_val = start_time + acts[permutation[i].getValue()]->processing;

    // compute tardiness
    int tardiness_penalty = acts[permutation[i].getValue()]->weight *
      MAX(0, makespan_val - acts[permutation[i].getValue()]->duedate);
    tardiness += tardiness_penalty;
    max_tardiness = MAX(max_tardiness, tardiness_penalty);

    weighted_completion += acts[permutation[i].getValue()]->weight * makespan_val;
  }

  // cout << "Permutation: " << endl;
  // for( int i = 0; i < permutation.getSize(); ++i ) {
  //   cout << "\t" << permutation[i].getValue() << endl;
  // }
  //cout << "sum-setup: " << sum_setup_val << endl << endl;

  // update objective function
  switch( obj_type ) {

  case Makespan:
    cp.add( obj_var >= makespan_val );
    break;

  case SumSetupTimes:
    // cout << "sum setup=" << sum_setup_val << " - obj = " << obj_var << endl;
    // if (obj_var_lag.getImpl() != NULL) {
    //   cout << "lagrangian = " << obj_var_lag << endl;
    // }
    cp.add( obj_var == sum_setup_val );
    break;

  case TotalTardiness:
    cp.add( obj_var >= tardiness );
    break;

  case MaxTardiness:
    cp.add( obj_var >= max_tardiness );
    break;

  case Latency:
    cp.add( obj_var >= weighted_completion );
    break;

  default:
    break;
  }



  // -----------------------------------------------------------
  // Begin Node Branching
  // -----------------------------------------------------------

  // set all layers and MDD as exact
  for (int l = 1; l < get_mdd()->get_num_layers()-1; ++l) {
    is_layer_exact_v[l]->setValue(cp, true);
  }
  is_mdd_exact_v.setValue(cp, true);

  // -----------------------------------------------------------
  // End Node Branching
  // -----------------------------------------------------------


}





