//============================================================================
// Sequence MDD Propagator - Implementation File
//============================================================================

#include <algorithm>
#include <queue>
#include "sequence_mdd.hpp"

using namespace std;


/**
 * Reserve node state/pre-processing memory
 */
void SequenceMDDPropagator::reserve_memory() {

	IloCP cp = get_CP();

	//cout << "[SequenceMDD] Reserving memory... \n";

	// Values in/out sequence set
	int max_val = vars.getMaxMax();

	in_set = new( cp.getHeap() ) bool[max_val+1];
	memset(in_set, false, sizeof(bool)*(max_val+1));

	in_set_val = new( cp.getHeap() ) int[max_val+1];
	memset(in_set_val, 0, sizeof(int)*(max_val+1));

	for( vector<int>::iterator val = values.begin(); val != values.end(); val++ ) {
		assert( *val >= 0 && *val <= max_val );
		in_set[*val] = true;
		in_set_val[*val] = 1;
	}

	// Node states

	RevMDD* mdd = get_mdd();
	int num_layers = mdd->get_num_layers();

	states = new( cp.getHeap() ) SequenceState**[num_layers];

	// root and terminal nodes
	states[0]    = new( cp.getHeap() ) SequenceState*[1];
	states[0][0] = new( cp.getHeap() ) SequenceState(cp, length, 0, 0);

	states[num_layers-1]    = new( cp.getHeap() ) SequenceState*[1];
	states[num_layers-1][0] = new( cp.getHeap() ) SequenceState(cp, length, 0, num_layers-1);

	// remaining nodes
	for( int l = 1; l < num_layers-1; l++ ) {
		states[l] = new( cp.getHeap() ) SequenceState*[mdd->get_max_width()];
		for( int n = 0; n < mdd->get_max_width(); n++) {
			states[l][n] = new( cp.getHeap() ) SequenceState(cp, length, 0, l);
		}
	}

	// layer states
	// TODO: do not allocate if stronger propagation (now it is been
	// allocated regardless for testing purposes)
	layer_lb = new( cp.getHeap() ) int[mdd->get_num_layers()];
	layer_ub = new( cp.getHeap() ) int[mdd->get_num_layers()];
	//for( int l = 0; l < mdd->get_num_layers(); l++ ) {
	//	layer_lb[l] = new( cp.getHeap() ) int;
	//	layer_ub[l] = new( cp.getHeap() ) IlcRevInt(cp, l);
	//}


	// Node sketches
	sketches = new( cp.getHeap() ) SequenceNodeSketch***[mdd->get_max_width()];
	sketch_init = new( cp.getHeap() ) SketchStatus**[mdd->get_max_width()];
	for( int i = 0; i < mdd->get_max_width(); i++ ) {
		sketches[i] = new( cp.getHeap() ) SequenceNodeSketch**[mdd->get_max_width()];
		sketch_init[i] = new( cp.getHeap() ) SketchStatus*[mdd->get_max_width()];
		for( int j = 0; j < mdd->get_max_width(); j++ ) {
			sketches[i][j] = new( cp.getHeap() ) SequenceNodeSketch*[2];
			sketch_init[i][j] = new( cp.getHeap() ) SketchStatus[2];
			for( int k = 0; k < 2; k++ ) {
				sketches[i][j][k] = new( cp.getHeap() ) SequenceNodeSketch;
				sketches[i][j][k]->timestamp = -1;
			}
		}
	}


	// others
	node_verified = new( cp.getHeap() ) bool[mdd->get_max_width()];
	arcs_to_move = new( cp.getHeap() ) SimpleList<RevMDD::Arc*>(cp, mdd->get_max_arc_id()+1);

	has_in_arc = new( cp.getHeap() ) bool**[mdd->get_max_width()];
	for( int w = 0; w < mdd->get_max_width(); w++ ) {
		has_in_arc[w] = new( cp.getHeap() ) bool*[mdd->get_max_width()];
		for( int u = 0; u < mdd->get_max_width(); u++ ) {
			has_in_arc[w][u] = new( cp.getHeap() ) bool[2];
		}
	}

	available_nodes = new( cp.getHeap() ) SimpleList<int>(cp, mdd->get_max_width());

	group_lb = new( cp.getHeap() ) int[mdd->get_max_width()];
	group_ub = new( cp.getHeap() ) int[mdd->get_max_width()];

	// arc --> sketch map
	arc_sketch = new( cp.getHeap() ) SequenceNodeSketch*[mdd->get_max_arc_id()+1];

	// auxiliaries for bounds computation
	aux_lb = new( cp.getHeap() ) int[length];
	aux_ub = new( cp.getHeap() ) int[length];

	// modified states per node
	state_modified = new( cp.getHeap() ) bool[mdd->get_max_width()];

	//cout << "[SequenceMDD] done.\n";
}




/**
 * Setup the state of a node according to its incoming arcs
 */
void SequenceMDDPropagator::setup_node_incoming(int layer, int node, RefinementGroup ref_group, bool new_node) {

  /*
  if (prop_id == 6) {
	cout << "\n[SequenceMDD] Setting state of node (layer: " << layer << ", id: " << node << ") - prop: " << prop_id << endl;
//
	cout << "\t\tIncoming arcs: ";
	for( int i = 0; i < ref_group->size; i++ ) {
		RevMDD::Arc* arc = ref_group->elements[i];
		cout << "(" << layer-1 << "," << arc->get_source() << "," << arc->get_val() << ") ";
	}
	cout << endl;
  }
  */

	// ===========================================================================================
	// First make sure that sketches of all incoming arcs of this node are updated. Also
	// mark sketches as unverified for state computation purposes.
	// ===========================================================================================

	for( int i = 0; i < ref_group->size; i++ ) {
		update_sketch_incoming(layer-1, ref_group->elements[i], new_node)->verified = false;
		assert( arc_sketch[ref_group->elements[i]->get_layer_id()]->feasible );
	}


	// ===========================================================================================
	// Compute new node state
	// ===========================================================================================

	// node state
	SequenceState* state = get_node_state(layer, node);
	SequenceState* source_state = NULL;

	// initialize states
	// TODO: special initialization for weaker propagation
	memset(aux_ub, 0, sizeof(int)*length);
	for( int l = 0; l < length; l++ ) {
		aux_lb[l] = INF;
	}

	// reset nodes considered
	if (prop_type == StrongProp) {
	  memset(node_verified, false, sizeof(bool)*get_mdd()->get_max_width());
	}
	
	// collect states from incoming arcs
	RevMDD::Arc* arc;
	int lb,ub;

	int a_lb, a_ub; // ancestor bounds

	for( int i = 0; i < ref_group->size; i++ ) {
		arc = ref_group->elements[i];

		if (prop_type == StrongProp) {

		  // source node was never verified, update ancestors
		  if( !node_verified[arc->get_source()] )  {  // TODO: change here!

		    // update ancestors information
		    source_state = get_node_state(layer-1, arc->get_source());
		    //if (prop_id == 6) {
		    //  cout << "\t\tAncestors from node " << arc->get_source() << endl;
		    //  for( int l = 1; l < length; l++ ) {
		    //cout << "\t\t\tl = " << l << " - LB = " << source_state->ancestor_lb[l-1]->getValue() << endl;
		    //	cout << "\t\t\tl = " << l << " - UB = " << source_state->ancestor_ub[l-1]->getValue() << endl;
		    //}
		    //}
		    for( int l = 1; l < length; l++ ) {
			  a_lb = source_state->ancestor_lb[l-1]->getValue();
			  a_ub = source_state->ancestor_ub[l-1]->getValue();
			  aux_lb[l] = MIN(aux_lb[l], a_lb);
			  aux_ub[l] = MAX(aux_ub[l], a_ub);
		    }
		    
		    node_verified[arc->get_source()] = true;
		  }
		}
		  
		// if arc type was not set, update bounds
		if( !arc_sketch[arc->get_layer_id()]->verified ) {

			arc_sketch[arc->get_layer_id()]->verified = true;

			lb = get_sketch_lb(arc);
			ub = get_sketch_ub(arc);
//
//			cout << "\t\tarc (";
//			cout << arc->get_source();
//			cout << "," << arc->get_val();
//			cout << ") - LB = " << lb;
//			cout << " - UB = " << ub;
//			cout << endl;

			aux_lb[0] = MIN(aux_lb[0], lb);
			aux_ub[0] = MAX(aux_ub[0], ub);
		}
	}

	// ===========================================================================================
	// Update final state. If node state has changed, then incoming and outgoing nodes must
	// be updated
	// ===========================================================================================

	// update final state
	IloCP cp = get_CP();

	//bool state_changed = false;

	if (prop_type == WeakProp) {
	  for( int l = 0; l < 1; l++ ) {
	    if( aux_lb[l] != state->ancestor_lb[l]->getValue() ) {
	      state->ancestor_lb[l]->setValue(cp, aux_lb[l]);
	      //state_changed = true;
	    }
	    if( aux_ub[l] != state->ancestor_ub[l]->getValue() ) {
	      state->ancestor_ub[l]->setValue(cp, aux_ub[l]);
	      //state_changed = true;
	    }
	  }

	} else if (prop_type == StrongProp) {
	  for( int l = 0; l < length; l++ ) {
	    if( aux_lb[l] != state->ancestor_lb[l]->getValue() ) {
	      state->ancestor_lb[l]->setValue(cp, aux_lb[l]);
	      //state_changed = true;
	    }
	    if( aux_ub[l] != state->ancestor_ub[l]->getValue() ) {
	      state->ancestor_ub[l]->setValue(cp, aux_ub[l]);
	      //state_changed = true;
	    }
	  }
	}


	// update layer info
	layer_lb[layer] = MIN( layer_lb[layer], state->y_lb.getValue() );
	layer_ub[layer] = MAX( layer_ub[layer], state->y_ub.getValue() );

	// cout << "Incoming: " << endl;
	// cout << "\tLayer " << layer << " - lb = " << layer_lb[layer];
	// cout << " - ub = " << layer_ub[layer] << endl;
}



/**
 * Called after a new node is created with its respective outgoing arcs
 */
void SequenceMDDPropagator::setup_node_outgoing(int layer, int node, RefinementGroup ref_group, bool new_node) {

	//cout << "\n[SequenceMDD] Setting state of node (layer: " << layer << ", id: " << node << ")" << endl;

	// ===========================================================================================
	// First make sure that sketches of all outgoing arcs of this node are updated. Also
	// mark sketches as unverified for state computation purposes.
	// ===========================================================================================

	for( int i = 0; i < ref_group->size; i++ ) {
		update_sketch_outgoing(layer, ref_group->elements[i])->verified = false;
		assert( arc_sketch[ref_group->elements[i]->get_layer_id()]->feasible );
	}


	// ===========================================================================================
	// Compute node state
	// ===========================================================================================

	// node state
	SequenceState* state = get_node_state(layer, node);
	SequenceState* target_state = NULL;

	// initialize states
	memset(aux_ub, 0, sizeof(int)*length);
	for( int l = 0; l < length; l++ ) {
		aux_lb[l] = INF;
	}

	// reset nodes considered
	if (prop_type == StrongProp) {
	  memset(node_verified, false, sizeof(bool)*get_mdd()->get_max_width());
	} 

	// collect states from incoming arcs
	RevMDD::Arc* arc;
	int lb,ub;
	int target_node;

	// descendant bounds
	int d_lb, d_ub;

	for( int i = 0; i < ref_group->size; i++ ) {

		arc = ref_group->elements[i];
		assert( arc->get_source() == node );

		target_node = arc->get_target();

		if (prop_type == StrongProp) {
		  // if target nodes was never verified, update ancestors
		  if( !node_verified[target_node] ) {   // TODO: change here!!
		    // update ancestors information
		    target_state = get_node_state(layer+1, target_node);
		    
		    //			cout << "\t\tAncestor for node " << w << endl;
		    //			for( int l = 1; l < length; l++ ) {
		    //				cout << "\t\t\tLB = " << source_state->ancestor_lb[l-1] << endl;
		    //				cout << "\t\t\tUB = " << source_state->ancestor_ub[l-1] << endl;
		    //			}
		    
		    for( int l = 1; l < length; l++ ) {
		      d_lb = target_state->descendant_lb[l-1]->getValue();
		      d_ub = target_state->descendant_ub[l-1]->getValue();
		      aux_lb[l] = MIN(aux_lb[l], d_lb);
		      aux_ub[l] = MAX(aux_ub[l], d_ub);     		}
		    
		    node_verified[target_node] = true;
		  }
		}
		  
		// if arc type was not set, update bounds
		//cout << "\t" << arc_sketch[arc->get_layer_id()]->n_types << endl;
		if( !arc_sketch[arc->get_layer_id()]->verified ) {

			arc_sketch[arc->get_layer_id()]->verified = true;

			lb = get_sketch_lb(arc);
			ub = get_sketch_ub(arc);
			aux_lb[0] = MIN(aux_lb[0], lb);
			aux_ub[0] = MAX(aux_ub[0], ub);

			//has_in_arc[arc->get_source()][arc->get_target()][in_set_val[arc->get_val()]] = true;
		}
	}

	// ===========================================================================================
	// Update final state. If node state has changed, then incoming and outgoing nodes must
	// be updated
	// ===========================================================================================

	// update final state
	IloCP cp = get_CP();

	//bool state_changed = false;

	if (prop_type == WeakProp) {
	  for( int l = 0; l < 1; l++ ) {
	    if( aux_lb[l] != state->descendant_lb[l]->getValue() ) {
	      state->descendant_lb[l]->setValue(cp, aux_lb[l]);
	      //state_changed = true;
	    }
	    if( aux_ub[l] != state->descendant_ub[l]->getValue() ) {
	      state->descendant_ub[l]->setValue(cp, aux_ub[l]);
	      //state_changed = true;
	    }
	  }

	} else if (prop_type == StrongProp) {
	  for( int l = 0; l < length; l++ ) {
	    if( aux_lb[l] != state->descendant_lb[l]->getValue() ) {
	      state->descendant_lb[l]->setValue(cp, aux_lb[l]);
	      //state_changed = true;
	    }
	    if( aux_ub[l] != state->descendant_ub[l]->getValue() ) {
	      state->descendant_ub[l]->setValue(cp, aux_ub[l]);
	      //state_changed = true;
	    }
	  }
	}

	// update layer info
	layer_lb[layer] = MIN( layer_lb[layer], state->y_lb.getValue() );
	layer_ub[layer] = MAX( layer_ub[layer], state->y_ub.getValue() );
}

/**
 * Partition refinement group (refinement). Max MDD width must be observed
 */
void SequenceMDDPropagator::partition_refinement_group(int layer, RefinementGroup* ref_groups, int &n_groups) {

  if (!must_refine) {
    return;
  }

	RevMDD* mdd = get_mdd();
	assert( n_groups < mdd->get_max_width( ));


	// ===========================================================================================
	// First make sure that sketches of all incoming arcs in this layer are updated.
	// ===========================================================================================

	for( RevMDD::Arc* arc = mdd->arcs_layer[layer].get_first();
			arc != NULL;
			arc = mdd->arcs_layer[layer].get_next(arc) )
	{
		assert( update_sketch_incoming(layer, arc)->feasible );
		update_sketch_incoming(layer, arc);
	}

	//cout << "\n[SequenceMDD] Refining layer... " << endl;


	// ------------------------------------------------------------------------
	// Refine groups according to lower LB / higher UB

	// create priority queue of groups
	RefGroup_LB_Comparator refgroup_comparator(group_lb, group_ub);
	priority_queue<int, vector<int>, RefGroup_LB_Comparator> group_queue(refgroup_comparator);

	// reset node availability
	available_nodes->clear();

	// create comparator for groups
	Arcs_LB_Descending group_comp(sketches, in_set_val);

	// sort each group according to lower bound
	for( int g = 0; g < mdd->get_max_width(); g++ ) {

		if( ref_groups[g]->is_empty() ) {

			// if group is empty, add to available nodes
			available_nodes->add(g);

		} else {

			// sort groups
			sort(ref_groups[g]->elements, ref_groups[g]->elements+ref_groups[g]->size, group_comp);

			// print
//			cout << "\tGroup " << g << endl;
//			for( int i = 0; i < ref_groups[g]->size; i++ ) {
//				cout << "\t\tarc (";
//				cout << ref_groups[g]->elements[i]->get_source();
//				cout << "," << ref_groups[g]->elements[i]->val;
//				cout << ") - LB = " << get_sketch_lb(ref_groups[g]->elements[i]);
//				cout << " - UB = " << get_sketch_ub(ref_groups[g]->elements[i]);
//				cout << endl;
//			}

			// if group has distinct LB/UB, add it to queue, taking into account smallest LB / largest UB
			if( group_distinct_lb(ref_groups[g]) || group_distinct_ub(ref_groups[g]) ) {
				group_lb[g] = get_sketch_lb(ref_groups[g]->get_last());
				group_ub[g] = get_sketch_ub(ref_groups[g]->get_last());

				group_queue.push(g);

//				cout << "\n\t\tadded to queue - LB = " << group_lb[g] << " - UB: " << group_ub[g] << endl;
			}

//			cout << endl;
		}
	}

	assert(mdd->get_max_width() == n_groups + available_nodes->size);

	int group_ref, new_group, bound;
	while( n_groups < mdd->get_max_width() && !group_queue.empty() ) {

		// select group to refine
		group_ref = group_queue.top();
		group_queue.pop();

//		cout << "\tGroup to refine: " << group_ref << endl;

		// take new node
		new_group = available_nodes->pop_up();

		// increment number of existing groups
		n_groups++;

		// transfer arcs to new group
		if( group_distinct_lb(ref_groups[group_ref]) ) {

			// transfer according to LB
			bound = get_sketch_lb(ref_groups[group_ref]->get_last());
			while( bound == get_sketch_lb(ref_groups[group_ref]->get_last()) ) {

			        // // we will globally remove an arc between two nodes: all states must be updated
				// mdd->modified_out_node(layer, ref_groups[group_ref]->get_last()->get_source());
				// mdd->modified_in_node(layer+1, group_ref);
				// state_modified[new_group] = true;

				// mdd->modified_in_node(layer+1, new_group);
				// mdd->modified_out_node(layer+1, new_group);

				ref_groups[new_group]->add( ref_groups[group_ref]->pop_up() );
			}

		} else {

			// transfer according to UB
			bound = get_sketch_ub(ref_groups[group_ref]->get_last());
			while( bound == get_sketch_ub(ref_groups[group_ref]->get_last()) ) {

				// mdd->modified_out_node(layer, ref_groups[group_ref]->get_last()->get_source());
				// mdd->modified_in_node(layer+1, group_ref);
				// state_modified[new_group] = true;

				// mdd->modified_in_node(layer+1, new_group);
				// mdd->modified_out_node(layer+1, new_group);

				ref_groups[new_group]->add( ref_groups[group_ref]->pop_up() );
			}
		}

		assert( ref_groups[group_ref]->size > 0 );
		assert( ref_groups[new_group]->size > 0 );

		// check if old group can be refined (notice it is already sorted)

//		cout << "\n\tOld group: " << group_ref << endl;
//		for( int i = 0; i < ref_groups[group_ref]->size; i++ ) {
//			cout << "\t\tarc (";
//			cout << ref_groups[group_ref]->elements[i]->get_source();
//			cout << "," << ref_groups[group_ref]->elements[i]->val;
//			cout << ") - LB = " << get_sketch_lb(ref_groups[group_ref]->elements[i]);
//			cout << " - UB = " << get_sketch_ub(ref_groups[group_ref]->elements[i]);
//			cout << endl;
//		}

		if( group_distinct_lb(ref_groups[group_ref]) || group_distinct_ub(ref_groups[group_ref]) ) {

			group_lb[group_ref] = get_sketch_lb(ref_groups[group_ref]->get_last());
			group_ub[group_ref] = get_sketch_ub(ref_groups[group_ref]->get_last());

			group_queue.push(group_ref);

//			cout << "\n\t\tadded to queue - LB = " << group_lb[group_ref] << " - UB: " << group_ub[group_ref] << endl;
		}


		// check if new group can be refined as well

		sort(ref_groups[new_group]->elements, ref_groups[new_group]->elements+ref_groups[new_group]->size, group_comp);

		// print
//		cout << "\n\tNew group: " << new_group << endl;
//		for( int i = 0; i < ref_groups[new_group]->size; i++ ) {
//			cout << "\t\tarc (";
//			cout << ref_groups[new_group]->elements[i]->get_source();
//			cout << "," << ref_groups[new_group]->elements[i]->val;
//			cout << ") - LB = " << get_sketch_lb(ref_groups[new_group]->elements[i]);
//			cout << " - UB = " << get_sketch_ub(ref_groups[new_group]->elements[i]);;
//			cout << endl;
//		}

		// if group has distinct LB/UB, add it to queue, taking into account smallest LB / largest UB
		if( group_distinct_lb(ref_groups[new_group]) || group_distinct_ub(ref_groups[new_group]) ) {
			group_lb[new_group] = get_sketch_lb(ref_groups[new_group]->get_last());
			group_ub[new_group] = get_sketch_ub(ref_groups[new_group]->get_last());

			group_queue.push(new_group);

//			cout << "\n\t\tadded to queue - LB = " << group_lb[new_group] << " - UB: " << group_ub[new_group] << endl;
		}

	}

//	cout << endl << endl;

	// ------------------------------------------------------------------------
}


/**
 * Process layer in a top-down perspective
 */
void SequenceMDDPropagator::filter_layer_topdown(int layer) {

  RevMDD* mdd = get_mdd();
  IloCP cp = get_CP();

  // pointer to the previous arc considered, for removal purposes
  RevMDD::Arc* previous = mdd->arcs_layer[layer].get_header();

  // remove infeasible arcs from layer
  for( RevMDD::Arc* arc = mdd->arcs_layer[layer].get_first();
       arc != NULL;
       arc = mdd->arcs_layer[layer].get_next(arc) )
    {
	if( verify_incoming_arc(layer, arc) ) {
	  // arc is feasible: update sketch info
	  previous = arc;

	} else {
	  // arc is infeasible: remove from layer
	  previous->next.setValue(cp, arc->next.getValue());
	  //cout << "[SequenceMDD] Removing arc (" << layer << "," << arc->get_source() << "," << arc->get_val() << ")" << endl;
	}
    }

  // clean modified state
  //memset( state_modified, false, sizeof(bool)*mdd->get_max_width());
}

/**
 * Process layer in a top-down perspective
 */
void SequenceMDDPropagator::filter_layer_bottomup(int layer) {

  RevMDD* mdd = get_mdd();
  IloCP cp = get_CP();

  // pointer to the previous arc considered, for removal purposes
  RevMDD::Arc* previous = mdd->arcs_layer[layer].get_header();

  //cout << "\n[SequenceMDD] Filter layer bottomup " << layer << endl;

  // remove infeasible arcs from layer
  for( RevMDD::Arc* arc = mdd->arcs_layer[layer].get_first();
       arc != NULL;
       arc = mdd->arcs_layer[layer].get_next(arc) )
    {
      if( verify_outgoing_arc(layer, arc) ) {

	// arc is feasible: update sketch info
	previous = arc;

      } else {
	// arc is infeasible; remove it from layer
	previous->next.setValue(cp, arc->next.getValue());
      }
    }

  // clean modified state
  //memset( state_modified, false, sizeof(bool)*mdd->get_max_width());
}


/**
 * Print node states
 **/
void SequenceMDDPropagator::print_states() {

	RevMDD* mdd = get_mdd();

	cout << "[SequenceMDD] States for propagator " << prop_id << endl;

	for( int l = 0; l < mdd->get_num_layers(); l++ ) {

		memset(state_modified, false, sizeof(bool)*mdd->get_max_width());

		cout << "=========================================" << endl;
		cout << "Layer " << l << endl;

		for( RevMDD::Arc* arc = mdd->arcs_layer[l].get_first();
				arc != NULL;
				arc = mdd->arcs_layer[l].get_next(arc) )
		{
			state_modified[arc->get_source()] = true;
		}

		if( l == mdd->get_num_layers() -1 ) {
			state_modified[0] = true;
		}

		for( int w = 0; w < mdd->get_max_width(); w++ ) {
			if( !state_modified[w] ) {
				continue;
			}

			cout << "\tNode " << w << endl;
			SequenceState* state = get_node_state(l, w);

			assert( state->ancestor_lb[0] == state->descendant_lb[0] );
			assert( state->ancestor_ub[0] == state->descendant_ub[0] );

			cout << "\t\tAncestors: " << endl;
			for( int l = 0; l < length; l++ ) {
				cout << "\t\t\tl: " << l << " - lb: " << state->ancestor_lb[l]->getValue() << " - ub: " << state->ancestor_ub[l]->getValue() << endl;
			}
			cout << "\t\tDescendants: " << endl;
			for( int l = 0; l < length; l++ ) {
				cout << "\t\t\tl: " << l << " - lb: " << state->descendant_lb[l]->getValue() << " - ub: " << state->descendant_ub[l]->getValue() << endl;
			}



			cout << endl;
		}
		cout << endl;
	}

}

/**
 * Update layer information
 **/
void SequenceMDDPropagator::update_layer(int layer, vector<int> &layer_nodes) {

  cout << "Not implemented" << endl;
  exit(1);

  if (prop_type == StrongProp) {
    return;
  }

	int lb = INF, lb_tmp;
	int ub = -1, ub_tmp;

	// obtain cumulative variable bound
	for( vector<int>::iterator n = layer_nodes.begin(); n != layer_nodes.end(); n++ ) {
		lb_tmp = get_node_state(layer, *n)->y_lb.getValue();
		ub_tmp = get_node_state(layer, *n)->y_ub.getValue();

		lb = MIN(lb, lb_tmp);
		ub = MAX(ub, ub_tmp);
	}

	//bool state_changed = false;

	//if( lb > layer_lb[layer]->getValue()) {
	  //state_changed = true;
	  //!! layer_lb[layer]->setValue(get_CP(), lb);
	//}

	//if( ub < layer_ub[layer]->getValue()) {
	  //state_changed = true;
	  //!! layer_ub[layer]->setValue(get_CP(), ub);
	//}

	// if( state_changed ) {

	// 	RevMDD* mdd = get_mdd();

	// 	// require ancestor update
	// 	if( layer + length < mdd->get_num_layers() ) {
	// 		for( int w = 0; w < mdd->get_max_width(); w++ ) {
	// 			modified_in_node(layer+length, w);
	// 		}
	// 	}

	// 	// require descendant update
	// 	if( layer - length >= 0 ) {
	// 		for( int w = 0; w < mdd->get_max_width(); w++ ) {
	// 			modified_out_node(layer-length, w);
	// 		}
	// 	}
	// }
}






