//============================================================================
// Sequence MDD Propagator - Header File
//============================================================================


#ifndef SEQUENCE_MDD_HPP_
#define SEQUENCE_MDD_HPP_

#include <cassert>
#include <ilcp/cpext.h>
#include <vector>
#include <list>

#include "../core/mdd_propagator.hpp"
#include "../util/util.hpp"
#include "../util/stats.hpp"

using namespace std;


/**
 * Sequence state
 */
struct SequenceState {

  IlcRevInt		y_lb;			/**< cumulative lower bound */
  IlcRevInt		y_ub;			/**< cumulative upper bound */

  IlcRevInt**		ancestor_lb;	/**< cumulative lower bound */
  IlcRevInt**		ancestor_ub;	/**< cumulative upper bound */

  IlcRevInt**		descendant_lb;	/**< cumulative lower bound */
  IlcRevInt**		descendant_ub;	/**< cumulative upper bound */

  int				initial_y_lb;	/**< store initial values */
  int				initial_y_ub;	/**< store initial values */

  /** Constructor */
  SequenceState(IloCP cp,  int length, int _initial_y_lb, int _initial_y_ub);

  /** Reset state */
  void reset(IloCP cp);
};


/**
 * Relaxed stated information for a potential node to be created in a layer.
 */

struct SequenceNodeSketch {
  int       		y_lb;	 		/**< lower bound of cumulative variable */
  int			    y_ub;    		/**< upper bound of cumulative variable */

  int				timestamp;		/**< last time layer was modified */
  bool			feasible;		/**< if sketch corresponds to a feasible arc */
  int				n_types;		/**< arcs of this type in layer */

  bool			verified;		/**< if sketch was already verified */
};


/**
 * Sketch status
 */
enum SketchStatus {
  NotInitialized, Feasible, Infeasible
};


/**
 * Sequence Propagator
 */
class SequenceMDDPropagator : public MDDPropagator {

public:


  /** Propagation type */
  enum PropagationType {
    WeakProp, StrongProp
  };

  /** Constructor */
  SequenceMDDPropagator(
			IloIntVarArray			&_vars,				/**< variables */
			int						_length,			/**< length of the sequence */
			int						_lb,     			/**< lower bound */
			int               		_ub,      			       /**< upper bound */
			vector<int>				&_values,			/**< values in the sequence set */
			PropagationType                    _prop_type = StrongProp,           /**< propagation type */
			bool                               _must_refine = true                     /**< if must refine */
			);

  /** Destructor */
  ~SequenceMDDPropagator() { }

  /*
   * -------------------------------------
   * Propagator Base Methods
   * -------------------------------------
   */

  /** Initialize constraint */
  void initialize();

  /** Initialize top-down pass */
  void initialize_topdown();

  /** Initialize bottom-up pass */
  void initialize_bottomup();

  /** Filter layer in a top-down perspective */
  void filter_layer_topdown(int layer);

  /** Filter layer in a bottom up perspective */
  void filter_layer_bottomup(int layer);

  /** Partition refinement group (refinement). Max MDD width must be observed */
  void partition_refinement_group(int layer, RefinementGroup* ref_groups, int &n_groups);

  /** Setup the state of a node according to its incoming arcs. Return if node was modified. */
  void setup_node_incoming (int layer, int node, RefinementGroup ref_group, bool new_node);

  /** Called after a new node is created with its outgoing arcs. Return if node was modified. */
  void setup_node_outgoing(int layer, int node, RefinementGroup ref_group, bool new_node);

  /** If constraint should be processed in a particular layer */
  bool in_layer(int layer);

  /** Print node states */
  void print_states();

  /** Update layer information */
  void update_layer(int layer, vector<int> &layer_nodes);

  /** Stats */
  Stats stats;


private:

  /** Check if an out arc is not necessarily infeasible **/
  bool verify_incoming_arc(int layer, RevMDD::Arc* arc);

  /** Check if an out arc is not necessarily infeasible **/
  bool verify_outgoing_arc(int layer, RevMDD::Arc* arc);

  /** Obtain node state */
  SequenceState* get_node_state(int layer, int index);

  /** Reserve node state/pre-processed memory */
  void reserve_memory();

  /** Check if group has distinct LB (assume elements are sorted) */
  bool group_distinct_lb(RefinementGroup ref_group);

  /** Check if group has distinct UB (assume elements are sorted) */
  bool group_distinct_ub(RefinementGroup ref_group);

  /** Get arc LB according to node sketch */
  int get_sketch_lb(RevMDD::Arc* arc);

  /** Get arc UB according to node sketch */
  int get_sketch_ub(RevMDD::Arc* arc);

  /** Update sketch of incoming arc */
  SequenceNodeSketch* update_sketch_incoming(int layer, RevMDD::Arc* arc);

  /** Update sketch of incoming arc; considers new nodes */
  SequenceNodeSketch* update_sketch_incoming(int layer, RevMDD::Arc* arc, bool new_node);

  /** Update sketch of outgoing arc */
  SequenceNodeSketch* update_sketch_outgoing(int layer, RevMDD::Arc* arc);


  /*
   * -------------------------------------
   * Sequence attributes
   * -------------------------------------
   */

  IloIntVarArray ilo_vars;   /**< model variables */
  IlcIntVarArray vars;	     /**< problem variables */
  int length;		     /**< length of the sequence */
  int lb;     		     /**< lower bound */
  int ub;      		     /**< upper bound */
  vector<int> values;	     /**< values in the sequence set */
  bool must_refine;          /**< if must refine */               

  int*	layer_lb;  /**< layer LB of cumulative variable */
  int*	layer_ub;  /**< layer UB of cumulative variable */
  

  /*
   * -------------------------------------
   * Propagation attributes
   * -------------------------------------
   */

  PropagationType                                       prop_type;               /**< propagation type */

  bool*						in_set;			 /**< if value is in sequence set */
  int*						in_set_val;		 /**< 1 if value is in sequence set, 0 otherwise	 */

  SequenceState***     		states;          /**< states per layer and node */

  SequenceNodeSketch****		sketches;		 /**< node sketches per source/target nodes and type (in set / out set) */
  SketchStatus***				sketch_init;	 /**< if sketch was initialized */

  bool*						node_verified;	 /**< if node was already verified */

  SimpleList<RevMDD::Arc*>*	arcs_to_move;	 /**< arcs to move from groups (refinement) */
  bool***						has_in_arc;		 /**< if has arc of a given type */

  SimpleList<int>*			available_nodes; /**< available nodes for refinement */

  int*						group_lb;		 /**< group lower bounds */
  int*						group_ub;		 /**< group lower bounds */

  SequenceNodeSketch**		arc_sketch;		 /**< sketch referring to arc */

  int*						aux_lb;			 /**< auxiliary to compute node state bounds */
  int*						aux_ub;			 /**< auxiliary to compute node state bounds */

  bool*						state_modified;	 /**< if node state was modified */
};


/**
 * --------------------------------------------
 * Comparators
 * --------------------------------------------
 */
struct Arcs_LB_Descending {

  SequenceNodeSketch**** sketch;
  int*				  in_set_val;

  Arcs_LB_Descending(SequenceNodeSketch**** _sketch, int* _in_set_val)
    : sketch(_sketch), in_set_val(_in_set_val) { }

  inline bool operator()(RevMDD::Arc* arcA, RevMDD::Arc* arcB) {
    if( sketch[arcA->get_source()][arcA->get_target()][in_set_val[arcA->get_val()]]->y_lb ==
	sketch[arcB->get_source()][arcB->get_target()][in_set_val[arcB->get_val()]]->y_lb )
      {
	if (sketch[arcA->get_source()][arcA->get_target()][in_set_val[arcA->get_val()]]->y_ub ==
	    sketch[arcB->get_source()][arcB->get_target()][in_set_val[arcB->get_val()]]->y_ub) 
	  {
	    return arcA->get_layer_id() < arcB->get_layer_id();
	  }

	return sketch[arcA->get_source()][arcA->get_target()][in_set_val[arcA->get_val()]]->y_ub <
	  sketch[arcB->get_source()][arcB->get_target()][in_set_val[arcB->get_val()]]->y_ub;
      }
    return sketch[arcA->get_source()][arcA->get_target()][in_set_val[arcA->get_val()]]->y_lb >
      sketch[arcB->get_source()][arcB->get_target()][in_set_val[arcB->get_val()]]->y_lb;
  }
};


struct RefGroup_LB_Comparator {

  int* lb;	// group lower bounds
  int* ub;	// group upper bounds

  RefGroup_LB_Comparator(int *_lb, int* _ub) : lb(_lb), ub(_ub) { }

  inline bool operator()(int gA, int gB) {
    if( lb[gA] == lb[gB] ) {
      if (ub[gA] == ub[gB]) {
	return gA < gB;
      }
      return ub[gA] > ub[gB];
    }
    return lb[gA] < lb[gB];
  }
};


/**
 * --------------------------------------------
 * Inline implementations
 * --------------------------------------------
 */

/**
 * Sequence Propagator Constructor
 */
inline SequenceMDDPropagator::SequenceMDDPropagator(
						    IloIntVarArray			&_vars,				/**< variables */
						    int						_length,			/**< length of the sequence */
						    int						_lb,     			/**< lower bound */
						    int               		_ub,      			/**< upper bound */
						    vector<int>				&_values,			/**< values in the sequence set */
						    PropagationType                      _prop_type,
						    bool _must_refine       /**< propagation type */
						    )
  : ilo_vars(_vars), length(_length), lb(_lb), ub(_ub), values(_values), prop_type(_prop_type), must_refine(_must_refine)
{
  num_state_updates = 0;
  num_changes_states = 0;

  if (prop_type == WeakProp) {
    cout << "[SequenceMDD] Using weak filtering - refine = " << must_refine << endl;	  
  } else if (prop_type == StrongProp) {
    cout << "[SequenceMDD] Using strong filtering - refine = " << must_refine << endl;
  }
	
}


/**
 * Sequence State Constructor
 **/
inline SequenceState::SequenceState(IloCP cp, int length, int _initial_y_lb, int _initial_y_ub)
  : initial_y_lb(_initial_y_lb), initial_y_ub(_initial_y_ub)
{
  // initialize ancestors / descendants
  ancestor_lb = new( cp.getHeap() ) IlcRevInt*[length];
  ancestor_ub = new( cp.getHeap() ) IlcRevInt*[length];

  descendant_lb = new( cp.getHeap() ) IlcRevInt*[length];
  descendant_ub = new( cp.getHeap() ) IlcRevInt*[length];

  // TODO: only perform this loop if strong propagation
  for( int i = 1; i < length; i++ ) {
    ancestor_lb[i] = new( cp.getHeap() ) IlcRevInt(cp, 0);
    ancestor_ub[i] = new( cp.getHeap() ) IlcRevInt(cp, length);

    descendant_lb[i] = new( cp.getHeap() ) IlcRevInt(cp, 0);
    descendant_ub[i] = new( cp.getHeap() ) IlcRevInt(cp, length);
  }

  // the first ancestor is the own node, and both ancestor/descendants
  // have the same value
  ancestor_lb[0] = &y_lb;
  ancestor_ub[0] = &y_ub;

  descendant_lb[0] = &y_lb;
  descendant_ub[0] = &y_ub;

  // reset values
  reset(cp);
}

/** Sequence State reset */
inline void SequenceState::reset(IloCP cp) {
  y_lb.setValue(cp, initial_y_lb);
  y_ub.setValue(cp, initial_y_ub);
}


/**
 * Initialize constraint
 */
inline void SequenceMDDPropagator::initialize() {

  IloCP cp = get_CP();

  // extract variables
  vars = cp.getIntVarArray(ilo_vars);

  // reserve pre-processing/node state memory
  reserve_memory();
}


/**
 * Obtain node state
 */
inline SequenceState* SequenceMDDPropagator::get_node_state(int layer, int index) {
  assert( layer >= 0 && layer < get_mdd()->get_num_layers() );
  assert( index >= 0 && index < get_mdd()->get_max_width() );
  return states[layer][index];
}



/**
 * Update sketch of incoming arc
 **/
inline SequenceNodeSketch* SequenceMDDPropagator::update_sketch_incoming(int layer, RevMDD::Arc* arc) {

  // cout << "Not updated" << endl;
  // exit(1);

  // get sketch
  SequenceNodeSketch* skt = sketches[arc->get_source()][arc->get_target()][in_set_val[arc->get_val()]];

  // 	// check if sketch has already been computed according to timestamp
  // 	if( skt->timestamp == get_timestamp() ) {

  // 		// associate arc with this sketch
  // 		arc_sketch[arc->get_layer_id()] = skt;

  // 		// return sketch
  // 		return skt;
  // 	}

  // //	cout << "\t[MDD-Sequence] Verifying in arc (" << layer << "," << arc->get_source() << "," << arc->get_val() << ")";
  // //	cout << " - prop: " << prop_id;
  // //	cout << endl;

  // 	// update sketch status
  // 	skt->timestamp = get_timestamp();


  // compute sketch bounds

  SequenceState* target_state = get_node_state(layer+1,arc->get_target());
  skt->y_lb = target_state->y_lb.getValue();
  skt->y_ub = target_state->y_ub.getValue();


  SequenceState* source_state = get_node_state(layer,arc->get_source());

  if( layer+1 >= length ) {
    if (prop_type == WeakProp) {
      skt->y_lb = MAX(skt->y_lb, lb + layer_lb[layer+1-length]);
      skt->y_ub = MIN(skt->y_ub, ub + layer_ub[layer+1-length]);

    } else if (prop_type == StrongProp) {
      skt->y_lb = MAX(skt->y_lb, lb + source_state->ancestor_lb[length-1]->getValue());
      skt->y_ub = MIN(skt->y_ub, ub + source_state->ancestor_ub[length-1]->getValue());
    }
  }

  // check if bounds are feasible
  int source_lb = source_state->y_lb.getValue() + in_set_val[arc->get_val()];
  int source_ub = source_state->y_ub.getValue() + in_set_val[arc->get_val()];

  if( skt->y_lb > skt->y_ub || source_lb > skt->y_ub || source_ub < skt->y_lb )	{

    //		cout << "\tarc infeasible: lb: " << skt->y_lb << " - ub: " << skt->y_ub;
    //		cout << " - source_lb: " << source_lb;
    //		cout << " - source_ub: " << source_ub;
    //		cout << endl << endl;

    skt->feasible = false;
    return skt;
  }

  // update bounds
  skt->y_lb = MAX(skt->y_lb, source_lb);
  skt->y_ub = MIN(skt->y_ub, source_ub);

  assert( skt->y_lb <= skt->y_ub );
  //cout << "\tsketch state: LB=" << skt->y_lb << " - UB=" << skt->y_ub << endl << endl;


  // set sketch as feasible
  skt->feasible = true;

  // associate arc with this sketch
  arc_sketch[arc->get_layer_id()] = skt;

  return skt;
}


/**
 * Update sketch of incoming arc - considers new node
 **/
inline SequenceNodeSketch* SequenceMDDPropagator::update_sketch_incoming(int layer, RevMDD::Arc* arc, bool new_node) {

  // get sketch
  SequenceNodeSketch* skt = sketches[arc->get_source()][arc->get_target()][in_set_val[arc->get_val()]];

  // 	// check if sketch has already been computed according to timestamp
  // 	if( skt->timestamp == get_timestamp() ) {

  // 		// associate arc with this sketch
  // 		arc_sketch[arc->get_layer_id()] = skt;

  // 		// return sketch
  // 		return skt;
  // 	}

  // //	cout << "\t[MDD-Sequence] Verifying in arc (" << layer << "," << arc->get_source() << "," << arc->get_val() << ")";
  // //	cout << " - prop: " << prop_id;
  // //	cout << endl;

  // 	// update sketch status
  // 	skt->timestamp = get_timestamp();


  // compute sketch bounds
  if( !new_node ) {
    SequenceState* target_state = get_node_state(layer+1,arc->get_target());
    skt->y_lb = target_state->y_lb.getValue();
    skt->y_ub = target_state->y_ub.getValue();
  } else {
    skt->y_lb = -1;
    skt->y_ub = INF;
  }


  SequenceState* source_state = get_node_state(layer,arc->get_source());

  if( layer+1 >= length ) {
    if (prop_type == WeakProp) {
      skt->y_lb = MAX(skt->y_lb, lb + layer_lb[layer+1-length]);
      skt->y_ub = MIN(skt->y_ub, ub + layer_ub[layer+1-length]);
	  
    } else if (prop_type == StrongProp) {
      skt->y_lb = MAX(skt->y_lb, lb + source_state->ancestor_lb[length-1]->getValue());
      skt->y_ub = MIN(skt->y_ub, ub + source_state->ancestor_ub[length-1]->getValue());
    }

    //int lb2 = MAX(skt->y_lb, lb + source_state->ancestor_lb[length-1]->getValue());
    //  assert(lb2 >= skt->y_lb);
	    
    //  int ub2 = MIN(skt->y_ub, ub + source_state->ancestor_ub[length-1]->getValue());
    //  assert(ub2 <= skt->y_ub);
  }

  // check if bounds are feasible
  int source_lb = source_state->y_lb.getValue() + in_set_val[arc->get_val()];
  int source_ub = source_state->y_ub.getValue() + in_set_val[arc->get_val()];

  if( skt->y_lb > skt->y_ub || source_lb > skt->y_ub	|| source_ub < skt->y_lb )	{

    //		cout << "\tarc infeasible: lb: " << skt->y_lb << " - ub: " << skt->y_ub;
    //		cout << " - source_lb: " << source_lb;
    //		cout << " - source_ub: " << source_ub;
    //		cout << endl << endl;

    skt->feasible = false;
    return skt;
  }

  // update bounds
  skt->y_lb = MAX(skt->y_lb, source_lb);
  skt->y_ub = MIN(skt->y_ub, source_ub);

  assert( skt->y_lb <= skt->y_ub );
  //cout << "\tsketch state: LB=" << skt->y_lb << " - UB=" << skt->y_ub << endl << endl;


  // set sketch as feasible
  skt->feasible = true;

  // associate arc with this sketch
  arc_sketch[arc->get_layer_id()] = skt;

  return skt;
}



/**
 * Update sketch of outgoing arc
 **/
inline SequenceNodeSketch* SequenceMDDPropagator::update_sketch_outgoing(int layer, RevMDD::Arc* arc) {

  // get sketch
  SequenceNodeSketch* skt = sketches[arc->get_source()][arc->get_target()][in_set_val[arc->get_val()]];

  // // check if sketch has already been verified according to time stamp
  // if( skt->timestamp == get_timestamp() ) {

  // 	// associate arc with this sketch
  // 	arc_sketch[arc->get_layer_id()] = skt;

  // 	// return sketch
  // 	return skt;
  // }

  // //cout << "\t[MDD-Sequence] Verifying out arc (" << layer << "," << arc->get_source() << "," << arc->get_val() << ")" << endl;

  // // update sketch status
  // skt->timestamp = get_timestamp();

  skt->n_types = 1;


  // compute sketch bounds

  SequenceState* source_state = get_node_state(layer, arc->get_source());

  skt->y_lb = source_state->y_lb.getValue();
  skt->y_ub = source_state->y_ub.getValue();

  SequenceState* target_state = get_node_state(layer+1,arc->get_target());

  int pos = layer + length;

  if( pos < get_mdd()->get_num_layers() ) {

    if (prop_type == WeakProp) {
      skt->y_lb = MAX(skt->y_lb, layer_lb[pos] - ub);
      skt->y_ub = MIN(skt->y_ub, layer_ub[pos] - lb);
	  
    } else if (prop_type == StrongProp) {
      skt->y_lb = MAX(skt->y_lb, target_state->descendant_lb[length-1]->getValue() - ub);
      skt->y_ub = MIN(skt->y_ub, target_state->descendant_ub[length-1]->getValue() - lb);
    }

    //int lb = MAX(skt->y_lb, target_state->descendant_lb[length-1]->getValue() - ub);
    //assert(lb >= skt->y_lb);

    //int ub = MIN(skt->y_ub, target_state->descendant_ub[length-1]->getValue() - lb);
    //assert(ub <= skt->y_ub);
  }

  // check if bounds are feasible
  int target_lb = target_state->y_lb.getValue() - in_set_val[arc->get_val()];
  int target_ub = target_state->y_ub.getValue() - in_set_val[arc->get_val()];

  if( skt->y_lb > skt->y_ub || target_lb > skt->y_ub || target_ub < skt->y_lb ) {
    skt->feasible = false;
    return skt;
  }

  // update bounds
  skt->y_lb = MAX(skt->y_lb, target_lb);
  skt->y_ub = MIN(skt->y_ub, target_ub);

  assert( skt->y_lb <= skt->y_ub );
  //cout << "\tsketch state: LB=" << skt->y_lb << " - UB=" << skt->y_ub << endl << endl;

  skt->feasible = true;
  arc_sketch[arc->get_layer_id()] = skt;

  return skt;
}




/**
 * Check if an out arc is not necessarily infeasible
 *
 **/
inline bool SequenceMDDPropagator::verify_incoming_arc(int layer, RevMDD::Arc* arc) {
  return ( update_sketch_incoming(layer, arc)->feasible );
}


/**
 * Check if an in arc is not necessarily infeasible
 **/
inline bool SequenceMDDPropagator::verify_outgoing_arc(int layer, RevMDD::Arc* arc) {
  return ( update_sketch_outgoing(layer, arc)->feasible );
}


/**
 * Initialize top-down pass
 **/
inline void SequenceMDDPropagator::initialize_topdown() {
  layer_lb[0] = 0;
  layer_ub[0] = 0;
  for (int l = 1; l < get_mdd()->get_num_layers(); ++l) {
    layer_lb[l] = INF;
    layer_ub[l] = -1;
  }

  //	// reset ancestors of root node
  //	SequenceState* state = get_node_state(0,0);
  //	memset(state->ancestor_lb, 0, sizeof(int)*length);
  //	memset(state->ancestor_ub, 0, sizeof(int)*length);

  // initialize node sketch timestamp
  //	for( int w = 0; w < get_mdd()->get_max_width(); w++ ) {
  //		for( int u = 0; u < get_mdd()->get_max_width(); u++ ) {
  //			sketches[w][u][0]->timestamp = -1;
  //			sketches[w][u][1]->timestamp = -1;
  //		}
  //	}

  // root node state is initially not modified
  //memset( state_modified, false, sizeof(bool)*get_mdd()->get_max_width() );
}

/**
 * Initialize bottom-up pass
 **/
inline void SequenceMDDPropagator::initialize_bottomup() {
  // layer_lb[get_mdd()->get_num_layers()-1] = 0;
  // layer_ub[get_mdd()->get_num_layers()-1] = 0;
  // for (int l = get_mdd()->get_num_layers()-2; l >= 0; --l) {
  //   layer_lb[l] = INF;
  //   layer_ub[l] = -1;
  // }

  // initialize node sketch timestamp
  //	for( int w = 0; w < get_mdd()->get_max_width(); w++ ) {
  //		for( int u = 0; u < get_mdd()->get_max_width(); u++ ) {
  //			sketches[w][u][0]->timestamp = -1;
  //			sketches[w][u][1]->timestamp = -1;
  //		}
  //	}

  // terminal node state set as initially modified, for the sake of consistency
  //memset( state_modified, false, sizeof(bool)*get_mdd()->get_max_width() );
  //state_modified[0] = true;
}

/**
 * Get arc LB according to node sketch
 **/
inline int SequenceMDDPropagator::get_sketch_lb(RevMDD::Arc* arc) {
  assert( arc != NULL );
  assert( arc->get_source() >= 0 && arc->get_source() < get_mdd()->get_max_width() );
  assert( arc->get_target() >= 0 && arc->get_target() < get_mdd()->get_max_width() );

  return arc_sketch[arc->get_layer_id()]->y_lb;
}

/**
 * Get arc UB according to node sketch
 **/
inline int SequenceMDDPropagator::get_sketch_ub(RevMDD::Arc* arc) {
  assert( arc != NULL );
  assert( arc->get_source() >= 0 && arc->get_source() < get_mdd()->get_max_width() );
  assert( arc->get_target() >= 0 && arc->get_target() < get_mdd()->get_max_width() );

  return arc_sketch[arc->get_layer_id()]->y_ub;
}


/**
 * Check if group has distinct LB (assume elements are sorted)
 **/
inline bool SequenceMDDPropagator::group_distinct_lb(RefinementGroup ref_group) {
  return( get_sketch_lb(ref_group->get_first()) != get_sketch_lb(ref_group->get_last()) );
}

/**
 * Check if group has distinct UB (assume elements are sorted)
 **/
inline bool SequenceMDDPropagator::group_distinct_ub(RefinementGroup ref_group) {
  return( get_sketch_ub(ref_group->get_first()) != get_sketch_ub(ref_group->get_last()) );
}

/**
 * If a constraint should be processed in a particular layer
 */
inline bool SequenceMDDPropagator::in_layer(int layer) {
  return true;
}




#endif // SEQUENCE_MDD_HPP

