// ============================================================================
// ILOG Search Goals for Scheduling
// ============================================================================

#include <ilcp/cpext.h>
#include <cstring>

#include "search_goals.hpp"


using namespace std;

/**
 * Set times from permutation variables
 */
ILCGOAL2(SetTimesPermutation, IlcIntVarArray, permutation, IlcIntervalVarArray, acts) {
  for( int i = 0; i < permutation.getSize(); ++i ) {
    acts[permutation[i].getValue()].setStart(acts[permutation[i].getValue()].getStartMin());
  }
  return IlcGoalTrue(getCP());
}

/**
 * Wrapper for set times permutation
 **/
ILOCPGOALWRAPPER2(IloSetTimesPermutation, cp,
		  IloIntVarArray, _permutation,
		  IloIntervalVarArray, _tasks)
{
  IlcIntervalVarArray tasks(cp, _tasks.getSize());
  for( int i = 0; i < _tasks.getSize(); i++ )
    tasks[i] = cp.getInterval(_tasks[i]);
  return SetTimesPermutation(cp, cp.getIntVarArray(_permutation), tasks);
}


/**
 * Fix variable to its minimum value
 */
ILCGOAL1(FixIntVarMin, IlcIntVar, var) {
  var.setValue(var.getMin());
  return IlcGoalTrue(getCP());
}

/**
 * Wrapper to fix variable to its minimum value
 **/
ILOCPGOALWRAPPER1(IloFixIntVarMin, cp, IloIntVar, var)
{
  return FixIntVarMin(cp, cp.getIntVar(var));
}


/**
 * Set times of a sequence var
 */
ILCGOAL1(SetTimesSequence, IlcIntervalSequenceVar, seq) {
  // fix start times
  for( IlcIntervalVar act = seq.getEarliestInHead(); act.getImpl() != NULL; act = seq.getOneLaterInHead(act) ) {
    act.setStart(act.getStartMin());
  }
  for( IlcIntervalVar act = seq.getEarliestInTail(); act.getImpl() != NULL; act = seq.getOneLaterInTail(act) ) {
    act.setStart(act.getStartMin());
  }
  return IlcGoalTrue(getCP());
}

/**
 * Wrapper for set times sequence
 **/
ILOCPGOALWRAPPER1(IloSetTimesSequence, cp, IloIntervalSequenceVar, _seq)
{
  return SetTimesSequence(cp, cp.getIntervalSequence(_seq));
}


/**
 * Set precedences for SOP problems
 */
ILCGOAL3(SetPrecedences, IlcIntervalVarArray, tasks, IlcIntervalSequenceVar, seq, bool**, precedes) {
  for( int i = 0; i < tasks.getSize(); i++ ) {
    for( int j = 0; j < tasks.getSize(); j++ ) {
      if( precedes[i][j] ) {
	seq.setBefore(tasks[i], tasks[j]);
      }
    }
  }
  return IlcGoalTrue(getCP());
}

/**
 * Wrapper for SOP precedences
 */
ILOCPGOALWRAPPER3(IloSetPrecedences, cp, IloIntervalVarArray, _tasks, IloIntervalSequenceVar, _seq, bool**, precedes)
{
  IlcIntervalVarArray tasks(cp, _tasks.getSize());
  for( int i = 0; i < _tasks.getSize(); i++ )
    tasks[i] = cp.getInterval(_tasks[i]);
  return SetPrecedences(cp, tasks, cp.getIntervalSequence(_seq), precedes);
}


/**
 * Implementation goal for lexicographic search
 */
ILCGOAL2(LexSchedule, IlcIntervalVarArray, tasks, IlcIntervalSequenceVar, seq) {
  IloCP cp = getCP();
  for( int i = 0; i < tasks.getSize(); i++ ) {
    if( seq.isCandidateHead(tasks[i]) ) {
      return IlcAnd(seq.tryExtendHead(tasks[i]), LexSchedule(cp, tasks, seq));
    }
  }
  return IlcGoalTrue(cp);
}


/**
 * Set precedences for SOP problems
 */
ILCGOAL1(SetMinStartTime, IlcIntervalVarArray, tasks) {
  for( int i = 0; i < tasks.getSize(); i++ ) {
    tasks[i].setStartMax(tasks[i].getStartMin());
  }
  return IlcGoalTrue(getCP());
}

/**
 * Wrapper for set min start times
 */
ILOCPGOALWRAPPER1(IloSetMinStartTime, cp, IloIntervalVarArray, _tasks)
{
  IlcIntervalVarArray tasks(cp, _tasks.getSize());
  for( int i = 0; i < _tasks.getSize(); i++ )
    tasks[i] = cp.getInterval(_tasks[i]);
  return SetMinStartTime(cp, tasks);
}


/**
 * Wrapper for lexicographic search
 */
ILOCPGOALWRAPPER2(IloLexSchedule, cp, IloIntervalVarArray, _tasks, IloIntervalSequenceVar, _seq)
{
  IlcIntervalVarArray tasks(cp, _tasks.getSize());
  for( int i = 0; i < _tasks.getSize(); i++ )
    tasks[i] = cp.getInterval(_tasks[i]);
  return LexSchedule(cp, tasks, cp.getIntervalSequence(_seq));
}


/**
 * Goal for testing
 */
ILCGOAL1(IlcTest, char*, message) {
  cout << message << endl;
  return IlcGoalTrue(getCP());
}


/**
 * Sequencing based on activity belonging to shortest path in the MDD
 * TODO: little optimization: also save last fixed position
 */
ILCGOAL2(ShortestPathMDD, IlcIntVarArray, permutation, DisjunctivePropagator*, disj_propagator) {

  // search first unfixed activity in M
  int unfixed = -1;
  for( int i = 0; i < permutation.getSize() && unfixed == -1; ++i ) {
    if( !permutation[i].isFixed() ) {
      unfixed = i;
    }
  }
  if( unfixed == -1 ) {
    // activities are fixed: return
    return IlcGoalTrue(getCP());
  }

  // get MDD shortest path
  IloCP cp = getCP();
  int* optimum = disj_propagator->compute_shortest_path();
  return( IlcOr(IlcAnd(permutation[unfixed] == optimum[unfixed], ShortestPathMDD(cp, permutation, disj_propagator)),
                IlcAnd(permutation[unfixed] != optimum[unfixed], ShortestPathMDD(cp, permutation, disj_propagator))) );
}


/**
 * Wrapper: Sequencing based on activity belonging to shortest path in the MDD
 */
ILOCPGOALWRAPPER2(IloShortestPathMDD, cp, IloIntVarArray, permutation, DisjunctivePropagator*, disj_propagator) {
  return ShortestPathMDD(cp, cp.getIntVarArray(permutation), disj_propagator);
}





/**
 * Wrapper: Sequencing based on relaxed sum of setup times in MDD
 */
ILCGOAL2(SumSetupTimesMDD, IlcIntVarArray, permutation, DisjunctivePropagator*, disj_propagator) {

  IloCP cp = getCP();

  int pos, act;
  disj_propagator->get_activity_min_sum_setup_T(pos, act);
  if( pos == -1 ) {
    return IlcGoalTrue(cp);
  }

  //cout << "Trying " << pos << " --> " << act << endl;

  return( IlcOr(IlcAnd(permutation[pos] == act, SumSetupTimesMDD(cp, permutation, disj_propagator)),
		IlcAnd(permutation[pos] != act, SumSetupTimesMDD(cp, permutation, disj_propagator))) );
}



/**
 * Wrapper: Sequencing based on relaxed sum of setup times in MDD
 */
ILOCPGOALWRAPPER2(IloSumSetupTimesMDD, cp, IloIntVarArray, permutation, DisjunctivePropagator*, disj_propagator) {
  return SumSetupTimesMDD(cp, cp.getIntVarArray(permutation), disj_propagator);
}



/**
 * Fix node of the MDD, removing all other nodes of the corresponding layer
 */
ILCGOAL3(FixMDDNode, DisjunctivePropagator*, disj_propagator, int, layer, int, node) {
  IloCP cp = getCP();
  disj_propagator->fix_node(layer, node);
  return IlcGoalTrue(cp);
}

/**
 * Fix node of the MDD, removing all other nodes of the corresponding layer
 */
ILCGOAL3(RemoveMDDNode, DisjunctivePropagator*, disj_propagator, int, layer, int, node) {
  IloCP cp = getCP();
  disj_propagator->remove_node(layer, node);
  return IlcGoalTrue(cp);
}


/**
 * Wrapper: Sequencing based on relaxed sum of setup times in MDD
 */
ILCGOAL2(MDDNodeBranch, IlcIntVarArray, permutation, DisjunctivePropagator*, disj_propagator) {

  IloCP cp = getCP();

  int layer, node;
  disj_propagator->get_node_min_sum_setup_T(layer, node);
  if( layer == -1 ) {
    return IlcGoalTrue(cp);
  }

  //cout << "Trying " << pos << " --> " << act << endl;

  return( IlcOr(IlcAnd(FixMDDNode(cp, disj_propagator, layer, node), MDDNodeBranch(cp, permutation, disj_propagator)),
		IlcAnd(RemoveMDDNode(cp, disj_propagator, layer, node), MDDNodeBranch(cp, permutation, disj_propagator))) );
}



/**
 * Wrapper: Sequencing based on relaxed sum of setup times in MDD
 */
ILOCPGOALWRAPPER2(IloMDDNodeBranch, cp, IloIntVarArray, permutation, DisjunctivePropagator*, disj_propagator) {
  return MDDNodeBranch(cp, cp.getIntVarArray(permutation), disj_propagator);
}




/**
 * Goal for lexicographic search
 */
ILOCPGOALWRAPPER1(IloTest, cp, char*, message) {
  char* new_message = new(cp.getHeap()) char[256];
  strcpy(new_message, message);
  return IlcTest(cp, new_message);
}


// -----------------------------------------------------------
// Begin Node Branching
// -----------------------------------------------------------

/**
 * Select an arbitrary node from the first non-exact layer
 * (Obs.: This selector is deterministic)
 */
pair<int, int>
SelectFromNonExactLayer::select_layer_node(
					   IloCP cp,
					   RevMDD* mdd,
					   DisjunctivePropagator* disj_propagator)
{
  int layer = -1;
  for (int l = 1; l < mdd->get_num_layers()-1; ++l) {
    if (!disj_propagator->is_layer_exact(l)) {
      layer = l;
      break;
    }
  }
  assert(layer != -1);

  // take node of the first arc in the layer
  RevMDD::Arc* first_arc = mdd->arcs_layer[layer].get_first();
  int node = first_arc->get_source();

  //	int node = disj_propagator->get_node_min_sum_setup_T(layer);
  return pair<int,int>(layer, node);
}


/**
 * Select node from the first non-exact layer
 * with min sum setup time to terminal node
 */
pair<int, int>
SelectMinSumSetupTNonExactLayer::select_layer_node(
						   IloCP cp,
						   RevMDD* mdd,
						   DisjunctivePropagator* disj_propagator)
{
  int layer = -1;
  for (int l = 1; l < mdd->get_num_layers()-1; ++l) {
    if (!disj_propagator->is_layer_exact(l)) {
      layer = l;
      break;
    }
  }
  assert(layer != -1);

  int node = disj_propagator->get_node_min_sum_setup_T(layer);
  return pair<int,int>(layer, node);
}

/**
 * Select node in the first inexact layer of the shortest path
 */
pair<int, int>
SelectFirstInexactNodeShortestPath::select_layer_node(
						   IloCP cp,
						   RevMDD* mdd,
						   DisjunctivePropagator* disj)
{
  // compute shortest path
  double sp_value;
  RevMDD::Arc** sp_arcs;
  int* opt = disj->compute_shortest_path((*cost), sp_value, sp_arcs);

  //cout << "SEARCH COST = " << sp_value << endl;

  // get first node of the SP that belongs to an inexact layer  
  int layer=-1, node=-1;
  for (int l = 0; l < mdd->get_num_layers(); ++l) {
    if (!disj->is_layer_exact(l)) {
      layer = l;
      node = sp_arcs[l]->get_source();
      break;
    }
  }

  return pair<int,int>(layer, node);
}



/**
 * Remove all but a particular node from a layer of the MDD
 */
ILCGOAL3(RemoveAllButNode, DisjunctivePropagator*, disj_propagator, int, layer, int, node) {
  IloCP cp = getCP();
  RevMDD* mdd = disj_propagator->get_mdd();

  //cout << "[BranchOnNodes] \tRemoving all BUT node in layer=" << layer << " - node=" << node << " - counter=" << counter << endl;

  //	char filename[256];
  //	sprintf(filename, "graphs/search_0_%d_before.gml", counter);
  //	mdd->export_to_gml(filename);


  // remove all arcs that leave nodes that are different than the provided one
  RevMDD::Arc* previous = mdd->arcs_layer[layer].get_header();
  for (RevMDD::Arc* arc = mdd->arcs_layer[layer].get_first();
       arc != NULL;
       arc = mdd->arcs_layer[layer].get_next(arc))
    {
      if (arc->get_source() != node) {
	previous->next.setValue(cp, arc->next.getValue());
      } else {
	previous = arc;
      }
    }


  // clean dangling nodes
  mdd->clean_dangling_nodes();

  //cout << "[BranchOnNodes] \tRemoved!" << endl;

  //	char filename[256];
  //	sprintf(filename, "graphs/search_0_%d_after.gml", counter);
  //	mdd->export_to_gml(filename);

  // force MDD propagation
  disj_propagator->requires_update();

  return IlcGoalTrue(cp);
}

/**
 * Remove an MDD node
 */
ILCGOAL3(RemoveNode, DisjunctivePropagator*, disj_propagator, int, layer, int, node) {
  IloCP cp = getCP();
  RevMDD* mdd = disj_propagator->get_mdd();

  //	cout << "[BranchOnNodes] \tRemoving node in layer=" << layer << " - node=" << node << " - counter=" << counter << endl;

  //	char filename[256];
  //	sprintf(filename, "graphs/search_1_%d_before.gml", counter);
  //	mdd->export_to_gml(filename);

  // remove all arcs that leave the node
  RevMDD::Arc* previous = mdd->arcs_layer[layer].get_header();
  for (RevMDD::Arc* arc = mdd->arcs_layer[layer].get_first();
       arc != NULL;
       arc = mdd->arcs_layer[layer].get_next(arc))
    {
      if (arc->get_source() == node) {
	previous->next.setValue(cp, arc->next.getValue());
      } else {
	previous = arc;
      }
    }

  // if removed all arcs in the layer, that means it was the only node
  if( mdd->arcs_layer[layer].is_empty() ) {
    return IlcGoalFail(cp);
  }

  // clean dangling nodes
  mdd->clean_dangling_nodes();

  //	cout << "[BranchOnNodes] \tRemoved!" << endl;

  //char filename[256];
  //sprintf(filename, "graphs/search_1_%d_after.gml", counter);
  //mdd->export_to_gml(filename);

  // force MDD propagation
  disj_propagator->requires_update();

  return IlcGoalTrue(cp);
}


/**
 * Wrapper: Branch on MDD Nodes
 */
ILCGOAL3(BranchOnMDDNodes, IlcIntVarArray, permutation,
	 DisjunctivePropagator*, disj_propagator,
	 MDDBranchNodeSelector*, node_selector)
{
  IloCP cp = getCP();
  RevMDD* mdd = disj_propagator->get_mdd();

  // check if MDD is exact
  if (disj_propagator->is_mdd_exact()) {

    //cout << "MDD is exact: no more branching needed" << endl;

    // Compute and set shortest path
    int* opt = disj_propagator->get_shortest_path();
    for (int i = 0; i < permutation.getSize(); ++i) {
      cp.add( permutation[i] == opt[i]);
    }

    // set remaining variables
    return IlcSimpleCompletionGoal(cp);
  }

  // // Check if shortest path is feasible
  // int* opt = disj_propagator->get_shortest_path();
  // bool feasible = true;
  // for (int i = 0; i < permutation.getSize() && feasible; ++i) {
  //   for (int j = i+1; j < permutation.getSize() && feasible; ++j ) {
  //     feasible = (opt[i] != opt[j]);
  //   }
  // }
  // // if feasible, set solution
  // if (feasible) {
  //   //cout << "Shortest path is feasible!" << endl;
  //   for (int i = 0; i < permutation.getSize(); ++i) {
  //     cp.add( permutation[i] ==  opt[i]);
  //   }

  //   // set remaining variables
  //   return IlcSimpleCompletionGoal(cp);
  // }

  // select an MDD node to branch on
  pair<int,int> layer_node =
    node_selector->select_layer_node(cp, mdd, disj_propagator);

  int layer = layer_node.first;
  int node = layer_node.second;

  //cout << "[BranchOnNodes] Selected layer=" << layer << " - node=" << node << " - count: " << counter << endl;

  //	char filename[256];
  //	sprintf(filename, "graphs/search_%d.gml", counter++);
  //	mdd->export_to_gml(filename);
  //if (count > 30) exit(1);

  return( IlcOr(
		// first branch: remove all but a particular node from a layer
		IlcAnd(RemoveAllButNode(cp, disj_propagator, layer, node),
		       BranchOnMDDNodes(cp, permutation, disj_propagator, node_selector)),

		// second branch: remove just selected node
		IlcAnd(RemoveNode(cp, disj_propagator, layer, node),
		       BranchOnMDDNodes(cp, permutation, disj_propagator, node_selector))) );
}

/**
 * Wrapper: Branch on MDD Nodes
 */
ILOCPGOALWRAPPER3(IloBranchOnMDDNodes, cp,
		  IloIntVarArray, permutation, DisjunctivePropagator*, disj_propagator,
		  MDDBranchNodeSelector*, node_selector)
{
  return BranchOnMDDNodes(cp,
			  cp.getIntVarArray(permutation),
			  disj_propagator,
			  node_selector);
}




/**
 * Wrapper: Fix next variable according to MDD shortest path
 */
ILCGOAL3(BranchOnMDDShortestPathAct, IlcIntVarArray, permutation,
	 DisjunctivePropagator*, disj_propagator,
	 int, var_index)
{
  IloCP cp = getCP();

  // search for next variable to fix
  while (var_index < permutation.getSize() && !permutation[var_index].isFixed()) {
    ++var_index;
  }

  // if all permutation variables are fixed, just complete remaining set
  if (var_index == permutation.getSize()) {
    return IlcSimpleCompletionGoal(cp);
  }

  // take activity in the MDD shortest path
  int* opt = disj_propagator->compute_shortest_path();
  
  return( IlcOr(
		// first branch: remove all but a particular node from a layer
		IlcAnd(permutation[var_index] == opt[var_index],
		       BranchOnMDDShortestPathAct(cp, permutation, disj_propagator, var_index+1)),

		// second branch: remove just selected node
		IlcAnd(permutation[var_index] != opt[var_index],
		       BranchOnMDDShortestPathAct(cp, permutation, disj_propagator, var_index))
		) );
}

/**
 * Wrapper: Branch on MDD Nodes
 */
ILOCPGOALWRAPPER2(IloBranchOnMDDShortestPathAct, cp,
		  IloIntVarArray, permutation, DisjunctivePropagator*, disj_propagator)
{
  return BranchOnMDDShortestPathAct(cp,
				    cp.getIntVarArray(permutation),
				    disj_propagator,
				    0);
}


// -----------------------------------------------------------
// End Node Branching
// -----------------------------------------------------------



/**
 * Wrapper: Fix next variable according to MDD shortest path
 */
ILCGOAL2(MDDMultipleDisjunctive, IloArray<IlcIntVarArray>, permutations,
	 IloArray<DisjunctivePropagator*>, disj)
{
  IloCP cp = getCP();

  int vehicle = -1;
  int pos = -1;
  int task = -1;
  int increase = INF;

  for (int i = 0; i < permutations.getSize(); ++i) {
    //cout << "Vehicle " << i << endl;
    int prev = 0;
    for (int j = 1; j < permutations[i].getSize(); ++j) {
      if (!permutations[i][j].isFixed()) {

	//cout << "\t" << i << "," << j << ": " << permutations[i][j] << endl;

	for (IlcIntExpIterator iter(permutations[i][j]); iter.ok(); ++iter) {
	  if (disj[i]->setup[prev][*iter] < increase && disj[i]->setup[prev][*iter] > 0) {
	    //cout << "\t" << prev << " --> " << *iter << " --> " << disj[i]->setup[prev][*iter] << endl;
	    increase = disj[i]->setup[prev][*iter];
	    vehicle = i;
	    pos = j;
	    task = *iter;
	  }
	}	
	break;
      }
      prev = permutations[i][j].getValue();
    }
  }

  //cout << "selected " << vehicle << "," << pos << "," << task << endl;

  if (vehicle == -1) {
    // set remaining variables
    return IlcSimpleCompletionGoal(cp);
  }

  
  return IlcOr(
	       IlcAnd(permutations[vehicle][pos] == task, MDDMultipleDisjunctive(cp, permutations, disj)),
	       IlcAnd(permutations[vehicle][pos] != task, MDDMultipleDisjunctive(cp, permutations, disj)));
}


/**
 * Wrapper: Multiple Disjunctive Search Goal
 */
ILOCPGOALWRAPPER2(IloMDDMultipleDisjunctive, cp,
		  IloArray<IloIntVarArray>, permutation, 
		  IloArray<DisjunctivePropagator*>, disj_propagator)
{
  IloArray<IlcIntVarArray> perm(permutation.getEnv(), disj_propagator.getSize());
  for (int i = 0; i < disj_propagator.getSize(); ++i) {
    perm[i] = cp.getIntVarArray(permutation[i]);
  }
  
  return MDDMultipleDisjunctive(cp, perm, disj_propagator);
}

