// ============================================================================
// ILOG Search Goals for Scheduling
// ============================================================================

#ifndef SEARCH_GOALS_HPP_
#define SEARCH_GOALS_HPP_

#include <ilcp/cpext.h>
#include "../constraints/disjunctive.hpp"

/**
 * Lexicographic sequencing of activities
 */
IloGoal IloLexSchedule(IloEnv env, IloIntervalVarArray activities, IloIntervalSequenceVar sequence);

/**
 * Set precedences for SOP problem
 */
IloGoal IloSetPrecedences(IloEnv env, IloIntervalVarArray activities, IloIntervalSequenceVar sequence, bool** precedes);


/**
 * Lexicographic sequencing of activities
 */
IloGoal IloLexSchedule(IloEnv env, IloIntervalVarArray activities, IloIntervalSequenceVar sequence);


/**
 * Set times of a disjunctive resource
 */
IloGoal IloSetTimesSequence(IloEnv env, IloIntervalSequenceVar seq);

/**
 * Set min start times for intervals
 */
IloGoal IloSetMinStartTime(IloEnv env, IloIntervalVarArray intervals);


/**
 * Set times of a disjunctive resource using permutation variables
 */
IloGoal IloSetTimesPermutation(IloEnv env, IloIntVarArray permutation, IloIntervalVarArray acts);


/**
 * Fix variable to its minimum value
 */
IloGoal IloFixIntVarMin(IloEnv env, IloIntVar var);

/**
 * Sequencing based on activity belonging to shortest path in the MDD
 */
IloGoal IloShortestPathMDD(IloEnv env, IloIntVarArray permutation, DisjunctivePropagator* disj_propagator);


/**
 * Sequencing based on sum of setup times
 */
IloGoal IloSumSetupTimesMDD(IloEnv env, IloIntVarArray permutation, DisjunctivePropagator* disj_propagator);


/**
 * Sequencing based on sum of setup times
 */
IloGoal IloMDDNodeBranch(IloEnv env, IloIntVarArray permutation, DisjunctivePropagator* disj_propagator);


/**
 * Sequencing of multiple resources based on MDD
 *
 */
IloGoal IloMDDMultipleDisjunctive(IloEnv env, IloArray<IloIntVarArray> permutation, 
				  IloArray<DisjunctivePropagator*> propagator);




// -----------------------------------------------------------
// Begin Node Branching
// -----------------------------------------------------------

/**
 * Node selector for branching
 */
class MDDBranchNodeSelector {
public:
	virtual ~MDDBranchNodeSelector() { }

	/** Select a pair (layer, node) to branch on */
	virtual pair<int, int>
	select_layer_node(
			IloCP cp,
			RevMDD* mdd,
			DisjunctivePropagator* disj_propagator) = 0;
};


/**
 * Select an arbitrary node from the first non-exact layer
 * (Obs.: This selector is deterministic)
 */
struct SelectFromNonExactLayer : public MDDBranchNodeSelector {
	pair<int, int>
	select_layer_node(IloCP cp, RevMDD* mdd, DisjunctivePropagator* disj_propagator);
};


/**
 * Select node with min sum setup time to terminal node from first non-exact layer
 */
struct SelectMinSumSetupTNonExactLayer : public MDDBranchNodeSelector {
	pair<int, int>
	select_layer_node(IloCP cp, RevMDD* mdd, DisjunctivePropagator* disj_propagator);
};

/**
 * Select node in the first inexact layer of the shortest path
 */
struct SelectFirstInexactNodeShortestPath : public MDDBranchNodeSelector {

  const int n_vars;
  double*** cost;

  pair<int, int>
  select_layer_node(IloCP cp, RevMDD* mdd, DisjunctivePropagator* disj);

  SelectFirstInexactNodeShortestPath(IloEnv env, int _n_vars, double*** _cost) 
    : n_vars(_n_vars), cost(_cost) {  }
};


/**
 * Branch on MDD nodes
 */
IloGoal IloBranchOnMDDNodes(
		IloEnv env,
		IloIntVarArray permutation,
		DisjunctivePropagator* disj_propagator,
		MDDBranchNodeSelector* node_selector);

/**
 * Branch on the activity in the shortest path
 */
IloGoal IloBranchOnMDDShortestPathAct(
				   IloEnv env,
				   IloIntVarArray permutation,
				   DisjunctivePropagator* disj_propagator);

// -----------------------------------------------------------
// End Node Branching
// -----------------------------------------------------------


#endif /* SEARCH_GOALS_HPP_ */
