//============================================================================
// Main TSPTW file
// Read and solve a TSPTW instance
//============================================================================

#include <ilcp/cpext.h>
#include <ilcplex/ilocplex.h>
#include <iostream>
#include <vector>
#include <fstream>
#include <ctime>

#include "core/mdd_handler.hpp"
#include "core/search_goals.hpp"
#include "constraints/disjunctive.hpp"
#include "util/stats.hpp"
#include "util/util.hpp"

#include "instances/tsptw_instances.hpp"


// input types
#define DUMAS_INPUT   0     // dumas tsptw
#define AFG_INPUT     1     // afg tsptw
#define WT_INPUT      2     // weighted total tardiness
#define WTS_INPUT     3     // weighted total tardiness with setup times
#define DISJ_INPUT    4     // simple disjunctive
#define SOP_INPUT     5     // sequential ordering problem
#define TSP_PD_INPUT  6     // TSP with pickup and delivery
#define RANDOM_INPUT  7     // randomly generated instances
#define TD_TSP_INPUT  8     // time-dependent TSP

// objective function types
#define FEASIBILITY 2

// search types
#define LEX_SEARCH 				0
#define DEFAULT_SEARCH 			1
#define MDD_SP_SEARCH 			2
#define MIN_SETUP_T_SEARCH 		3
#define NODEBRANCH_SEARCH 		4
#define DEFAULT_SEARCH_DFS		5
#define MDD_SHORTEST_PATH		6


// parameters
#define TIME_LIMIT 5400


using namespace std;


// ------------------------------------------
// TODO: remove
double** global_cost;
clock_t time_lagrangian = 0;
clock_t time_shortestpath = 0;
clock_t time_lp = 0;

// assignment constraint
vector<double> lambda_perm;	   // multipliers at current iteration
vector<double> gradient_perm;      // gradient
IlcRevFloat** lambda_perm_rev;     // reversible lambda multipliers
double** cost;                     // cost structure
bool performed_lagrangian = false; // if Lagrangian was already performed

double initial_relax = 0;    // initial relaxation
double lag_relax = 0;        // lagrangian relaxation

// ------------------------------------------


//
// Objective Function Propagator for MDDs
//
class ObjectiveFunctionPropagatorMDDI : public IlcConstraintI {


public:
  // Constructor
  ObjectiveFunctionPropagatorMDDI(IloCP cp,
				  IlcIntVarArray _vars,
				  DisjunctivePropagator* _disj,
				  IlcIntVar _objective)

    : IlcConstraintI(cp), vars(_vars), disj(_disj),
      objective(_objective), n_rounds(0), 
      num_acts(_vars.getSize()), num_pos(_vars.getSize())
  {
    // initialize cost structure
    cost = new(cp.getHeap()) double*[num_acts];
    for (int i = 0; i < num_acts; ++i) {
      cost[i] = new(cp.getHeap()) double[num_acts];
    }
    global_cost = cost;

    // initialize Lagrangian properties - permutation
    lambda_perm.resize(num_acts);
    gradient_perm.resize(num_acts);
    lambda_perm_rev = new(cp.getHeap()) IlcRevFloat*[num_acts];
    for (int i = 0; i < num_acts; ++i) {
      lambda_perm_rev[i] = new(cp.getHeap()) IlcRevFloat(cp, 0.0);
    }

    // initialize Lagrangian properties - precedences
    lambda_prec.resize(num_acts);
    gradient_prec.resize(num_acts);
    lambda_prec_rev = new(cp.getHeap()) IlcRevFloat***[num_acts];
    for (int i = 0; i < num_acts; ++i) {
      lambda_prec[i].resize(num_acts);
      gradient_prec[i].resize(num_acts);      
      lambda_prec_rev[i] = new(cp.getHeap()) IlcRevFloat**[num_pos];
      
      for (int j = 0; j < num_acts; ++j) {
	lambda_prec[i][j].resize(num_pos);
	gradient_prec[i][j].resize(num_pos);      
	lambda_prec_rev[i][j] = new(cp.getHeap()) IlcRevFloat*[num_pos];	   
	for (int t = 0; t < num_pos; ++t) {
	  lambda_prec_rev[i][j][t] = new(cp.getHeap()) IlcRevFloat(cp, 0.0);
	}
      }     
    }
  }

  // Post
  void post() {
    for (int i = 0; i < num_acts; ++i) {
      vars[i].whenValue(this);
    }
  }

  // Propagation
  void propagate();

  // Filter MDD using resulting lambdas
  void filter();

private:
  IlcIntVarArray  vars;            // problem variables
  DisjunctivePropagator* disj;     // disjunctive propagator
  IlcIntVar objective;             // objective function

  // precedence constraint
  vector< vector< vector<double> > > lambda_prec;	    // multipliers at current iteration
  vector< vector< vector<double> > > gradient_prec;   // gradient
  IlcRevFloat****	   lambda_prec_rev; // reversible lambda multipliers

  int 		  n_rounds;        // lagrangian rounds
  const int       num_acts;        // number of activities
  const int       num_pos;         // number of positions

  double sum_lambdas;
};


//
// Main Lagrangian iteration
//
void ObjectiveFunctionPropagatorMDDI::propagate() {

  // if (performed_lagrangian) {
  //   disj->set_cost_val_layer(cost);
  //   return;
  // }

  // !!!!!!!!!!!!!!!!!!!!!!!
  if (n_rounds > 0) {
    return;
  }

  // if objective is fixed, no need to run
  if (objective.isFixed()) {
    return;
  }

  // cases where lagrangian should be performed
  //if (n_rounds > 0 && (objective.getMax() -objective.getMin() > 2000)) {
  //return;
  //}

  clock_t time_lagrangian_local = clock();
  clock_t time_shortestpath_local;
  clock_t time_lp_local;

  // adjust MDD arcs
  RevMDD* mdd = disj->get_mdd();
  for( int l = 0; l < mdd->get_num_layers()-1; ++l) {
    for( RevMDD::Arc* arc = mdd->arcs_layer[l].get_first(); 
	 arc != NULL; 
	 arc = mdd->arcs_layer[l].get_next(arc) ) 
      {
	arc->target_node = arc->get_target_rev();
      }
  }


  // ------------------------------------------------------------------------
  // LP model initialization
  // ------------------------------------------------------------------------

  IloEnv env;
  
  // variables
  IloNumVarArray lambda_perm_var(env, num_acts, -IloInfinity, IloInfinity);

  /* !!
     IloArray< IloArray<IloNumVarArray> > lambda_prec_var(env, num_acts);
     for (int i = 0; i < num_acts; ++i) {
     lambda_prec_var[i] = IloArray<IloNumVarArray>(env, num_acts);
     for (int j = 0; j < num_acts; ++j) {
     lambda_prec_var[i][j] = IloNumVarArray(env, num_pos, -IloInfinity, 0);
     }    
     }
  */

  // model
  IloModel model(env);

  // objective
  IloNumVar obj(env, 0, objective.getMax());
  //IloNumVar obj(env, 0, 30000);
  model.add( IloMaximize(env, obj) );

  IloCplex cplex(model);
  cplex.setParam(IloCplex::Threads, 1);
  cplex.setOut(env.getNullStream());
  
  // ------------------------------------------------------------------------

  // set initial lambdas

  // permutation
  for (int i = 0; i < num_acts; ++i) {
    lambda_perm[i] = lambda_perm_rev[i]->getValue();
  }

  // precedences
  /* !!
     for (int i = 0; i < num_acts; ++i) {
     for (int j = 0; j < num_acts; ++j) {
     if (i != j && disj->must_precede(i,j)) {
     for (int t = 0; t < num_pos; ++t) {
     lambda_prec[i][j][t] = lambda_prec_rev[i][j][t]->getValue();
     }
     }
     }
     }
  */

  double primal_cost = -1;
  double dual_cost = -1;
  double max_primal = -1;

  int max_iterations = 100000;
  //int max_iterations = 18;
  int iteration;
  bool feasible;
  for (iteration = 0; iteration < max_iterations; ++iteration) {
    
    // Update cost structure for current lambdas

    // permutation
    primal_cost = 0;
    for (int i = 0; i < num_acts; ++i) {
      for (int j = 0; j < num_pos; ++j) {
  	cost[i][j] = -lambda_perm[i];
      }
      primal_cost += lambda_perm[i];
    }
    
    // precedences
    /* !!
       for (int i = 0; i < num_acts; ++i) {
       for (int j = 0; j < num_acts; ++j) {
       if (i != j && disj->must_precede(i,j)) {       
       for (int t = 0; t < num_pos; ++t) {

       primal_cost += lambda_prec[i][j][t];

       cost[i][t] -= lambda_prec[i][j][t];
       for (int t_prime = 0; t_prime < t; ++t_prime) {
       cost[j][t_prime] -= lambda_prec[i][j][t];
       }
       }
       }
       }
       }
    */
  
    // Compute shortest path

    time_shortestpath_local = clock();
    double sp_value;

    int* shortest_path = disj->compute_shortest_path(cost, sp_value);
    primal_cost += sp_value;

    max_primal = std::max(primal_cost, max_primal);
    time_shortestpath += clock() - time_shortestpath_local;

    // cout << "SP = ";
    // for (int i = 0; i < num_vars; ++i) {
    //   cout << shortest_path[i] << " ";
    // }
    // cout << endl;
    // cout << "Primal cost = " << sp_value << endl;


    // Compute primal value of feasible solution. Also check feasibility.
    double cx = 0;
    feasible = false;    
    for (int i = 0; i < num_acts; ++i) {
      // for (int j = i+1; j < num_acts; ++j) {
      // 	feasible &= (shortest_path[i] != shortest_path[j]);
      // 	feasible &= !disj->must_precede(shortest_path[j], shortest_path[i]);     
      // }
      if (i > 0) {
  	cx += disj->mult[i] * disj->setup[shortest_path[i-1]][shortest_path[i]];
      }
    }

    // If feasible, we are done: fix solution
    if (feasible) {
      cout << "Feasible!" << endl;
      for (int i = 0; i < num_acts; ++i) {
  	getCP().add( vars[i] == shortest_path[i] );
      }
      break;
    }

    // stopping conditions
    if (n_rounds > 0) {
      break;
    }
    //if (n_rounds > 0 && iteration > 10) {      
    //  primal_cost = max_primal;
    //  //cout << "\tprimal_cost =" << primal_cost << " - max_primal=" << max_primal << endl;
    // break;
    //}

    // Compute gradient

    // permutation
    std::fill(gradient_perm.begin(), gradient_perm.end(), -1);
    for (int i = 0; i < num_acts; ++i) {
      gradient_perm[shortest_path[i]]++;
    }

    // precedences
    /* !!
       for (int i = 0; i < num_acts; ++i) {
       for (int j = 0; j < num_acts; ++j) {
       if (i != j && disj->must_precede(i,j)) {
       for (int t = 0; t < num_pos; ++t) {
       gradient_prec[i][j][t] = -1;

       gradient_prec[i][j][t] += (shortest_path[t] == i ? 1.0 : 0.0);
       for (int t_prime = 0; t_prime < t; ++t_prime) {
       gradient_prec[i][j][t] += (shortest_path[t_prime] == j ? 1.0 : 0.0);
       }
       }
       }
       }
       }
    */
    
    // cout << "Gradient: " << endl;
    // for (int i = 0; i < num_vars; ++i) {
    //   cout << "\tval " << i << " - grad = " << gradient[i] << endl;
    // }

    // Add new cut to LP model and compute dual solution
    try {

      time_lp_local = clock();

      IloExpr lagrangian_term(env);

      // permutation
      for (int i = 0; i < num_acts; ++i) {
  	lagrangian_term -= lambda_perm_var[i] * gradient_perm[i];
      }

      // precedences
      /* !!
	 for (int i = 0; i < num_acts; ++i) {
	 for (int j = 0; j < num_acts; ++j) {
	 if (i != j && disj->must_precede(i,j)) {
	 for (int t = 0; t < num_pos; ++t) {
	 lagrangian_term -= lambda_prec_var[i][j][t]*gradient_prec[i][j][t];
	 }
	 }
	 }
	 }
      */

      model.add( obj <= cx + lagrangian_term );
      //cplex.exportModel("test.lp");

      bool solved = cplex.solve();
      if (!solved) {
  	cout << "ERROR: LP solution not found" << endl;	
  	exit(1);
  	//env.end();
  	//getCP().fail();
      }
      time_lp = clock() - time_lp_local;

      dual_cost = cplex.getObjValue();
      //cout << "Dual solution value = " << dual_cost << endl;


      // Check optimality conditions
      if (std::abs(primal_cost - dual_cost) <= 0.001) {
  	//cout << "OPT: " << primal_cost << " <--> " << dual_cost << endl;
  	break;
      } 

      //if (dual_cost-primal_cost <= 1.0) {
      //break;
      //}

      double gap = 100*((dual_cost - primal_cost)/primal_cost);
      if (primal_cost > 0 && dual_cost > 0 && gap <= 1) {
       	break;
      }

     

      // // check breaking conditions
      // if (n_rounds > 0 && iteration > 50 && primal_cost >= objective.getMin()) {
      //  	break;
      // }

      // permutation
      for (int i = 0; i < num_acts; ++i) {
  	if (cplex.isExtracted(lambda_perm_var[i])) {
  	  lambda_perm[i] = cplex.getValue(lambda_perm_var[i]);
  	  //cout << "\t\t" << j << "," << t << " --> " << lambda[j][t] << endl;
  	} else {
  	  lambda_perm[i]= 0;
  	}
      }

      // precedences
      /* !!
	 for (int i = 0; i < num_acts; ++i) {
	 for (int j = 0; j < num_acts; ++j) {
	 if (i != j && disj->must_precede(i,j)) {
	 for (int t = 0; t < num_pos; ++t) {
	 if (cplex.isExtracted(lambda_prec_var[i][j][t])) {
	 lambda_prec[i][j][t] = cplex.getValue(lambda_prec_var[i][j][t]);
	 } else {
	 lambda_prec[i][j][t]= 0;
	 }     
	 }
	 }
	 }
	 }
      */
   
      // Check breaking conditions
      //if (n_rounds > 0 && primal_cost >= objective.getMin()) {
      //break;
      //}

      if (n_rounds == 0) {
  	cout << "iter=" << iteration << "\tprimal=" << primal_cost << "\tdual=" << dual_cost;
  	cout << " \tgap=" << gap << "%";
  	cout << endl;
      }

      //cplex.end();
      // for (int i = 0; i < num_acts; ++i) {
      // 	cout << lambda_perm[i] << "\t";
      // 	if (i%11 == 0)
      // 	  cout << endl;
      // }
      // cout << endl;

      

    } catch (IloException &ex) {
      cout << "Ilog error: " << ex << endl;
      exit(1);
    }
  }
  env.end();

  // if feasible, then it is optimal
  if (feasible) {
    return;
  }

  // set initial lambdas
  
  // permutation
  for (int i = 0; i < num_acts; ++i) {
    lambda_perm_rev[i]->setValue(getCP(), lambda_perm[i]);
  }

  // precedences
  /* !!
     for (int i = 0; i < num_acts; ++i) {
     for (int j = 0; j < num_acts; ++j) {
     if (i == j) {
     continue;
     }
     for (int t = 0; t < num_pos; ++t) {
     lambda_prec_rev[i][j][t]->setValue(getCP(), lambda_prec[i][j][t]);
     }
     }
     }
  */



  time_lagrangian += clock() - time_lagrangian_local;
  lag_relax = primal_cost;

  if (n_rounds == 0) {
    cout << "\nLagrangian round=" << n_rounds;
    cout << " - primal=" << primal_cost;
    cout << " - dual=" << dual_cost;
    cout << " - iterations=" << iteration;
    cout << " - obj_min=" << objective.getMin();
    cout << " - time sp=" << ((double)time_shortestpath/CLOCKS_PER_SEC);
    cout << " - time lp=" << ((double)time_lp/CLOCKS_PER_SEC);
    cout << " - total time= " << ((double)time_lagrangian/CLOCKS_PER_SEC);
    //cout << " - gap = " << 100*((dual_cost - primal_cost)/primal_cost) << "%";
    cout << endl;
    cout << endl;
  }

  //exit(1);

  // // feasibility condition w.r.t. upper bound
  // if ((int)primal_cost > objective.getMax()) {
  //   getCP().fail();
  // }

  //if (objective.getMin() <= primal_cost) {

  //cout << "Setting costs..." << endl;


  // set costs
  sum_lambdas = 0;

  // permutation
  for (int i = 0; i < num_acts; ++i) {
    sum_lambdas += lambda_perm[i];
  }    
  
  // add costs to initial arcs
  for (int i = 0; i < num_acts; ++i) {
    cost[i][0] += sum_lambdas;
  }

  disj->set_cost_val_layer(cost);
  performed_lagrangian = true;

  // double test;
  // disj->compute_shortest_path(cost, test);
  // cout << "TEST = " << test << endl;
  

  n_rounds++;

  //exit(1);
}

IlcConstraint IlcObjectiveFunctionPropagatorMDD(IlcIntVarArray _vars,
						DisjunctivePropagator* _disj,
						IlcIntVar _objective)
{
  IloCP cp = _vars[0].getCP();
  return new (cp.getHeap()) ObjectiveFunctionPropagatorMDDI(cp, _vars, _disj, _objective);
}


ILOCPCONSTRAINTWRAPPER3(ObjectiveFunctionPropagatorMDDCt,
			cp,
			IloIntVarArray, _vars,
			DisjunctivePropagator*, _disj,
			IloIntVar, objective)
{
  return IlcObjectiveFunctionPropagatorMDD(cp.getIntVarArray(_vars), _disj, cp.getIntVar(objective));
}




//
// -----------------------------------------------------------------------
//


/**
 * Simple constraint to impose precedences into a
 * sequence var object
 */
class ImposePrecedences : public IlcConstraintI {

public:

  /** Constructors */
  ImposePrecedences(IloCP cp,
		    IlcIntervalVarArray _interval_vars,
		    IlcIntervalSequenceVar _seq_var,
		    bool** _precedes, int _num_tasks)
    : IlcConstraintI(cp), interval_vars(_interval_vars), seq_var(_seq_var),
      precedes(_precedes), num_tasks(_num_tasks)
  { }

  /** ILOG search control */
  virtual void propagate() {
    for( int i = 0; i < num_tasks; ++i ) {
      for( int j = 0; j < num_tasks; ++j ) {
	if( i != j && precedes[i][j] ) {
	  //cout << "activity " << i << " precedes " << j << endl;
	  seq_var.setBefore(interval_vars[i], interval_vars[j]);
	}
      }
    }
  }
  virtual void post() { }
  virtual IlcBool isViolated() const { return false;}

  /** Empty destructor */
  ~ImposePrecedences(){ };

private:
  IlcIntervalVarArray interval_vars;
  IlcIntervalSequenceVar seq_var;
  bool** precedes;
  int num_tasks;
};

/**
 * Wrapper 1
 */
inline IlcConstraint IlcImposePrecedences(
					  IloCP cp,
					  IlcIntervalVarArray _interval_vars,
					  IlcIntervalSequenceVar _seq_var, bool** precedes, int num_tasks)
{
  return new(cp.getHeap()) ImposePrecedences(cp, _interval_vars, _seq_var, precedes, num_tasks);
}



/**
 * Wrapper for ILO object
 */
ILOCPCONSTRAINTWRAPPER4(IloImposePrecedences,
			cp,
			IloIntervalVarArray, _interval_vars,
			IloIntervalSequenceVar, _seq_var, bool**, precedes, int, num_tasks)
{
  // compose sequence var
  IlcIntervalSequenceVar seq_var;
  if( _seq_var.getImpl() != NULL ) {
    use(cp, _seq_var);
    seq_var = cp.getIntervalSequence(_seq_var);
  }

  // compose interval var
  IlcIntervalVarArray interval_vars(cp, _interval_vars.getSize());
  for( int i = 0; i < interval_vars.getSize(); i++ ) {
    use(cp, _interval_vars[i]);
    interval_vars[i] = cp.getInterval(_interval_vars[i]);
  }

  return IlcImposePrecedences(cp, interval_vars, seq_var, precedes, num_tasks);
}


/**
 * Solution
 */
struct Solution {
  vector<int> permutation;    // activity permutation
  double      time;           // time to reach solution
  int         branches;       // number of branches
  int         fails;          // number of fails
  int         makespan;       // makespan
  int         sum_setup;      // sum of setup times
  int         value;		    // solution objective function value
};


/**
 * Print solution to file
 */
void print_solution(
		    ActivityArray &activities,
		    IloCP& cp,
		    IloIntVarArray& permutation,
		    IloIntVar &makespan,
		    IloIntVar &sum_setup,
		    ofstream &out)
{
  // solution
  for( int i = 0; i < permutation.getSize(); i++ ) {
    out << cp.getValue(permutation[i]) << " ";
  }
  out << "\t[" << cp.getValue(makespan) << "]";
  out << "\t[" << cp.getValue(sum_setup) << "]";

  // search info
  out << "\t" << cp.getInfo(IloCP::NumberOfBranches);
  out << "\t" << cp.getInfo(IloCP::NumberOfFails);
  if( cp.getInfo(IloCP::FailStatus) == IloCP::SearchStoppedByLimit ) {
    out << "\tTL";
  } else {
    out << "\t" << cp.getTime();
  }

  out << endl;
}


/**
 * Main
 */
//int main_tsptw(int argc, char* argv[]) {
int main(int argc, char* argv[]) {

  if( argc < 6 || ( atoi(argv[5]) > 0 && argc < 7 ) ) {
    cout << "\nUsage: disjunctive [filename] [input-type] "
      "[obj-function] [search-type] [propagation-type] [width] "
      " [opt: time limit, INF = -1] [opt: refinement-type] [opt: lagrangian]\n" << endl;
    cout << "\twhere: " << endl;
    cout << "\t\tinput_type = 0: tsptw instances" << endl;
    cout << "\t\tinput_type = 1: modified AFG instances" << endl;
    cout << "\t\tinput_type = 2: WT instances" << endl;
    cout << "\t\tinput_type = 3: WTS instances" << endl;
    cout << "\t\tinput_type = 4: disjunctive instances" << endl;
    cout << "\t\tinput_type = 5: sequential ordering problem" << endl;
    cout << "\t\tinput_type = 6: TSP with Pickup and Delivery" << endl;
    cout << "\t\tinput_type = 7: random instances" << endl;
    cout << "\t\tinput_type = 8: Time-Dependent TSP" << endl;
    cout << endl;
    //cout << "\t\tobj-function = 0: minimize makespan" << endl;
    cout << "\t\tobj-function = 1: minimize sum of setup times" << endl;
    //cout << "\t\tobj-function = 2: minimize total tardiness" << endl;
    //cout << "\t\tobj-function = 3: minimize max tardiness" << endl;
    //cout << "\t\tobj-function = 4: minimize latency" << endl;
    //cout << "\t\tobj-function = 5: none - only feasibility" << endl;
    cout << endl;
    cout << "\t\tsearch-type = 0: lexicographic search" << endl;
    cout << "\t\tsearch-type = 1: default search" << endl;
    cout << "\t\tsearch-type = 2: MDD shortest path search" << endl;
    cout << "\t\tsearch-type = 3: MDD min setup T search" << endl;
    cout << "\t\tsearch-type = 4: branch on MDD nodes" << endl;
    cout << "\t\tsearch-type = 5: default search with DFS" << endl;
    cout << "\t\tsearch-type = 6: fix next activity from shortest path" << endl;
    cout << endl;
    cout << "\t\tpropagation-type = 0: pure CP" << endl;
    cout << "\t\tpropagation-type = 1: pure MDD" << endl;
    cout << "\t\tpropagation-type = 2: CP+MDD" << endl;
    cout << endl;
    cout << "\t\trefinement-type = 0: deterministic according to objective (default)" << endl;
    cout << "\t\trefinement-type = 1: random" << endl;
    cout << "\t\trefinement-type = 2: read from file [order-file]" << endl;
    cout << "\n\t\t** Obs.: This version only works with obj-function=1: minimize sum of setup times\n";
    cout << endl << endl;

    exit(1);
  }


  try {

    IloEnv env;
    IloModel model(env);

    // ----------------------------------------------
    // Read input
    // ----------------------------------------------

    char inst_name[256];
    extract_instance_name(argv[1], inst_name);

    int input_type = atoi(argv[2]);

    DisjunctiveObjective obj_type = static_cast<DisjunctiveObjective>(atoi(argv[3]));
    int search_type = atoi(argv[4]);
    int propagation_type = atoi(argv[5]);

    if (obj_type != SumSetupTimes) {
      cout << "\nError: this version (MDD Lagrangian paper) only works for minimizing sum of setup times." << endl << endl;
    }
    
    // time limit (if set)
    int time_limit = TIME_LIMIT;
    if (argc >= 8) {
      time_limit =  atoi(argv[7]);
    }

    int width = 0;

    ActivityArray acts;
    int** setup;
    bool** input_precedence;

    switch( input_type ) {
    case 0:
      read_dumas_input(env, argv[1], acts, setup);
      break;
    case 1:
      read_afg_input(env, argv[1], acts, setup);
      break;
    case WT_INPUT:
      read_wt_input(env, argv[1], acts, setup);
      break;
    case WTS_INPUT:
      read_wts_input(env, argv[1], acts, setup);
      break;
    case DISJ_INPUT:
      read_disj_input(env, argv[1], acts, setup);
      break;
    case SOP_INPUT:
      read_sop_input(env, argv[1], acts, setup, input_precedence);
      break;
    case TSP_PD_INPUT:
      read_tsppd_input(env, argv[1], acts, setup, input_precedence);
      break;
    case RANDOM_INPUT:
      read_random_input(env, argv[1], acts, setup);
      break;
    case TD_TSP_INPUT:
      read_td_tsp_input(env, argv[1], acts, setup);
      break;
    default:
      cout << "Error: instance type cannot be identified." << endl;
      exit(1);
    }

    cout << "\nInstance name: " << argv[1] << endl;

    // -----------------------------------------------
    // ILOG CPO Model
    // -----------------------------------------------

    /*
     * Impose release/deadline constraint
     */

    // set of ILO activities
    IloIntervalVarArray activities(env);

    // end times of activities
    IloIntExprArray end_times(env);

    for( ActivityArray::iterator act = acts.begin(); act != acts.end(); ++act ) {

      // release/deadline constraint
      model.add( IloStartOf((*act)->ilo_start_time) >= (*act)->release );
      model.add( IloEndOf((*act)->ilo_start_time) <= (*act)->deadline );

      end_times.add( IloEndOf((*act)->ilo_start_time) );

      // add to set of activities
      activities.add((*act)->ilo_start_time);
    }


    /*
     * Permutation variables
     */
    IloIntVarArray permutation;
    char varname[256];
    if( propagation_type == 1 || propagation_type == 2 ) {

      //permutation = IloIntVarArray(env, acts.size(), 0, acts.size()-1);

      permutation = IloIntVarArray(env, acts.size());
      for( int i = 0; i < (int)acts.size(); ++i ) {
	sprintf(varname, "permutation[%d]", i);
	//permutation[i] = IloIntVar(env, 0, acts.size()-1, varname);
	permutation[i] = IloIntVar(env, 0, acts.size()-1);
      }

      if( input_type == DUMAS_INPUT || input_type == AFG_INPUT || input_type == TSP_PD_INPUT 
	  || input_type == TD_TSP_INPUT)
	{
	  // first and last cities are depot
	  permutation[0].setMax(0);
	  permutation[acts.size()-1].setMin(acts.size()-1);
	} else if( input_type == WTS_INPUT ) {
	permutation[0].setMax(0);
      }

      // Enforce that all permutation variables must be distinct
      // Note: this is added outside the disjunctive constraint to allow the alldiff CPO propagator
      // to be performed first
      model.add( IloAllDiff(env, permutation) );
    }


    // objective function variable (if any)
    IloIntVar obj_var;
    if( obj_type != None ) {
      sprintf(varname, "obj_var[%d]", obj_type);
      obj_var = IloIntVar(env, 0, INF, varname);
      //obj_var = IloIntVar(env, 144494, 144494);
      //obj_var = IloIntVar(env, 144495, 144500);
      model.add( IloMinimize(env, obj_var) );
    }
    //model.add( obj_var <= 7500 );

    // create sequence variable with empty handle
    IloIntervalSequenceVar sequence;

    // ------------------------------------------------------
    // ILOG CP disjunctive propagator
    // ------------------------------------------------------
    if( propagation_type == 0 || propagation_type == 2 ) {

      // define activity times
      IloIntArray types(env, acts.size());
      for( unsigned int i = 0; i < acts.size(); i++ ) {
	types[i] = i;
      }

      // create sequence variables
      sequence = IloIntervalSequenceVar(env, activities, types);

      // TODO: change!! This just force extraction... (because of a deprecated function
      // in version 12.5
      for( unsigned int i = 0; i < acts.size(); i++ ) {
	model.add( IloTypeOfNext(sequence, acts[i]->ilo_start_time, -1) >= -100 );
      }


      if( input_type == DUMAS_INPUT || input_type == AFG_INPUT || input_type == TSP_PD_INPUT || 
	  input_type == TD_TSP_INPUT) {
	// we fix first and last activity in the sequence
	model.add( IloFirst(env, sequence, acts[0]->ilo_start_time) );
	model.add( IloLast(env, sequence, acts.back()->ilo_start_time) );
      } else if( input_type == WTS_INPUT ) {
	model.add( IloFirst(env, sequence, acts[0]->ilo_start_time) );
      }

      // create transition cost matrix
      IloIntArray2   		  setup_costs(env, acts.size());
      IloTransitionDistance setup_times(env, acts.size());

      // Type of next for the last activity on the machine
      for( unsigned int i = 0; i < acts.size(); i++ ) {
	setup_costs[i] = IloIntArray(env, acts.size()+1);
	for( unsigned int j = 0; j < acts.size(); j++ ) {
	  setup_times.setValue(i, j, setup[i][j]);
	  setup_costs[i][j] = setup[i][j];

	  if( input_type == SOP_INPUT ||  input_type == TSP_PD_INPUT ) {
	    if( i != j && input_precedence[j][i] ) {
	      setup_times.setValue(i, j, IloIntervalMax);
	    }
	  }
	  ///cout << i << "," << j << " --> " << setup_times.getValue(i,j) << endl;
	}
	setup_costs[i][acts.size()] = 0;
      }

      // check input precedences
      if( input_type == SOP_INPUT || input_type == TSP_PD_INPUT ) {
	//model.add( IloImposePrecedences(env, activities, sequence, input_precedence, acts.size()) );
	for( unsigned int i = 0; i < acts.size(); ++i ) {
	  for( unsigned int j = i+1; j < acts.size(); ++j ) {

	    if( input_precedence[i][j] ) {
	      //cout << acts[i]->ilo_start_time << " << " << acts[j]->ilo_start_time << endl;

	      model.add( IloBefore(env, sequence,
				   acts[i]->ilo_start_time, acts[j]->ilo_start_time ) );
	      model.add( IloEndBeforeStart(env,
					   acts[i]->ilo_start_time,
					   acts[j]->ilo_start_time) );

	    } else if( input_precedence[j][i] ) {
	      //cout << acts[j]->ilo_start_time << " << " << acts[i]->ilo_start_time << endl;

	      model.add( IloBefore(env, sequence,
				   acts[j]->ilo_start_time, acts[i]->ilo_start_time ) );
	      model.add( IloEndBeforeStart(env,
					   acts[j]->ilo_start_time,
					   acts[i]->ilo_start_time) );
	    }
	  }
	}
      }

      // add CP optimizer disjunctive constraint
      model.add( IloNoOverlap(env, sequence, setup_times, IloTrue) );

      // create objective function
      switch( obj_type ) {

      case Makespan:
	model.add( obj_var == IloMax(end_times) );
	break;

      case SumSetupTimes: {

	if ( input_type == TSP_PD_INPUT || input_type == TD_TSP_INPUT) {
	  model.add( obj_var >= IloMax(end_times) );
	} else {
	  IloIntExpr obj(env);
	  for( int i = 0; i < (int)acts.size(); i++ ) {
	    obj += setup_costs[types[i]][IloTypeOfNext(sequence, acts[i]->ilo_start_time, acts.size())];
	  }
	  model.add( obj_var == obj);
	}
	break;
      }

      case TotalTardiness: {
	IloIntExpr tardiness_sum(env);
	for( int i = 0; i < (int)acts.size(); ++i ) {
	  tardiness_sum += (acts[i]->weight) * IloMax(0, IloEndOf(acts[i]->ilo_start_time) - acts[i]->duedate);
	}
	model.add( obj_var == tardiness_sum );
	break;
      }

      case MaxTardiness: {

	// compute upper bound on max tardiness
	int max_release = 0;
	int max_weight = 0;
	int sum_p = 0;
	for( ActivityArray::iterator it = acts.begin(); it != acts.end(); ++it ) {
	  max_release = MAX( max_release, (*it)->release );
	  max_weight = MAX( max_weight, (*it)->weight );
	  sum_p += (*it)->processing;
	}
	obj_var.setMax( (max_release + sum_p)*max_weight );

	IloIntExprArray tardiness_v(env);
	for( int i = 0; i < (int)acts.size(); ++i ) {
	  tardiness_v.add((acts[i]->weight) * IloMax(0, IloEndOf(acts[i]->ilo_start_time) - acts[i]->duedate));
	}
	model.add( obj_var == IloMax( tardiness_v ) );
	break;
      }


      case Latency: {
	IloIntExpr latency_sum(env);
	for( int i = 0; i < (int)acts.size(); ++i ) {
	  latency_sum += (acts[i]->weight) * IloEndOf(acts[i]->ilo_start_time);
	}
	model.add( obj_var == latency_sum );
	break;
      }


      default:
	break;
      }

      // add an an additional obj. function constraint for sequential ordering problem
      if( input_type == SOP_INPUT ||  input_type == TD_TSP_INPUT) {
	model.add( obj_var == IloMax(end_times) );
      }
    }




    //		// Link with next variables
    //		if( propagation_type == 2 ) {
    //
    //			IloIntVarArray next_vars(env, acts.size()-1, 1, acts.size()-1);
    //			for( int i = 0; i < (int)acts.size()-1; ++i ) {
    //				model.add( next_vars[i] == IloTypeOfNext(sequence, activities[i], -1) );
    //			}
    //
    //			for( int i = 0; i < (int)acts.size()-2; ++i ) {
    //				model.add( next_vars[ permutation[i] ] == permutation[i+1] );
    //			}
    //
    //		}


    
    // ------------------------------------------------------
    // MDD Propagator
    // ------------------------------------------------------

    DisjunctivePropagator* disj_prop = NULL;
    if( propagation_type == 1 || propagation_type == 2 ) {

      // read width
      width = atoi(argv[6]);
      if( width <= 0 ) {
	cout << "\nError: invalid width" << endl << endl;
	exit(1);
      }
      cout << "width: " << width << endl << endl;

      // create MDD disjunctive propagator

      IloArray<MDDPropagator*> propagators(env);

      if( obj_type != None ) {
	disj_prop = new DisjunctivePropagator(acts, sequence, permutation, setup,
					      obj_type, obj_var);
      } else {
	disj_prop = new DisjunctivePropagator(acts, sequence, permutation, setup);
      }
      propagators.add( disj_prop );

      // create MDD handler with propagators and add it to model
      model.add( IloMDDHandlerScheduling(env, permutation, activities, sequence, propagators, width, "mdd-handler") );
			
			
      // check input precedences
      if( input_type == SOP_INPUT|| input_type == TSP_PD_INPUT ) {
	for( int i = 0; i < (int)acts.size(); ++i ) {
	  for( int j = 0; j < (int)acts.size(); ++j ) {
	    if( i == j ) continue;
	    if( input_precedence[i][j] ) {
	      disj_prop->set_precedence(i, j);
	    }
	  }
	}
      }

      // add objective function propagator
      //if (input_type == SOP_INPUT || input_type == TD_TSP_INPUT) {
      if (input_type == TD_TSP_INPUT) {
	disj_prop->set_td_multipliers();
      }

      // lagrangian
      if (argc >= 10 && atoi(argv[9]) == 1) {

	cout << endl;
	cout << endl;
	cout << "Creating auxiliary model..." << endl;

	// create auxiliary model to compute lagrangian multipliers
	IloModel lagModel(env);
	lagModel.add(model);
            
	// propagate model to create initial MDDs
	IloCP cp(lagModel);
	cp.setParameter(IloCP::Workers, 1); // only one core
	cp.setParameter(IloCP::DefaultInferenceLevel, IloCP::Extended);
	cp.propagate();

	initial_relax = cp.getIntVar(obj_var).getMin();

	cout << "** Initial objective function = " << cp.getIntVar(obj_var);	
	cout << endl << endl;

	// compute multipliers
	ObjectiveFunctionPropagatorMDDI obj(cp, 
					    cp.getIntVarArray(permutation),
					    disj_prop,
					    cp.getIntVar(obj_var));

	obj.propagate();
	cp.end();
	lagModel.end();

	// create lagrangian objective function variable
	char varname[256];
	sprintf(varname, "obj_var_lag");
	IloIntVar obj_var_lag(env, -10000000, 10000000, varname);
	disj_prop->set_obj_var_lagrangian(obj_var_lag);
	   
	// In this case, Lagrangian objective corresponds to exact cost
	model.add( obj_var_lag == obj_var );

	//cout << "\tdone computing Lagrangian multipliers" << endl;
	cout << endl;

	//exit(1);
      }


      // adjustment for some objective functions
      switch( obj_type ) {

      case Makespan:
	// TODO: check this!!
	model.add( obj_var == IloMax(end_times) );
	break;

      default:
	break;
      }

      // refinement type (if set)
      if( argc >= 9 ) {
	if( argv[8][0] == '1' ) {
	  disj_prop->set_random_exact_refinement();
	} else if( argv[8][0] == '2' ) {
	  vector<int> order( acts.size() );
	  ifstream input_file( argv[9] );
	  for( int i = 0; i < (int)acts.size(); ++i ) {
	    input_file >> order[i];
	  }
	  input_file.close();
	  disj_prop->set_exact_act_order( order );
	}
      }

      
    }


    // srand(0);	       
    // vector<int> order(acts.size());
    // for (size_t i = 0; i < acts.size(); ++i) {
    //   order[i] = i;
    // }

    // int best_bound = -1;
    // vector<int> best_order;

    // for (int i = 0; i < 100; ++i) {      

    //   std::random_shuffle(order.begin(), order.end());
    //   disj_prop->set_exact_act_order(order);

    //   IloCP cp(model);  
    //   cp.setParameter(IloCP::Workers, 1);
    //   cp.setParameter(IloCP::DefaultInferenceLevel, IloCP::Extended);
      
    //   cp.propagate();
    //   cout << "bound = " << cp.getIntVar(obj_var).getMin() << endl;
    //   if (cp.getIntVar(obj_var).getMin() > best_bound) {
    // 	best_bound = cp.getIntVar(obj_var).getMin();
    // 	best_order = order;
    //   }

    //   cp.end();
    // }

    // cout << "Best bound = " << best_bound << endl;
    // disj_prop->set_exact_act_order(best_order);
    
    //    exit(1);

    // model.add( permutation[0] == 0 );
    // model.add( permutation[1] == 9 );
    // model.add( permutation[2] == 11 );
    // model.add( permutation[3] == 2 );
    // model.add( permutation[4] == 15 );
    // model.add( permutation[5] == 16 );
    // model.add( permutation[6] == 8 );
    // model.add( permutation[7] == 4 );
    // model.add( permutation[8] == 1 );
    // model.add( permutation[9] == 5 );
    // model.add( permutation[10] == 6 );
    // model.add( permutation[11] == 10 );
    // model.add( permutation[12] == 14 );
    // model.add( permutation[13] == 18 );
    // model.add( permutation[14] == 13 );
    // model.add( permutation[15] == 7 );
    // model.add( permutation[16] == 3 );
    // model.add( permutation[17] == 17 );
    // model.add( permutation[18] == 12 );
    // model.add( permutation[19] == 20 );
    // model.add( permutation[20] == 19 );
    // model.add( permutation[21] == 21 );


    // model.add( permutation[8] == 1 );
    // model.add( permutation[9] == 20 );
    // model.add( permutation[10] == 19 );
    // model.add( permutation[11] == 9 );
    // model.add( permutation[12] == 12 );
    // model.add( permutation[13] == 23 );
    // model.add( permutation[14] == 7 );
    // model.add( permutation[15] == 26 );
    // model.add( permutation[16] == 15 );
    // model.add( permutation[17] == 18 );    
    // model.add( permutation[18] == 3 );
    // model.add( permutation[19] == 14 );
    // model.add( permutation[20] == 17 );
    // model.add( permutation[21] == 13 );
    // model.add( permutation[22] == 16 );
    // model.add( permutation[23] == 21 );
    // model.add( permutation[24] == 10 );
    // model.add( permutation[25] == 24 );
    // model.add( permutation[26] == 6 );
    // model.add( permutation[27] == 22 );
    // model.add( permutation[28] == 2 );
    // model.add( permutation[29] == 29 );


      
    // -----------------------------------------------
    // Solve
    // -----------------------------------------------
    
    IloCP cp(model);
    
    // set parameters
    cp.setParameter(IloCP::Workers, 1); // only one core
    if (time_limit != -1) {
      cp.setParameter(IloCP::TimeLimit, time_limit);
    }
    cp.setParameter(IloCP::DefaultInferenceLevel, IloCP::Extended);
    cp.setParameter(IloCP::RelativeOptimalityTolerance, 1e-8);
    //		cp.setParameter(IloCP::SearchType, IloCP::Restart);

    if( search_type != DEFAULT_SEARCH ) {
      cp.setParameter(IloCP::SearchType, IloCP::DepthFirst);
    }
    //cp.setParameter(IloCP::RestartFailLimit, 500);
    //cp.setParameter(IloCP::RestartProofEmphasis, 1000);
    //cp.setParameter(IloCP::LogVerbosity, IloCP::Quiet);

    //cp.setParameter(IloCP::RestartProofEmphasis, 1);

    //
    //		// generate all solutions
    //		cp.setParameter(IloCP::LogVerbosity, IloCP::Quiet);
    //		long int sol_count = 0;
    //		ofstream sols("sols.txt");
    //		cp.startNewSearch(search_procedure);
    //		while( cp.next() ) {
    //			if( propagation_type == 1 ) {
    //				for( int i = 0; i < permutation.getSize(); i++ ) {
    //					sols << cp.getValue(permutation[i]) << " ";
    //				}
    //				sols << endl;
    //			} else {
    //				IlcIntervalSequenceVar ilcseq = cp.getIntervalSequence(sequence);
    //				for( IloIntervalVar act = cp.getFirst(sequence); act.getImpl() != NULL; act = cp.getNext(sequence, act) ) {
    //				  sols << ilcseq.getType(cp.getInterval(act)) << " ";
    //				}
    //				//sols << " " << cp.getIntVar(makespan) << " " << cp.getIntVar(sum_setup);
    //				sols << endl;
    //			}
    //			sol_count++;
    //		}
    //		if( sol_count == 0 ) {
    //		  sols << "-1" << endl;
    //		}
    //		sols.close();
    //		exit(1);


    //		if( cp.propagate() ) {
    //			cout << cp.getIntVar(makespan) << endl;
    //			for( int i = 0; i < permutation.getSize(); i++ ) {
    //				cout << " " << cp.getIntVar(permutation[i]);
    //			}
    //		} else {
    //			cout << endl << endl << "Infeasible!" << endl << endl;
    //		}
    //		cout << endl;
    //		exit(1);

    //        if( cp.solve(search_phase) ) {


    // -----------------------------------------------
    // Search

    //		if( cp.propagate() ) {
    //			cout << "makespan: " << cp.getIntVar(obj_var) << endl;
    //			if( permutation.getImpl() != NULL ) {
    //				for( int i = 0; i < permutation.getSize(); ++i ) {
    //					cout << "\t" << cp.getIntVar(permutation[i]) << endl;
    //				}
    //			}
    //			for( int i = 0; i < acts.size(); ++i ) {
    //				cout << "\t" << cp.getInterval(acts[i]->ilo_start_time) << endl;
    //			}
    //		}
    //		exit(1);

    vector<Solution> all_solutions;   // all solutions found during search
    all_solutions.reserve(1000000);

    // set search strategy
    IloGoal search_procedure;
    bool is_goal = false;

    IloSearchPhase search_phase;
    bool is_phase = false;

    if( search_type == LEX_SEARCH ) {

      // create search method
      if( propagation_type == 1 ) {

	// Pure MDD: perform lexicographic search on permutation variables
	search_procedure = IloGenerate(env, permutation)
	  && IloSetTimesPermutation(env, permutation, activities);
	if( obj_type != None ) {
	  search_procedure = search_procedure && IloFixIntVarMin(env, obj_var);
	}
	is_goal = true;

      } else {

	// CP or CP+MDD: lexicographic search on sequence of activities
	search_procedure = IloLexSchedule(env, activities, sequence)
	  && IloSetTimesSequence(env, sequence);

	if( obj_type != None ) {
	  search_procedure = search_procedure && IloFixIntVarMin(env, obj_var);
	}
	is_goal = true;
      }
      //cp.startNewSearch(search_procedure);

    } else if( search_type == DEFAULT_SEARCH || search_type == DEFAULT_SEARCH_DFS ) {

      if( propagation_type == 1 ) {
	search_phase = IloSearchPhase(env, permutation);
      } else {
	IloIntervalSequenceVarArray seqs_array(env);
	seqs_array.add(sequence);
	search_phase = IloSearchPhase(env, seqs_array);
      }
      //cp.startNewSearch(search_phase);
      is_phase = true;

    } else if( search_type == MDD_SP_SEARCH ) {

      if( propagation_type == 0 ) {
	cout << "Error: Propagation type must include MDD" << endl;
	exit(1);
      }

      IloGoal search_procedure = IloShortestPathMDD(env, permutation, disj_prop)
	&& IloSetTimesPermutation(env, permutation, activities);

      if( obj_type != None ) {
	search_procedure = search_procedure && IloFixIntVarMin(env, obj_var);
      }
      is_goal = true;

      //cp.startNewSearch(search_procedure);

    } else if( search_type == MIN_SETUP_T_SEARCH ) {

      if( propagation_type == 0 ) {
	cout << "Error: Propagation type must include MDD" << endl;
	exit(1);
      }

      search_procedure = IloSumSetupTimesMDD(env, permutation, disj_prop)
	&& IloSetTimesPermutation(env, permutation, activities);

      if( obj_type != None ) {
	search_procedure = search_procedure && IloFixIntVarMin(env, obj_var);
      }
      is_goal = true;

      //cp.startNewSearch(search_procedure);


    } else if( search_type == NODEBRANCH_SEARCH ) {

      if( propagation_type == 0 || width <= 1 ) {
	cout << "Error: Propagation type must include MDD with witdh > 1" << endl;
	exit(1);
      }


      // -----------------------------------------------------------
      // Begin Node Branching
      // -----------------------------------------------------------

      // // Node selector
      // global_cost = new(env) double*[acts.size()];
      // for (size_t i = 0; i < acts.size(); ++i) {
      // 	global_cost[i] = new(env) double[acts.size()];
      // }
      search_procedure = IloBranchOnMDDNodes(
					     env,
					     permutation,
					     disj_prop,
					     new(env) SelectFirstInexactNodeShortestPath(env, acts.size(), &global_cost))
	&&
	IloSetTimesPermutation(env, permutation, activities);


      // commented out:
      // search_procedure = IloMDDNodeBranch(env, permutation, disj_prop)
      //      		&& IloSetTimesPermutation(env, permutation, activities);

      // -----------------------------------------------------------
      // End Node Branching
      // -----------------------------------------------------------


      if( obj_type != None ) {
	search_procedure = search_procedure && IloFixIntVarMin(env, obj_var);
      }
      is_goal = true;

      //cp.startNewSearch(search_procedure);
    } else if (search_type == MDD_SHORTEST_PATH) {

      if( propagation_type == 0 || width <= 1 ) {
	cout << "Error: Propagation type must include MDD with witdh > 1" << endl;
	exit(1);
      }

      search_procedure = IloBranchOnMDDShortestPathAct(env, permutation, disj_prop);      
      is_goal = true;     
    }

    bool feasible = false;
    if( is_goal ) {
      feasible = cp.solve( search_procedure );

    } else if( is_phase ) {
      feasible = cp.solve( search_phase );
    }

    //		bool feasible = false;
    //		while( cp.next() ) {
    //
    //			feasible = true;
    //			if( obj_type == None ) {
    //				break;
    //			}
    //
    //			// store solution
    //			Solution solution;
    //			all_solutions.push_back(solution);
    //			all_solutions.back().permutation.reserve(acts.size());
    //
    //			if( propagation_type != 0 ) {
    //				for( int i = 0; i < permutation.getSize(); i++ ) {
    //					all_solutions.back().permutation.push_back(cp.getValue(permutation[i]));
    //				}
    //			} else {
    //				IlcIntervalSequenceVar ilcseq = cp.getIntervalSequence(sequence);
    //				for( IloIntervalVar act = cp.getFirst(sequence); act.getImpl() != NULL; act = cp.getNext(sequence, act) ) {
    //					all_solutions.back().permutation.push_back(ilcseq.getType(cp.getInterval(act)));
    //				}
    //			}
    //			//          cout << "Makespan: " << cp.getIntVar(makespan) << endl;
    //			//          cout << "Sum setup times: " << cp.getIntVar(sum_setup) << endl;
    //			//          cout << "Total tardiness: " << cp.getIntVar(total_tardiness) << endl;
    //			//          cout << endl;
    //			all_solutions.back().value = cp.getValue(obj_var);
    //			all_solutions.back().branches = cp.getInfo(IloCP::NumberOfBranches);
    //			all_solutions.back().fails = cp.getInfo(IloCP::NumberOfFails);
    //			all_solutions.back().time = cp.getTime();
    //			//all_solutions.back().makespan = cp.getValue(makespan);
    //		}

    int makespan = 0;
    int sum_setup = 0;

    if( feasible ) {
      if( propagation_type != 0 ) {
	cout << "\nSequence: " << endl;
	for( int i = 0; i < permutation.getSize(); i++ ) {
	  //cout << cp.getValue(permutation[i]) << " ";
	  cout << "\t";
	  cout << "Activity " << cp.getValue(permutation[i]);
	  cout << " :: " << cp.getInterval(activities[cp.getValue(permutation[i])]);
	  cout << endl;

	  makespan = MAX(makespan, cp.getInterval(activities[cp.getValue(permutation[i])]).getEndMax());
	  if( i > 0 ) {
	    sum_setup += setup[cp.getValue(permutation[i-1])][cp.getValue(permutation[i])];
	  }
	}
	cout << endl;

      } else {
	cout << "\nSequence: " << endl;
	IlcIntervalSequenceVar ilcseq = cp.getIntervalSequence(sequence);
	int i = 0;
	int pos = 0, current, prev=-1;
	for( IloIntervalVar act = cp.getFirst(sequence); act.getImpl() != NULL; act = cp.getNext(sequence, act) ) {
	  cout << "\t";
	  cout << "Activity " << ilcseq.getType(cp.getInterval(act));
	  if( permutation.getImpl() != NULL ) {
	    cout << " (" << cp.getValue(permutation[i++]) << ")";
	  }
	  cout << " :: " << cp.getInterval(act);
	  cout << endl;

	  current = ilcseq.getType(cp.getInterval(act));
	  makespan = MAX(makespan, cp.getInterval(activities[current]).getEndMax());
	  if( pos > 0 ) {
	    sum_setup += setup[prev][current];
	  }
	  pos++;
	  prev = current;
	}
	cout << endl;
      }

      cout << "\nMakespan: " << makespan << endl;
      cout << "Sum setup times: " << sum_setup << endl;

      if( obj_type != None ) {
	cout << "\nBest solution: " << cp.getObjValue() << endl;

	// if (input_type == TD_TSP_INPUT) {	
	//   double total = 0;
	//   for (int i = 1; i < (int)acts.size(); ++i) {
	//     cout << "\t";
	//     cout << cp.getValue(permutation[i-1]);
	//     cout << " --> ";
	//     cout << cp.getValue(permutation[i]);
	//     cout << " - setup=";
	//     cout << setup[cp.getValue(permutation[i-1])][cp.getValue(permutation[i])];
	//     cout << " - mult=";
	//     cout << (acts.size() - i);
	//     cout << " - result=";
	//     cout << (acts.size() - i)*setup[cp.getValue(permutation[i-1])][cp.getValue(permutation[i])];
	//     cout << endl;

	//     total +=(acts.size() - i)*setup[cp.getValue(permutation[i-1])][cp.getValue(permutation[i])];	 
	//   }
	//   cout << "Total distance = " << total << endl;
	//   exit(1);
	// }

      } else {
	//cout << "\nMakespan: " << cp.getValue(makespan) << endl;
	//cout << "Sum of setup times: " << cp.getValue(sum_setup) << endl;
      }
    } else {
      cout << "\nProblem is infeasible." << endl;
    }
    cout << "\nBranches: " << cp.getInfo(IloCP::NumberOfBranches) << endl;
    cout << "Fails: " << cp.getInfo(IloCP::NumberOfFails) << endl;
    cout << "Time: " << cp.getTime() << endl << endl;

    // statistics
    char statname[256];
    sprintf(statname, "stats-%d-%d.txt", propagation_type, width);

    //ofstream statsfile("stats.txt", ios::app);
    ofstream statsfile(statname, ios::app);
    statsfile << argv[1];

    // objective function
    switch( obj_type ) {
    case Makespan:
      statsfile << "\tMakespan";
      break;

    case SumSetupTimes: {
      statsfile << "\tSumSetup";
      break;
    }

    case TotalTardiness: {
      statsfile << "\tTotalTardiness";
      break;
    }

    case MaxTardiness: {
      statsfile << "\tMaxTardiness";
      break;
    }

    case Latency: {
      statsfile << "\tLatency";
      break;
    }
    default:
      break;
    }


    switch( propagation_type ) {
    case 0:
      statsfile << "\tCP";
      break;
    case 1:
      statsfile << "\tMDD";
      break;
    case 2:
      statsfile << "\tCP+MDD";
      break;
    }

    switch (search_type) {
    case 0:
      statsfile << "\tLEX";
      break;
    case 1:
      statsfile << "\tDEFAULT";
      break;
    case 5:
      statsfile << "\tDEFAULT+DFS";
      break;
    }

     
    if( propagation_type == 0 ) {
      statsfile << "\tnowidth";
    } else {
      statsfile << "\t" << width;
    }

    if( obj_type != None && feasible ) {
      statsfile << "\t[*" << cp.getValue(obj_var) << "]";
      statsfile << "\t[" << makespan << "]";
      statsfile << "\t[" << sum_setup << "]";
    } else if( feasible ) {
      statsfile << "\t[feasible]";
      statsfile << "\t[" << makespan << "]";
      statsfile << "\t[" << sum_setup << "]";
    } else {
      statsfile << "\t[infeasible]";
    }

    statsfile << "\t" << cp.getInfo(IloCP::NumberOfBranches);
    statsfile << "\t" << cp.getInfo(IloCP::NumberOfFails);

    //if( cp.getTime() >= TIME_LIMIT - 15) {
    if( cp.getInfo(IloCP::FailStatus) == IloCP::SearchStoppedByLimit ) {
      statsfile << "\tTL(" << time_limit << ")";
    } else {
      statsfile << "\t" << cp.getTime();
    }

    // lagrangian info
    statsfile << "\t" << initial_relax;
    statsfile << "\t" << lag_relax;
    statsfile << "\t" << ((double)time_shortestpath)/CLOCKS_PER_SEC;
    statsfile << "\t" << ((double)time_lp)/CLOCKS_PER_SEC;
    statsfile << "\t" << ((double)time_lagrangian)/CLOCKS_PER_SEC;


    
    //if( argc >= 8 ) {
    //  statsfile << "\t" << argv[7];
    //}

    statsfile << endl;
    statsfile.close();

    // print all solutions
    char allsol_filename[256];
    sprintf(allsol_filename, "sols/sol_p%d_w%d_o%d_%s", propagation_type, width, obj_type, inst_name);

    ofstream allsols_file(allsol_filename);
    for( vector<Solution>::iterator sol = all_solutions.begin(); sol != all_solutions.end(); ++sol ) {
      for( int i = 0; i < (int)acts.size(); ++i ) {
	allsols_file << sol->permutation[i] << " ";
      }
      allsols_file << "\t[" << sol->value << "]";
      allsols_file << "\t" << sol->branches;
      allsols_file << "\t" << sol->fails;
      allsols_file << "\t" << sol->time;
      allsols_file << endl;
    }

  } catch (IloException &ex) {
    cout << "CPO Error: " << ex << endl;
  }

  return 0;
}
