/*
 * ------------------------------------------------------------
 * MDD Data Structure Implementations
 * ------------------------------------------------------------
 */

#include "mdd.hpp"


/**
 * Constructor for Reversible MDD
 *
 * Variable ordering is imposed by the order presented in 'vars' argument
 */
RevMDD::RevMDD(IloCP _cp, IlcIntVarArray _vars, int _max_width, bool create_path)
  : cp(_cp), vars(_vars), max_width(_max_width), num_layers(_vars.getSize()+1), time_clean(0)
{
  reserve_memory();
  if( create_path )
    construct_path();
}


/**
 * Allocate MDD memory
 */
void RevMDD::reserve_memory() {

  //cout << "\t[MDD] Reserving memory..." << endl;

  //cout << "\t\tArc buffers...\n";

  min_dom_val = vars.getMinMin();
  max_dom_val = vars.getMaxMax();

  arc_buffer = new( cp.getHeap() ) Arc***[num_layers];
  arc_layer  = new( cp.getHeap() ) Arc**[num_layers];
  max_arc_id = new( cp.getHeap() ) int[num_layers];

  for( int l = 0; l < num_layers; l++ ) {
    int layer_id = 0;
    arc_buffer[l] = new( cp.getHeap() ) Arc**[max_width];
    for( int w = 0; w < max_width; w++ ) {
      arc_buffer[l][w] = new( cp.getHeap() ) Arc*[max_dom_val+1];
      for( int v = min_dom_val; v <= max_dom_val; v++ ) {
	arc_buffer[l][w][v] = new( cp.getHeap() ) Arc(cp, layer_id++, v, w);
      }
    }
    max_arc_id[l] = layer_id;
  }


  for( int l = 0; l < num_layers; l++ ) {
    int layer_id = 0;
    arc_layer[l] = new( cp.getHeap() ) Arc*[max_arc_id[l]];
    for( int w = 0; w < max_width; w++ ) {
      for( int v = min_dom_val; v <= max_dom_val; v++ ) {
	arc_layer[l][layer_id] = arc_buffer[l][w][v];
	layer_id++;
      }
    }
  }


  //cout << "\t\tArc layer list... " << endl;
  arcs_layer = new( cp.getHeap() ) RevLinkedList<Arc>[num_layers];
  for( int l = 0; l < num_layers; l++ ) {
    Arc* header = new( cp.getHeap() ) Arc(cp, -1, -1, -1);
    arcs_layer[l].initialize(cp, header, num_layers * max_width * max_width);
  }

  // other initializations
  aux_set.resize(0, max_dom_val, false);
  has_value = new( cp.getHeap() ) bool[max_dom_val+1];


  // -----------------------------------------------------------
  // Begin Node Branching
  // -----------------------------------------------------------

  // degree of a node
  node_degree = new (cp.getHeap()) int*[num_layers];
  for (int l = 0; l < num_layers; ++l) {
    node_degree[l] = new(cp.getHeap()) int[max_width];
    for (int w = 0; w < max_width; ++w) {
      node_degree[l][w] = -1;
    }
  }

  // -----------------------------------------------------------
  // End Node Branching
  // -----------------------------------------------------------

  domain_checker = new(cp.getHeap()) SimpleChecker(cp, max_dom_val+1);

  //    cout << "\t[MDD] done." << endl;
}


/**
 * Construct initial path MDD
 */
void RevMDD::construct_path() {

  //cout << "\t[MDD] Constructing path MDD...\n";

  // just add arcs corresponding to variables domains in active arc list
  Arc* previous;

  for( int l = 0; l < num_layers-1; l++ ) {

    previous = arcs_layer[l].get_header();

    // add each var domain value
    for( IlcIntExpIterator iter(vars[l]); iter.ok(); ++iter ) {
      previous = arcs_layer[l].extend(previous, get_arc(l, 0, *iter));
      previous->set_target_rev(cp, 0);
    }

    arcs_layer[l].close(previous);
  }

  //cout << "\t[MDD] Exporting to graphs/initial_mdd.gml...\n";
  //export_to_gml("graphs/initial_mdd.gml");

  //    cout << "\t[MDD] done.\n";
}


/**
 * Update variable domains according to changes in arcs
 * TODO: only update layers that had some change
 */
void RevMDD::update_variable_domains() {

  for( int l = 0; l < num_layers-1; l++ ) {
    domain_checker->clear();
    // get the union of arc values
    //memset(has_value, false, sizeof(bool)*(max_dom_val+1));
    for( Arc* arc = arcs_layer[l].get_first(); arc != NULL; arc = arcs_layer[l].get_next(arc) ) {
      //has_value[arc->val] = true;
      domain_checker->add(arc->val);
    }

    // remove values that do not belong to arc domain
    for( IlcIntExpIterator val(vars[l]); val.ok(); ++val ) {
      //if( !has_value[*val] ) {
      if (!domain_checker->contains(*val)) {
	//cout << "\t\t[MDD] Removing val " << *val << " from variable " << l << endl;
	vars[l].removeValue(*val);
      }
    }
  }
}


/**
 * Export to GML file
 */
void RevMDD::export_to_gml(const char* filename) {

  ofstream mdd_file(filename);
  mdd_file << "graph [\n";
  //mdd_file << "comment \"automatically generated MDD graph \"\n";
  mdd_file << "\t directed 1\n";
  mdd_file << "\t hierarchic 1\n";


  // --- first create nodes for each layer ---

  // create remaining nodes
  bool** node_created = new bool*[num_layers];
  for( int i = 0; i < num_layers; i++ ) {
    node_created[i] = new bool[max_width];
    memset(node_created[i], false, sizeof(bool)*max_width);
  }


  for( int l = 0; l < num_layers-1; l++ ) {
    for( Arc* arc = arcs_layer[l].get_first(); arc != NULL; 
	 arc = arcs_layer[l].get_next(arc) ) {
	  
      // fix arc target
      arc->target_node = arc->target_node_id.getValue();

      if( !node_created[l][arc->get_source()] ) {
	node_created[l][arc->get_source()] = true;
	mdd_file << "node [ \n";
	mdd_file << "\t id " << l*max_width + arc->get_source() << "\n";
	mdd_file << "\t label \"" << l << "," << arc->get_source() << "\"\n";

	mdd_file << "\t graphics [ \n";
	mdd_file << "\t\t type \"ellipse\"\n";
	mdd_file << "\t\t hasFill 0 \n";
	mdd_file << "\t\t ] \n";
	mdd_file << "\t ]\n";
      }

      if( !node_created[l+1][arc->get_target()] ) {
	node_created[l+1][arc->get_target()] = true;
	mdd_file << "node [ \n";
	mdd_file << "\t id " << (l+1)*max_width + arc->get_target() << "\n";
	mdd_file << "\t label \"" << (l+1) << "," << arc->get_target() << "\"\n";

	mdd_file << "\t graphics [ \n";
	mdd_file << "\t\t type \"ellipse\"\n";
	mdd_file << "\t\t hasFill 0 \n";
	mdd_file << "\t\t ] \n";
	mdd_file << "\t ]\n";
      }
    }
  }
  for( int i = 0; i < num_layers; i++ ) {
    delete[] node_created[i];
  }
  delete[] node_created;


  // --- now create arcs ---
  for( int l = 0; l < num_layers; l++ ) {
    for( Arc* arc = arcs_layer[l].get_first(); arc != NULL; arc = arcs_layer[l].get_next(arc) ) {
      mdd_file << "edge [ \n";
      mdd_file << "\t source " << l*max_width + arc->get_source() << "\n";
      mdd_file << "\t target " << (l+1)*max_width + arc->get_target() << "\n";
      mdd_file << "\t label \"" << arc->val << "\"\n";
      mdd_file << "\t ]\n";
    }
  }

  //    // create arcs
  //    for( IlcInt l = 0; l < num_layers; l++ ) {
  //        //cout << "layer " << l << " - active_nodes: " << active_nodes[l] << endl;
  //        Node* node = nodes_layer[l].get_first();
  //        while( node != NULL ) {
  //
  //            ArcLL* arc = node->out_arcs.get_first();
  //            while( arc != NULL ) {
  //
  //                mdd_file << "edge [ \n";
  //                mdd_file << "\t source " << l*max_width + node->id << "\n";
  //                mdd_file << "\t target " << (l+1)*max_width + arc->node_id << "\n";
  //                mdd_file << "\t label \"" << *(arc->get_domain()) << "\"\n";
  //                mdd_file << "\t ]\n";
  //
  //                arc = node->out_arcs.get_next(arc);
  //            }
  //
  //            node = nodes_layer[l].get_next(node);
  //        }
  //    }
  //
  mdd_file << "]"; // graph [
  mdd_file.close();
}


/**
 * Clean dangling nodes
 **/
void RevMDD::clean_dangling_nodes(bool fail) {

  /*
   * Initialization: fix arc targets
   */
  for( int l = 0; l < num_layers-1; l++ ) {
    for( Arc* arc = arcs_layer[l].get_first(); arc != NULL; arc = arcs_layer[l].get_next(arc) ) {
      arc->target_node = arc->target_node_id.getValue();
    }
  }


  /**
   * 1. Top-down pass to remove nodes without incoming arcs
   */
  if (time_clean == INF) {
    for (int l = 0; l < get_num_layers(); ++l) {
      for (int w = 0; w < max_width; ++w) {
	node_degree[l][w] = -1;
      }
    } 
    time_clean = 0;
  } else {
    time_clean++;	  
  }
	
  // reset in_arcs of root node
  node_degree[0][0] = time_clean;

  // we keep a pointer to the previous arc considered for removal purposes
  Arc* previous;

  //export_to_gml("before_cleaning.gml");

  for( int l = 0; l < get_num_layers()-1; l++ ) {
    //memset(node_degree[l+1], 0, sizeof(int)*max_width);

    previous = arcs_layer[l].get_header();
    for( Arc* arc = arcs_layer[l].get_first();
	 arc != NULL;
	 arc = arcs_layer[l].get_next(arc) )
      {
	if( node_degree[l][arc->get_source()] != time_clean ) {
	  // if source does not have incoming nodes, we remove the arc from the MDD
	  previous->next.setValue(cp, arc->next.getValue());
	} else {
	  // count number of incoming arcs of the next layer
	  node_degree[l+1][arc->get_target()] = time_clean;
	  previous = arc;
	}
	// fail if resulting layer is empty
	if( arcs_layer[l].is_empty() ) {
	  if (fail) {
	    cp.fail();
	  }
	  cout << "ERROR: Fail in cleaning dangling nodes (incoming): MDD is empty" << endl;
	  //export_to_gml("after_cleaning.gml");
	  exit(1);
	}
      }
  }


  /**
   * 2. Bottom-up pass to remove nodes without outgoing arcs
   */

  if (time_clean == INF) {
    for (int l = 0; l < get_num_layers(); ++l) {
      for (int w = 0; w < max_width; ++w) {
	node_degree[l][w] = -1;
      }
    } 
    time_clean = 0;
  } else {
    time_clean++;	  
  }


  // reset out_arcs of terminal node
  node_degree[get_num_layers()-1][0] = time_clean;

  for (int l = get_num_layers()-2; l >= 0; --l) {
    //memset(node_degree[l], 0, sizeof(int)*get_max_width());

    previous = arcs_layer[l].get_header();
    for( Arc* arc = arcs_layer[l].get_first();
	 arc != NULL;
	 arc = arcs_layer[l].get_next(arc) )
      {
	if( node_degree[l+1][arc->get_target()] != time_clean ) {
	  // if source does not have incoming nodes, we remove the arc from the MDD
	  previous->next.setValue(cp, arc->next.getValue());
	  //cout << "effective!" << endl;
	} else {
	  // count number of incoming arcs of the next layer
	  node_degree[l][arc->get_source()] = time_clean;
	  previous = arc;
	}
      }
    // fail if resulting layer is empty
    if( arcs_layer[l].is_empty() ) {
      if (fail) {
	cp.fail();
      }
      cout << "ERROR: Fail in cleaning dangling nodes (outgoing): MDD is empty" << endl;
      export_to_gml("after_cleaning.gml");
      exit(1);
    }
  }

}
